/**
 * Copyright (c) Members of the EGEE Collaboration. 2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 * Authors: Oscar Koeroo, Mischa Sall\'e, Aram Verstegen
 *      NIKHEF Amsterdam, the Netherlands
 *      <grid-mw-security@nikhef.nl>	
 */

#ifndef _ENVIRON_H_
#define _ENVIRON_H_

#include <unistd.h>
#include <stdarg.h>
extern char **environ;

/* Define environment list type: note that we CANNOT initialize it as
 *  env_t myenv={"..","..",NULL} etc. */
typedef char **env_t;

/* Create type for const char * to ease casts and functions prototypes */
typedef const char *const_string_t;

/* Create type for fully constant environment blob */
typedef const const_string_t *const_env_t;

#ifndef MAXENVVARNAMELEN
#define MAXENVVARNAMELEN	16384
#endif

/**
 * returns number of entries in NULL terminated string array, such as argv and
 * environ(7). In case list is NULL it also returns 0.
 */
size_t cgul_strarrlen(const_env_t list);

/**
 * returns the number of entries in list which start with pattern. If pattern of
 * list is NULL it also returns 0.
 */
size_t cgul_patternentries(const_env_t list, const_string_t pattern);

/**
 * Returns the length of the name part of the var=value pair or -1 when there is
 * no '=' sign.
 */
int cgul_getvarnamelen(const_string_t varname);

/**
 * Finds name for given var=value pair in namevalue. varname is a buffer
 * which will contain the result and has space for maxvarnamelen chars.
 * Returns integer, negative on error, -1 on no '=' sign, -2 name too long;
 * 0 success
 */
int cgul_getvarname(char *varname, const int maxvarnamelen,
		    const_string_t namevalue);

/**
 * Finds index in **src matching variable name, in case of no match (including
 * env==NULL) returns -1
 */
int cgul_getvarindex(const_env_t src, const_string_t varname);

/**
 * Finds the value of varname in src list. Returns NULL when no match, incl.
 * NULL valued src or varname.
 */
char *cgul_getenv_src(env_t src, char *varname);

/**
 * As unsetenv(3) but using var=value pair, specified in namevalue. The maximum
 * var length is MAXVARNAMELEN. See also putenv(3). 
 * Returns integer, 0 on success, negative on error.
 */
int cgul_unsetenvpair(const_string_t namevalue);

/**
 * Removes a name=value pair from dst as in cgul_unsetenvpair() with a 'external
 * environment'.
 */
int cgul_unsetenvpair_dst(env_t *dst, const_string_t namevalue);

/**
 * Clears the entire environment
 * Returns integer, -1 on error, 0 on success
 */
int cgul_clear_env(void);

/*
 * Clears the environment of all variables starting with pattern, case
 * INsensitive.
 * Returns number of cleaned entries, -1 on error.
 */
int cgul_clear_env_pattern(const_string_t pattern);

/**
 * Removes all variables starting with pattern from dst as in
 * cgul_clear_env_pattern() with an 'external environment'.
 * Returns number of cleaned entries, -1 on error.
 */
int cgul_clear_env_pattern_dst(env_t *dst, const_string_t pattern);

/**
 * Creates a duplicate of src. In case *dst!=NULL it will be used and should be
 * big enough. Otherwise space is malloc-ed. dst is only updated upon success.
 * Returns -1 upon error, 0 on success.
 */
int cgul_copy_env(env_t *dst, const_env_t src);

/**
 * Same as copy_env() but first unsets all environment variables starting with
 * MALLOC, case INsensitive. dst is again only updated upon success.
 * Returns -1 upon error, 0 on success.
 */
int cgul_safe_copy_env(env_t *dst, const_env_t src);

/**
 * Creates a list of the var=value pairs from src which have varname in list.
 * If it fails NULL is returned. This can also be done using add_src_list()
 * using dst==NULL, but this is more efficiently coded. All new entries are
 * strdup-ed from those in src.
 */
env_t cgul_copy_src_list(const_env_t src, const_env_t list);

/**
 * Adds to dst the var=value pairs from src which have varname in list.
 * All entries will be strdup-ed from those in src.
 * If *dst is NULL, it will be malloc-ed, otherwise it will be realloc-ed.
 * Upon success returns 0. On error -1 is returned. If part of the environment
 * has been successfully copied, they will stay part of the environment.
 * See also copy_src_list() which is more efficient when dst is NULL !
 */
int cgul_add_src_list(env_t *dst, const_env_t src, const_env_t list);

/**
 * Adds to dst the var=value pairs from src which start with pattern.
 * All entries will be strdup-ed from those in src.
 * If *dst is NULL, it will be malloc-ed, otherwise it will be realloc-ed.
 * Upon success returns the number of matches. On error -1 is returned, and dst
 * of the environment has been successfully copied, they will stay part of the
 * environment.
 * An empty pattern (either "" or NULL) or src is NOT considered error.
 * See also cgul_add_src_list()
 */
int cgul_add_src_pattern(env_t *dst, const_env_t src, const_string_t pattern);

/**
 * adds var=value pair to dst. If the variable already exists, it's entry is
 * replaced. Otherwise *dst is realloc-ed and a new entry is added, the
 * resulting array is returned.
 * Upon success returns 0. In case of error -1 is returned and dst remains
 * unchanged. namevalue is strdup-ed into dst and hence can be freed by the
 * caller (unlike putenv(3p) )
 */
int cgul_add_namevalue(env_t *dst, const_string_t namevalue);

/**
 * creates a name=value pair from fmt and the remaining arguments as a printf
 * statement, then adds the name=value pair to the environment dst
 */
int cgul_env_printf_dst(env_t *dst, const char *fmt, ...);

/**
 * creates a name=value pair from fmt and the remaining arguments as a vprintf
 * statement, then adds the name=value pair to the environment dst
 */
int cgul_env_vprintf_dst(env_t *dst, const char *fmt, va_list ap);

/**
 * adds name=value pair to dst, as in setenv() with a 'external environment' .
 * If the variable already exists, it's entry is replaced. Otherwise *dst is
 * realloc-ed and a new entry is added.
 * Upon success returns 0. In case of error -1 is returned and dst remains
 * unchanged. name and value can be freed by the caller (unlike putenv(3p) )
 */
int cgul_setenv_dst(env_t *dst, const_string_t name, const_string_t value);

/**
 * Removes a variable from dst as in unsetenv() with a 'external environment'.
 */
int cgul_unsetenv_dst(env_t *dst, const_string_t name);

/**
 * Adds to the current environment the var=value pairs from src which have
 * varname in list. Returns number of entries added or -1 on error.
 * NOTE: we do a strdup before the actual putenv, so even though putenv doesn't
 * reserve new memory, we stil can free src !
 */
int cgul_putenv_src_list(const_env_t src, const_env_t list);

/**
 * Adds to the current environment the var=value pairs from src which start with
 * pattern. Returns number of entries added or -1 on error.
 * NOTE: we do a strdup before the actual putenv, so even though putenv doesn't
 * reserve new memory, we stil can free src !
 */
int cgul_putenv_src_pattern(const_env_t src, const_string_t pattern);

#endif
