/**
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Adapted from bsd-closefrom.c appearing in openssh
 *
 * Authors:
 * 2009-
 *      Oscar Koeroo <okoeroo@nikhef.nl>
 *      Mischa Sall\'e <msalle@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *      Gerben Venekamp <venekamp@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 */

/**
 * Copyright (c) 2004-2005 Todd C. Miller <Todd.Miller@courtesan.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "glexec_config.h"
#include "glexec_closefrom.h"

#include <sys/types.h>
#include <sys/param.h>
#include <unistd.h>
#include <stdio.h>
#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif
#include <limits.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#ifdef HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlefd(dirp)n
# ifdef HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# ifdef HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# ifdef HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#ifndef OPEN_MAX
# define OPEN_MAX	256
#endif

#ifdef PATH_MAX
# define _CLOSEFROM_PATH_MAX	PATH_MAX
#else
# define _CLOSEFROM_PATH_MAX	4096
#endif

#ifdef _GLEXEC_CLOSEFROM_NEEDS_INIT
/* Allows calling opendir before setuid to work around old RedHat systems. Note
 * that the filedescriptor belonging to it is automatically closed since opendir
 * implicitly uses FD_CLOEXEC, see opendir(3p) */
static DIR *dirp=NULL;

/*!
 * Opens the directory DIR to /proc/$$/fd or /dev/fd, can be used to work around
 * a old RedHat problem where opendir() gives a permission denied after a
 * userswitch.
 */
void glexec_init_closefrom(void)	{
    char fdpath[_CLOSEFROM_PATH_MAX];
    int len;

#if HAVE_PROC_PID
    /* Check for a /proc/$$/fd directory. */
    len = snprintf(fdpath, sizeof(fdpath), "/proc/%ld/fd", (long)getpid());
#else /* HAVE_DEV_FD */
    /* Check for a /dev/fd directory. */
    len = snprintf(fdpath, sizeof(fdpath), "/dev/fd");
#endif /* HAVE_PROC_PID */
    if (len > 0 && (size_t)len <= sizeof(fdpath))
	dirp=opendir(fdpath);
}

/*!
 * Closes the directory DIR to /proc/$$/fd or /dev/fd, only needed on non-POSIX
 * platforms
 * \return 0 success, 1 failure
 */
int glexec_close_dirp(void)	{
    if (dirp && closedir(dirp))
	return 1;

    return 0;
}
#endif /* _GLEXEC_CLOSEFROM_NEEDS_INIT */

/*
 * Close all file descriptors greater than or equal to lowfd.
 */
#ifdef HAVE_CLOSEFROM
int glexec_closefrom(int lowfd) {
    errno=0;
    /* On SunOS closefrom is void closefrom(int) */
    closefrom(lowfd);
#ifdef sun
    /* On SunOS closefrom has no defined errors (but might mess up errno) */
    return 0;
#else
    return errno==0 ? 0 : 1;
#endif
}
#elif defined HAVE_FCNTL_CLOSEM
int glexec_closefrom(int lowfd) {
    return fcntl(lowfd, F_CLOSEM, 0);
}
#else /* neither HAVE_CLOSEFROM nor HAVE_FCNTL_CLOSEM */
int glexec_closefrom(int lowfd) {
    long fd, maxfd;
    int rc=0;
    /* NOTE: MacOS (certain version) have dirfd as macro */
#if ( HAVE_DIRFD || defined(dirfd) ) && ( HAVE_PROC_PID || HAVE_DEV_FD )
    char *endp;
    struct dirent *dent;
    int fddir=-1;

    /* If dirp has succesfully been opened try getting its file descriptor and
     * use it. Note that opendir automatically enforces setting the FD_CLOEXEC
     * on the filedescriptor if there is one. */
    if (dirp && (fddir=dirfd(dirp)) !=-1 ) {
	while ((dent = readdir(dirp)) != NULL) {
	    errno=0;
	    fd = strtol(dent->d_name, &endp, 10);
	    if (errno==0 &&
		dent->d_name != endp &&
		*endp == '\0' &&
		fd >= 0 &&
		fd < INT_MAX &&
		fd >= lowfd &&
		fd != fddir)
	    {
		if (close((int) fd))
		    rc=1;
	    }
	}
	if (closedir(dirp))
	    rc=1;
    } else
#endif
    {
	/*
	 * Fall back on sysconf() or getdtablesize().  We avoid checking
	 * resource limits since it is possible to open a file descriptor
	 * and then drop the rlimit such that it is below the open fd.
	 */
#ifdef HAVE_SYSCONF
	maxfd = sysconf(_SC_OPEN_MAX);
#else
	maxfd = getdtablesize();
#endif /* HAVE_SYSCONF */
	if (maxfd < 0)
	    maxfd = OPEN_MAX;

	for (fd = lowfd; fd < maxfd; fd++)
	    if (close((int) fd) && errno!=EBADF)
		rc=1;
    }
    return rc;
}
#endif /* neither HAVE_CLOSEFROM nor HAVE_FCNTL_CLOSEM */
