/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 * Authors:
 * 2009-
 *	Oscar Koeroo <okoeroo@nikhef.nl>
 *	Mischa Sall\'e <msalle@nikhef.nl>
 *	NIKHEF Amsterdam, the Netherlands
 * 2006-2009
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 */


/*!
 *  \file glexec_conf.c
 *
 */

#include "glexec_conf.h"

#include "glexec.h"
#include "glexec_util.h"
#include "glexec_log.h"
#include "fileutil/fileutil.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
/* For strcasecmp */
#include <strings.h>

/*************************************************************************/
/* DEFINES/TYPEDEFS                                                      */
/*************************************************************************/

/*!
 *  \brief Single linked list of key value pairs.
 *
 *  Group a key and its value together and remember on which line in
 *  the configuration file they were found.
 */ 
typedef struct glexec_conf_pair_s {
  unsigned int lineno;               /*!<  Line number of the key value pair.*/
  char *key;                         /*!<  Key.*/
  char *value;                       /*!<  Value associated with the key.*/
} glexec_conf_pair_t;


/*************************************************************************/
/* GLOBAL VARIABLES                                                      */
/*************************************************************************/

/*!
 *  \brief Internal representation of a config file.
 *
 *  The internal representation of the configuration file is a list of
 *  glexec_conf_pair_t combined with a maximum size and number of actual
 *  entries.
 *  The size of the table can be determined at run time.
 *  \see glexec_conf_init().
 */
static size_t global_conf_configsize=0;  /*!< Number of elements of the table.*/
static size_t global_conf_configentries=0;/*!< Number of used elements. */
static glexec_conf_pair_t *global_conf_configlist=NULL;	/*!< All elements.*/


/*************************************************************************/
/* FUNCTION PROTOTYPES                                                   */
/*************************************************************************/
 
/*!
 *  Initialisation of the memory area newly allocated for storing the results
 *  of parsing the glexec configuration file. When called after already
 *  initialized, it will do a realloc of the table with the new size.
 *  \param size    Size (number of elements) of the table.
 */
static void glexec_conf_init(size_t size);

/*!
 * Looks up entry for given key.
 * \param key
 * \return key/value pair or NULL when unset
 */
static glexec_conf_pair_t *glexec_conf_entry(const char *key);

/*!
 *  Parse the configuration file and return the result.
 *  \param  config_buffer  char array containing entire config file, will be
 *          destroyed during parsing
 *  \return 0 or -1 on failure
 */
static int glexec_conf_parse_config(char *config_buffer);

/*!
 * Parses line from config buffer: Might be section name (will be set) or
 * key/value pair (will be stored).
 * \param tag	already set or new section name
 * \param line  line from config buffer
 * \param lineno current line (or startline with continuation chars)
 * \return 0 on success, -1 on error
 */
static int glexec_conf_parse_line(char **tag, char *line, unsigned int lineno);

/*!
 *  When a key value pair has successfully been extracted from the configuration
 *  file, it needs to be added to the internal representation of the
 *  configuration file.
 *  \param section  current section
 *  \param key
 *  \param value
 *  \param lineno   line in config file
 *  \return Error.
 */
static int glexec_conf_add_key_value_pair(char *section, char *key, char *value,
				   unsigned int lineno);


/************************************************************************/
/* FUNCTIONS                                                            */
/************************************************************************/

/*!
 *  Read the glexec configuration file.
 *
 *  Before reading is started it is checked if the effective user of the
 *  process has not root privileges. The config file is also checked for
 *  its permissions and ownership.
 *
 *  \param config_file  Name of the config file to read.
 *  \param config       Parsed config file.
 *
 *  \return Error.
 *
 */
GLEXEC_ERROR glexec_conf_read_config(const char *config_file,
			    uid_t unpriv_uid, gid_t unpriv_gid)	{
    const char *real_config_file=config_file ? config_file : GLEXEC_CONFIG_FILE;
    char *config_buffer=NULL;
    int rc1;
    int rc2;
    gid_t gid=getgid(),egid=getegid();

    /* Read the config file */
    rc1=cgul_read_config(real_config_file, &(config_buffer),
			unpriv_uid, unpriv_gid);
    /* Parse the return code */
    switch(rc1)	{
	case 0:
	    break;
	case -1:    /* I/O error */
	    glexec_log(GLEXEC_LOG_ERROR,"Failed to read configfile %s.\n",
		    real_config_file);
	    glexec_log(GLEXEC_LOG_ERROR,"  errno (%d): %s\n",
		    errno, strerror(errno));
	    return GLEXEC_CONFIG_READING_ERROR;
	case -2:    /* privilege-drop error */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Could not drop priviliges when reading config file.\n");
	    return GLEXEC_INSUFFICIENT_PRIVILEGES;
	case -3:    /* permission error (untrusted path) */
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error: the config file is NOT trusted (one of its " \
		"pathcomponents is writable for others than *users* root or %s).\n",
		GLEXEC_USER);
            return GLEXEC_WRONG_PERMISSION;
	case -10:    /* permission error (unconfidential path), only when macro
			DEMAND_CONFIG_IS_CONFIDENTIAL is defined */
	    if (egid!=gid && egid!=unpriv_gid)  {
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: the config file is trusted but NOT confidential " \
		    "(it can be viewed by untrusted users and/or groups, " \
		    "i.e. other than user root or %s, and groupids %u and %u).\n",
		    GLEXEC_USER, unpriv_gid, egid);
	    } else {
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: The config file is trusted but NOT confidential " \
		    "(it can be viewed by untrusted users and/or groups, "\
		    "i.e. other than user root or %s, and groupid %u).\n",
		    GLEXEC_USER, unpriv_gid);
	    }
            return GLEXEC_WRONG_PERMISSION;
	case -4:    /* memory error */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Out of memory while reading config file.\n");
            return GLEXEC_OUT_OF_MEMORY;
	case -5:    /* unknown or safefile error */
	default:
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error while reading config file.\n");
	    return GLEXEC_CONFIG_READING_ERROR;
    }

    /* Now parse the configuration: this will destroy the config_buffer */
    rc2=glexec_conf_parse_config(config_buffer);
    /* Clean up read config file buffer */
    free(config_buffer);

    /* Check return code */
    if (rc2)	{
	glexec_conf_close();
	return GLEXEC_CONFIG_READING_ERROR;
    }

    return GLEXEC_SUCCESS;
}

/*!
 * Looks up value for given key.
 * \param key
 * \return value or NULL when unset.
 */
char *glexec_conf_value(const char *key) {
    glexec_conf_pair_t *p;

    if ((p=glexec_conf_entry(key))!=NULL)   {
	glexec_log(GLEXEC_LOG_DEBUG, "Found key '%s' with value '%s'.\n",
		   key, p->value);
	return p->value;
    }
    glexec_log(GLEXEC_LOG_DEBUG, "Key '%s' was not found.\n", key);
    return NULL;
}

/*!
 * Looks up value for given key which may be yes/true or no/false. When value is
 * different an error is logged and glexec exits.
 * \param key
 * \param def_value: return value when no value is given
 * \return 1 when yes/true, 0 when no/false
 */
int glexec_conf_boolean(const char *key, int def_value) {
    char *value=glexec_conf_value(key);

    /* Handle unset value */
    if (!value)	{
	glexec_log(GLEXEC_LOG_DEBUG,
		"Using default value '%s' for key '%s'\n",
		def_value ? "yes" : "no", key);
	return def_value;
    }

    /* Handle yes */
    if ( (strcasecmp(value, "no")==0) || (strcasecmp(value, "false")==0) )
	return 0;

    /* Handle yes */
    if ( (strcasecmp(value, "yes")==0) || (strcasecmp(value, "true")==0) )
	return 1;

    glexec_log(GLEXEC_LOG_ERROR,
	    "Value '%s' for key '%s' was not understood.\n", value, key);
    glexec_exit(GLEXEC_CONF_VALUE_MALFORMED, GLEXEC_NO_CHILD_EXIT_CODE);

    /* we will never get here. */
    return def_value;
}

/*!
 *  Return the value from the configuration that is associated with the
 *  specified key. When no key is given, or its value cannot be retrieved,
 *  return the default value for this key.
 *  \param key            Search key.
 *  \param default_value  Default value for the search key.
 *  \return Value of the key in case it was found, otherwise the default
 *          value will be returned.
 *
 */
const char *glexec_conf_value_default(const char *key,
				      const char *default_value)   {
    char *value = NULL;

    if (!key || !(value = glexec_conf_value(key))) {
	return default_value;
    }

    return value;
}

/*!
 *  Free the resources allocated with the internal representation of
 *  the glexec config file.
 */
void glexec_conf_close(void) {
    size_t i;
  
    if (global_conf_configsize==0)
	return; /* Nothing to be done */

    for (i=0; i<global_conf_configentries; i++) {
	free(global_conf_configlist[i].key);
	global_conf_configlist[i].key=NULL;
	free(global_conf_configlist[i].value);
	global_conf_configlist[i].value=NULL;
    }
    global_conf_configentries=0;
    free(global_conf_configlist); global_conf_configlist=NULL;
    global_conf_configsize=0;
}


/************************************************************************/
/* INTERNAL FUNCTIONS                                                   */
/************************************************************************/

/*!
 *  Allocate memory area for storing the results of parsing the glexec
 *  configuration file.
 *  \param size    Size (number of elements) of the table.
 */
static void glexec_conf_init(size_t size) {
    /* NOTE: gcalloc never returns on error, just exits */
    if (global_conf_configsize==0)
	global_conf_configlist=
	    (glexec_conf_pair_t *)gmalloc(size*sizeof(glexec_conf_pair_t));
    else
	global_conf_configlist=
	    (glexec_conf_pair_t *)grealloc(global_conf_configlist,size*sizeof(glexec_conf_pair_t));
    
    global_conf_configsize=size;
}

/*!
 * Looks up entry for given key.
 * \param key
 * \return key/value pair or NULL when unset
 */
static glexec_conf_pair_t *glexec_conf_entry(const char *key)	{
    size_t i;

    if (global_conf_configsize==0 || key==NULL)
	/* config uninitialized or key empty */
	return NULL;

    for (i=0; i<global_conf_configentries; i++)   {
	if (strcmp(key, global_conf_configlist[i].key) == 0)
	    return &(global_conf_configlist[i]);
    }
    return NULL;
}

/*!
 *  Parse the configuration file and return the result.
 *  \param  config_buffer  char array containing entire config file, will be
 *          destroyed during parsing
 *  \return 0 or -1 on failure
 */
static int glexec_conf_parse_config(char *config_buffer)	{
    char *config=config_buffer,*section=NULL;
    char *line=NULL,*nextline=config;
    char *pos=NULL,*posdst,*possrc;
    int continuation;
    unsigned lineno=0,baselineno=0;

    /* Malloc memory to hold the list */
    glexec_conf_init(GLEXEC_CONF_ELEM_MAX);

    do	{
	/* Get new line, find end of line, and beginning of next line */
	line=nextline;
	lineno++;
	baselineno=lineno;
	do  {
	    continuation=0;
	    nextline=strchr(line,'\n');
	    if (nextline!=NULL)	{   /* Found end of line */
		*(nextline++)='\0';
		/* If new nextline is end of buffer: it should also be NULL */
		if (*nextline=='\0')
		    nextline=NULL;
	    }

	    /* Remove everything after # or ; */
	    if ((pos=strchr(line,'#'))!=NULL) *pos='\0';
	    if ((pos=strchr(line,';'))!=NULL) *pos='\0';

	    /* Remove leading whitespace */
	    while (*line==' ' || *line=='\t') line++;
	    
	    /* Check there is anything on the line */
	    if (*line=='\0') break;

	    /* Remove trailing whitespace: at this point line has non-white */
	    pos=line+strlen(line)-1;
	    while (*pos==' ' || *pos=='\t') pos--;
	    pos[1]='\0';

	    /* test whether last char is continuation character */
	    if (*pos=='\\')	{
		lineno++;
		continuation=1;
		if (nextline==NULL)	{
		    glexec_log(GLEXEC_LOG_ERROR,
			    "Illegal continuation character at last line %d "
			    "in config file.\n",lineno);
		    return -1;
		}
		/* Find first non-whitespace in nextline (if any) */
		while (*nextline==' ' || *nextline=='\t') nextline++;
		/* Move everything forward from nextline-1 down to line */
		posdst=nextline; possrc=pos;
		while (possrc>line)	{
		    posdst--; possrc--;
		    *posdst=*possrc;
		}
		line=posdst;
	    }
	} while (continuation);

	/* Check there is anything on the line */
	if (*line=='\0') continue;

	/* Now parse the line and optionally store the key/value pair */
	if (glexec_conf_parse_line(&section, line, baselineno))
	    return -1;

    } while (nextline);

    return 0;
}

/*!
 * Parses line from config buffer: Might be section name (will be set) or
 * key/value pair (will be stored).
 * \param tag	already set or new section name
 * \param line  line from config buffer
 * \param lineno current line (or startline with continuation chars)
 * \return 0 on success, -1 on error
 */
static int glexec_conf_parse_line(char **section, char *line, unsigned int lineno)  {
    char *key=NULL,*value=NULL;
    char *pos;
    size_t valuelen;

    /* Now we have line it should be either a tag or key=value pair */
    if (*line=='[')   {	/* Found tag, find closing tag */
	*section=line+1;
	pos=strchr(*section,']');
	if (pos==NULL || pos[1]!='\0')  {  /* Illegal line */
	    glexec_log(GLEXEC_LOG_ERROR,
		"Illegal section at line %u in config file: "
		"no closing ']' or trailing garbage\n", lineno);
	    return -1;
	}
	*pos='\0';
	return 0;
    }

    /* It's not a section tag, so it should be key=value pair */
    if (*section==NULL)  {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Illegal key/value at line %u in config file: "
	    "section is still unset\n",lineno);
	return -1;
    }

    /* Find = character, it should not be the first */
    pos=strchr(line,'=');
    if (pos==NULL || pos==line)  {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Illegal key/value at line %u in config file.\n", lineno);
	return -1;
    }

    /* remove leading whitespace from value */
    value=pos+1;
    while (*value==' ' || *value=='\t') value++;
    /* remove trailing whitespace from key: it cannot be empty,
     * because line has leading whitespace removed */
    key=line;
    pos--;
    while (*pos==' ' || *pos=='\t') pos--;
    pos[1]='\0';
    /* if value is quoted, remove them */
    while ((valuelen=strlen(value))>1)	{
	if (value[0]=='"' && value[valuelen-1]=='"')	{
	    value[valuelen-1]='\0'; value++;
	    valuelen=strlen(value);
	} else if (value[0]=='\'' && value[valuelen-1]=='\'')	{
	    value[valuelen-1]='\0'; value++;
	    valuelen=strlen(value);
	} else
	    break;
    }
    /* Now store tag/key/value pair */
    if (glexec_conf_add_key_value_pair(*section,key,value,lineno))
	return -1;

    return 0;
}

/*!
 *  When a key value pair has successfully been extracted from the configuration
 *  file, it needs to be added to the internal representation of the
 *  configuration file.
 *  \param section  current section
 *  \param key
 *  \param value
 *  \param lineno   line in config file
 *  \return Error.
 */
static int glexec_conf_add_key_value_pair(char *section, char *key, char *value,
				   unsigned int lineno) {
    char *longkey;
    size_t longkeylen=strlen(section)+strlen(key)+2;
    glexec_conf_pair_t *p;

    /* Create long key */
    longkey=gmalloc(longkeylen);
    snprintf(longkey,longkeylen,"%s:%s",section,key);

    /* Check we don't have key yet */
    if ((p=glexec_conf_entry(longkey))!=NULL)  {
	/* Replace */
	glexec_log(GLEXEC_LOG_WARNING,
		"Warning: key %s defined twice at lines %d and %d! "
		"Previous key and value are disregarded.\n",
		key, p->lineno, lineno);
	glexec_log(GLEXEC_LOG_DEBUG,
		"  key='%s', old value = '%s', new value = '%s'\n",
		p->key, p->value, value);
	p->lineno=lineno;
	free(p->value); p->value=gstrdup(value);
	free(longkey);
    } else {
	/* New: check we have space for a new entry */
	if (global_conf_configentries==global_conf_configsize)  {
	    /* Add an additional 10 elements and log on DEBUG it's too small */
	    glexec_log(GLEXEC_LOG_DEBUG,
		    "Adding 10 entries to config table size, "
		    "builtin value=%d is too small\n",
		    global_conf_configsize);
	    glexec_conf_init(global_conf_configsize+10);
	}
	global_conf_configlist[global_conf_configentries].lineno=lineno;
	global_conf_configlist[global_conf_configentries].key=longkey;
	global_conf_configlist[global_conf_configentries].value=gstrdup(value);
	global_conf_configentries++;
    }

    return 0;
}

