/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors:
 * 2009-
 *	Oscar Koeroo <okoeroo@nikhef.nl>
 *	Mischa Sall\'e <msalle@nikhef.nl>
 *	NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *	Gerben Venekamp <venekamp@nikhef.nl>
 *	NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 *  \file glexec_environ.c
 *
 *  Provides environment and current working directory related functions
 */

#include "glexec_environ.h"

#include "glexec.h"
#include "glexec_exit.h"
#include "glexec_log.h"
#include "glexec_util.h"
#include "glexec_conf.h"
#include "environ/environ.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <regex.h>

/************************************************************************/
/* DEFINES	                                                        */
/************************************************************************/

/* Pattern of safe variables for gLExec itself */
#ifndef GLEXEC_ENV_PATTERN
#define GLEXEC_ENV_PATTERN  "GLEXEC_"
#endif

/* Whitelist of safe variables for gLExec itself */
#ifndef GLEXEC_ENV_WHITELIST
#define GLEXEC_ENV_WHITELIST	{"X509_USER_PROXY",NULL}
#endif

/************************************************************************/
/* GLOBAL VARIABLES							*/
/************************************************************************/

/* Master backup for the calling environment, can be freed using
 * glexec_free_env() */
static env_t env_backup=NULL;

/* Environment for the target process, can be obtained using
 * glexec_get_target_env() and freed using glexec_free_env() */
static env_t env_target=NULL;

/* Environment for the epilogue process, can be obtained using
 * glexec_get_epilog_env() and freed using glexec_free_env() */
static env_t env_epilog=NULL;

/* Current working directory at calling time, initialize as "" */
static char caller_cwd[AP_MAXPATH]="";

/************************************************************************/
/* INTERNAL FUNCTION PROTOTYPES                                         */
/************************************************************************/

/*!
 *  Check all environment variables that make it to log files, i.e. their value
 *  is printed in a log file, for illegal characters such as \0 or \n
 */
static int glexec_sane_environment(void);

/*!
 * checks if a variable is safe, i.e. doesn't contain unsafe characters.
 * \return 1 if safe, otherwise 0.
 */
static int glexec_sane_variable(const char *);

/************************************************************************/
/* FUNCTIONS								*/
/************************************************************************/

/*!
 * Free-s the internal static env_t variables (backup of original env, target
 * env and epilog env).
 */
void glexec_free_env(void) {
    /* IMPORT NOTE: we can even free backup, even though some variables have
     * been put in the current environment with putenv, but we have strdupped
     * those */
    glexec_free_list(env_backup);
    glexec_free_list(env_target);
    glexec_free_list(env_epilog);
}

/*!
 * Getter function for the internal target environment
 * \return target environment
 */
env_t glexec_get_target_env(void)   {
    return env_target;
}

/*!
 * Getter function for the internal epilogue environment
 * \return epilogue environment
 */
env_t glexec_get_epilog_env(void)   {
    return env_epilog;
}

/*!
 * Getter function for the internal caller current working directory
 * \return caller current working directory
 */
char *glexec_get_caller_cwd(void)   {
    return caller_cwd;
}

/*!
 * Sets the glexec runtime current working directory to the safe
 * GLEXEC_RUNTIME_CWD instead of the caller's working directory. The original
 * working directory is saved and used for finding the input proxies.
 * Also sets the umask to a safe value, which will be reverted only for the
 * payload.
 * \return new current working directory
 */
const char *glexec_set_runtime_cwd_umask(void)	{
    const char *runtime_cwd=GLEXEC_RUNTIME_CWD;

    /* Set umask to GLEXEC_RUNTIME_UMASK and save old value for payload */
    glexec_set_caller_umask(umask(GLEXEC_RUNTIME_UMASK));

    /* Safe the current working directory in the static variable. */
    if (getcwd(caller_cwd, (size_t)AP_MAXPATH)==NULL)	{
	caller_cwd[0]='\0'; /* prevent undefined values */
	glexec_log(GLEXEC_LOG_ERROR, "Cannot get (and store) current dir: %s.\n",
		    strerror(errno));
	glexec_exit(GLEXEC_STAT_ERROR_CURRENT_WORK_DIR, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Now change to the safe directory */
    if (chdir(runtime_cwd))  {
	glexec_log(GLEXEC_LOG_ERROR,
		"Cannot change to safe directory '%s': %s.\n", runtime_cwd,
		strerror(errno));
	glexec_exit(GLEXEC_ERROR_CHDIR, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    
    glexec_log(GLEXEC_LOG_DEBUG,"Current working directory is now %s.\n",
	    runtime_cwd);

    return runtime_cwd;
}

/*!
 * Prepends value of specified variable with cwd of calling user.
 * \param var should contain
 * \param value optionally the (old) value of the variable, if NULL, it will be
 * obtained dynamically.
 * \param *newvalue new value, when different from value it points to a static
 * buffer overwritten at next call. On error newvalue is unchanged.
 * \return 0 on success, -1 on error, in which case errno is set. When setting
 * the environment fails, gLExec will exit with either GLEXEC_OUT_OF_MEMORY
 * (when applicable) or GLEXEC_SETENV_ERROR.
 */
int glexec_make_var_absolute(const char *var,char *value,char **newvalue) {
    static char filename[AP_MAXPATH];
    int rc_sn;

    /* Check if var is valid, or we might get a SEGFAULT */
    if (var==NULL)  {
	errno=EINVAL;
	return -1; /* error */
    }

    /* If value isn't specified, obtain it */
    if (value==NULL)	{
	/* Check var is set */
	if ( (value=getenv(var))==NULL )    {
	    *newvalue=NULL;
	    return 0; /* var is unset, not error */
	}
    }

    /* If value is already absolute, nothing to do */
    if (value[0]=='/')	{
	*newvalue=value;
	return 0;
    }

    /* Make path absolute */
    rc_sn=snprintf(filename,(size_t)AP_MAXPATH,"%s/%s", caller_cwd,value);
    if (rc_sn<0)    {
	glexec_log(GLEXEC_LOG_ERROR, "Cannot create full path of %s.\n",var);
	errno=EINVAL;
	return -1;
    }
    if (rc_sn>=AP_MAXPATH) {
	glexec_log(GLEXEC_LOG_ERROR, "Full path %s/%s of %s is too long.\n",
		   caller_cwd,value,var);
	errno=ENAMETOOLONG;
	return -1;
    }

    /* Update the variable */
    if (setenv(var, filename, 1))	{
	glexec_log(GLEXEC_LOG_ERROR, "Cannot set %s to %s.\n",var, filename);
	glexec_exit( (errno==ENOMEM ? GLEXEC_OUT_OF_MEMORY : 
				      GLEXEC_SETENV_ERROR),
		    GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Now return the new variable */
    *newvalue=filename;
    return 0;
}

/*!
 * Makes a backup of the original environment, then completely clean the
 * internal gLExec environment except for those in GLEXEC_ENV_WHITELIST or
 * matched against GLEXEC_ENV_PATTERN. It also does some basic sanity checks on
 * the result. This function should be called as soon as possible.
 * \return original environment
 */
env_t glexec_clean_environment(void) {
    /* Whitelist of safe variables for gLExec itself */
    const char *glexec_env_whitelist[]=GLEXEC_ENV_WHITELIST;
 
    /* Backup the environment, so that we can clean it COMPLETELY! */
    if (cgul_safe_copy_env(&env_backup,(const_env_t)environ) < 0)	{
	fprintf(stderr,"[gLExec]:  Error while making backup of the environment.\n");
	glexec_exit(GLEXEC_ENV_CLEAN_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    /* Clean the environment completely */
    if (cgul_clear_env())	{
	fprintf(stderr,"[gLExec]:  Error while cleaning the environment.\n");
	glexec_exit(GLEXEC_ENV_CLEAN_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Reinsert glexec whitelist and pattern into environment */
    if (cgul_putenv_src_list((const_env_t)env_backup, glexec_env_whitelist)<0 ||
	cgul_putenv_src_pattern((const_env_t)env_backup, GLEXEC_ENV_PATTERN)<0) {
	fprintf(stderr,"[gLExec]:  Error while reinserting safe environment.\n");
	glexec_exit(GLEXEC_ENV_CLEAN_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
    }
   
    /* Check the Environment */
    if (!glexec_sane_environment()) {
	glexec_log(GLEXEC_LOG_DEBUG, "There was a problem found with at least one environment variable.\n");
	glexec_exit(GLEXEC_REJECTED_ENV_VAR, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    return env_backup;
}

/*!
 * Sets up a clean new environment for the target environment from the original
 * backup using the whitelist as passed in. NOTE that this variable will be
 * freed.
 * In addition those variables matching GLEXEC_ENV_PATTERN are added.
 * \param preserve_env_variables string containing all to-be-preserved
 * variables, comma separated, NULL or "" when none.
 * \return target environment
 */
env_t glexec_setup_target_environment(const char *preserve_env_variables)  {
    /* Convert the string into a list, needs freeing */
    env_t env_whitelist=glexec_convert_csv_to_list(preserve_env_variables);

    /* Make a clean environment based on the whitelist and the backup */
    env_target=cgul_copy_src_list((const_env_t)env_backup,
				  (const_env_t)env_whitelist);
   
    /* Cleanup the whitelist */
    glexec_free_list(env_whitelist); env_whitelist=NULL;

    /* Check we have been able to setup the env */
    if (env_target == NULL)	{
	glexec_log(GLEXEC_LOG_ERROR, "Cannot create target environment.\n");
	glexec_exit(GLEXEC_TARGET_ENV_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
    }
   
    /*
     * Now add the GLEXEC_ variables to the target environment. Note that
     * GLEXEC_TARGET_PROXY will explicitly be set later if not copied here.
     */
    if (cgul_add_src_pattern(&env_target,(const_env_t)env_backup,GLEXEC_ENV_PATTERN)<0) {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Cannot add "
	    GLEXEC_ENV_PATTERN
	    " variables to target environment.\n");
	glexec_exit(GLEXEC_TARGET_ENV_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    return env_target;
}

/*!
 * Sets up a clean new environment for the epilogue environment from the
 * target environment and original backup explicitly cleaning the variables
 * matching epilog_pattern are added.
 * \param epilog_pattern environment variables blacklist
 * \return epilogue environment
 */
env_t glexec_setup_epilog_environment(const char *epilog_pattern) {
    /* start from target_env, this includes all starting with GLEXEC_ */
    if (cgul_copy_env(&env_epilog,(const_env_t)env_target))	{
	glexec_log(GLEXEC_LOG_ERROR, "Cannot create epilogue environment.\n");
	glexec_exit(GLEXEC_EPILOG_SETUP_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* remove GLEXEC_EPILOG_ variables */
    if (cgul_clear_env_pattern_dst(&env_epilog,epilog_pattern)==-1) {
	glexec_log(GLEXEC_LOG_ERROR,
		"Cannot remove variables \"%s\" from epilogue environment.\n",
		epilog_pattern);
	glexec_exit(GLEXEC_EPILOG_SETUP_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    return env_epilog;
}

/*!
 * Adds variable=int pair to epilog environment
 * \param childrc -1 for none, otherwise >=0
 * \param fmt format string for name=value, remaining arguments are passed as
 * arguments for the format string.
 * \return new epilogue environment
 */
env_t glexec_epilog_printf_env(int childrc, const char *fmt,...)    {
    va_list ap;

    va_start(ap,fmt);
    if (cgul_env_vprintf_dst(&env_epilog,fmt,ap)) {
	va_end(ap);
	glexec_log(GLEXEC_LOG_ERROR,
		"Error adding var/value pair to epilogue env.\n");
	glexec_exit(GLEXEC_PREMATURE_EXIT,
		    childrc>0 ? childrc : GLEXEC_NO_CHILD_EXIT_CODE);
    }
    va_end(ap);
    return env_epilog;
}

/*!
 *  Set an environment variable to the specified value, only if the value found
 *  matches the regular expression
 *  \param var    Environment variable that needs to be settings
 *  \param value  The new value for the environment variable
 *  \param regex  Regular expression that defines valid values.
 *  \return value set or NULL when nothing was set. Note that when the regex
 *  itself is invalid (not when it doesn't match), the value will still be set.
 */
const char *glexec_setenv_regex(const char* var,const char* value,const char *regex) {
    regex_t reg;
    int status;

    /* First compile a regular expression...*/
    if (regcomp(&reg, regex, REG_EXTENDED|REG_NOSUB)==0) { /* Success */
	/* ...and then execute it. */
	status = regexec(&reg, value, (size_t)0, NULL, 0);
	regfree(&reg);

	/* Check the status of the execution. If successful it means that
	 * the check string is valid and we either should do nothing in case
	 * the environment variable was already set, or set the variable when
	 * it did not exist. */
	if (status!=0) {
	    glexec_log(GLEXEC_LOG_WARNING,
		"Value '%s' for environment variable '%s' did not pass the regular expression test '%s'.\n",
		value, var, regex);
	    return NULL;
	}
    } else {
	    /* So, the regular expression turns out to be invalid. This really
	     * shouldn't happen. What can we do but log it, so someone might
	     * figure out what went wrong. Aborting might be a be too harsh
	     * in a production environment. */
	    glexec_log(GLEXEC_LOG_WARNING,
		"Invalid regular expression found '%s'.\n", regex);
    }
	    
    /* Set the environment variable. */
    glexec_log(GLEXEC_LOG_DEBUG,
		"Setting environment variable '%s' to '%s'.\n", var, value);
    setenv(var, value, 1);

    return value;
}

/*!
 * Sets a var - value pair into the target environment
 * \param var variable
 * \param value
 * \return (new) target_environment
 */
env_t glexec_target_setenv(const char *var, const char *value)    {
    /* Handle var==NULL separately */
    if (var==NULL)
	return env_target;

    /* Since var != NULL cgul_setenv_dst can only fail due to out-of-mem */
    if (cgul_setenv_dst(&env_target,var,value) < 0) {
	glexec_log(GLEXEC_LOG_ERROR,
		   "Cannot set %s in target environment.\n",var);
	glexec_exit(GLEXEC_OUT_OF_MEMORY, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    return env_target;
}

/*!
 * Unsets a var from the target environment
 * \param var variable
 * \return (new) target_environment
 */
env_t glexec_target_unsetenv(const char *var)    {
    /* Handle var==NULL separately */
    if (var==NULL || env_target==NULL)
	return env_target;

    /* Since var != NULL cgul_unsetenv_dst cannot fail */
    cgul_unsetenv_dst(&env_target,var);

    return env_target;
}

/*!
 * Sets a var - value pair into the epilogue environment
 * \param var variable
 * \param value
 * \return (new) epilogue environment
 */
env_t glexec_epilog_setenv(const char *var, const char *value)    {
    /* Handle var==NULL separately */
    if (var==NULL)
	return env_epilog;

    /* Since var != NULL cgul_setenv_dst can only fail due to out-of-mem */
    if (cgul_setenv_dst(&env_epilog,var,value) < 0) {
	glexec_log(GLEXEC_LOG_ERROR,
		   "Cannot set %s in epilogue environment.\n",var);
	glexec_exit(GLEXEC_OUT_OF_MEMORY, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    return env_epilog;
}


/************************************************************************/
/* INTERNAL FUNCTIONS                                                   */
/************************************************************************/

/*!
 *  Check all environment variables that make it to log files, i.e. their value
 *  is printed in a log file.
 *
 */
static int glexec_sane_environment(void)
{
  int rc = 1;

  rc &= glexec_sane_variable("GLEXEC_ID");
  rc &= glexec_sane_variable("GLEXEC_SOURCE_PROXY");
  rc &= glexec_sane_variable("GLEXEC_TARGET_PROXY");
  rc &= glexec_sane_variable("X509_USER_PROXY");

  return rc;
}
/*!
 *  Check a string for the occurrence of certain characters. This is
 *  specifically for the checking of environment variables that make it to a log
 *  file. The newline character '\n' is not allowed to appear in it as it allows
 *  reformatting the intended layout of the log file and may cause a potential
 *  exploitation.
 *  \param variable  Variable to be checked
 *  \return true, if the the variable is found to be sane, false otherwise.
 */
static int glexec_sane_variable(const char *variable) {
    int i=0;
    const char *value;

    if (!(value = getenv(variable)))
	return 1;

    while (value[i] != '\0') {
	if (!isprint(value[i])) {
	    glexec_log(GLEXEC_LOG_ERROR,
		"Environment variable '%s' contains %s.\n", variable,
		(value[i] == '\n' ? "a newline '\\n' character" :
				    "non-printable character(s)") );
	    return 0;
	}
	i++;
    }
    return 1;
}
