/**
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors:
 * 2009-
 *      Oscar Koeroo <okoeroo@nikhef.nl>
 *      Mischa Sall\'e <msalle@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *      Gerben Venekamp <venekamp@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 *  \file glexec_epilog.c
 *
 *  Provides functionality for epilogue script
 */

#include "glexec_config.h"

/* Define _XOPEN_SOURCE as the prototype is only enabled in unistd.h when
 * _XOPEN_SOURCE is at least 500. Since we need _POSIX_C_SOURCE 200112L for
 * geteuid, we use 600 which covers both. */
#if HAVE_GETSID && defined(__linux__)
# define _XOPEN_SOURCE	600
#endif

#include "glexec_epilog.h"
#include "glexec_environ.h"
#include "glexec_conf.h"
#include "glexec_account.h"
#include "glexec_log.h"
#include "glexec_exit.h"
#include "glexec_util.h"
#include "glexec_ipc.h"
#include "fileutil/fileutil.h"
#include "realpath/realpath.h"

/* For NULL */
#include <stdio.h>
/* For free(), exit() and strtol */
#include <stdlib.h>
/* For execve, getpid, pid_t getsid */
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
/* For UINT_MAX */
#include <limits.h> 


/************************************************************************/
/* GLOBAL VARIABLES                                                     */
/************************************************************************/

/* Non-NULL if an epilog should run */
static char *epilogpath=NULL;

/* UID of epilogue user: epilogue will run as this user, and (s)he will be
 * trusted for the ownership/writability */
static uid_t epilog_uid=0;

/* Whether a trusted group is specified */
static int epilog_group_present=0;

/* GID of epilogue user: epilogue will run as this group IFF
 * epilog_group_present is non-zero, this group will be then be trusted for the
 * ownership/writability. In any case, for testing the trustedness, this group
 * is used as effective gid */
static gid_t epilog_gid=0;

/*************************************************************************/
/* FUNCTION PROTOTYPES                                                   */
/*************************************************************************/

/*************************************************************************/
/* FUNCTION DECLARATIONS                                                 */
/*************************************************************************/

/*!
 * Setup epilogue:
 * - whether it should run
 * - optional user/group 
 * - whether it's trusted
 * - environment variables
 * \param caller, target account information
 * \param argc, argv payload command
 */
void glexec_setup_epilog(glexec_account_t *caller, glexec_account_t *target,
			 int argc, char *argv[])  {
    int i;
    const char *value;
    char *string;
    struct passwd *pw;
    struct group *gr;
    long int epilog_timeout;

    /* Obtain absolute canonical path for epilogue script */
    if ((epilogpath=glexec_conf_value("glexec:epilogue"))==NULL)
	return;	/* No epilog script specified */
    if (epilogpath[0]!='/')	{
	glexec_log(GLEXEC_LOG_ERROR,"epilog MUST point to an absolute path\n");
	glexec_exit(GLEXEC_EPILOG_SETUP_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Check epilog user */
    if ((value=glexec_conf_value("glexec:epilogue_user"))!=NULL)    {
	if ((pw=glexec_getpwnam(value))==NULL)	{
	    glexec_log(GLEXEC_LOG_ERROR,
		    "User epilog_user=%s doesn't exist on system.\n",value);
	    glexec_exit(GLEXEC_EPILOG_SETUP_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
	}
	epilog_uid=pw->pw_uid;
    }

    /* Check epilog group */
    if ((value=glexec_conf_value("glexec:epilogue_group"))!=NULL)    {
	if ((gr=glexec_getgrnam(value))==NULL)	{
	    glexec_log(GLEXEC_LOG_ERROR,
		    "User epilog_group=%s doesn't exist on system.\n",value);
	    glexec_exit(GLEXEC_EPILOG_SETUP_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
	}
	epilog_group_present=1;
	epilog_gid=gr->gr_gid;
    }

    /* Check epilog timeout, any mistake: consider fatal error */
    if ((value=glexec_conf_value("glexec:epilogue_timeout"))!=NULL)    {
	errno=0;
	epilog_timeout=strtol(value,NULL,10);
	if (errno!=0 || epilog_timeout<0 || (long unsigned)epilog_timeout>UINT_MAX)    {
	    glexec_log(GLEXEC_LOG_WARNING,
		"Cannot convert epilog_timeout \"%s\" to an unsigned int\n",
		value);
	    glexec_exit(GLEXEC_EPILOG_SETUP_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);
	}
	glexec_set_epilog_timeout((unsigned)epilog_timeout);
    }

    /* Test for trustability of epilogue script, for given epilog_uid/_gid, will
     * exit if it fails. This will be tested again just before executing it. */
    if (!glexec_check_epilog_trusted(epilogpath))
	glexec_exit(GLEXEC_EPILOG_SETUP_ERROR,GLEXEC_NO_CHILD_EXIT_CODE);

    glexec_log(GLEXEC_LOG_DEBUG,"Epilogue %s is trusted.\n",
	    epilogpath);

    /* Make sure signal handling code knows it has to run */
    glexec_set_epilog_should_run();

    /* Epilogue is ok, initialize its environment */
    glexec_setup_epilog_environment(GLEXEC_EPILOG_PATTERN);

    /* caller's CWD, USER and GROUP, target's USER and GROUP */
    glexec_epilog_setenv("GLEXEC_EPILOG_GLEXEC_CWD",glexec_get_caller_cwd());
    glexec_epilog_setenv("GLEXEC_EPILOG_GLEXEC_USER",caller->uname);
    glexec_epilog_setenv("GLEXEC_EPILOG_GLEXEC_GROUP",caller->gname);
    glexec_epilog_setenv("GLEXEC_EPILOG_TARGET_USER",target->uname);
    glexec_epilog_setenv("GLEXEC_EPILOG_TARGET_GROUP",target->gname);

    /* GLEXEC_EPILOG_GLEXEC_PID, CWD, USER, GROUP, UID, GID and payload UID,
     * GID, argc */
    glexec_epilog_printf_env(-1,"GLEXEC_EPILOG_GLEXEC_PID=%d",getpid());
#if HAVE_GETSID
    glexec_epilog_printf_env(-1,"GLEXEC_EPILOG_GLEXEC_SID=%d",getsid(0));
#endif
    glexec_epilog_printf_env(-1,"GLEXEC_EPILOG_GLEXEC_PGID=%d",getpgrp());
    glexec_epilog_printf_env(-1,"GLEXEC_EPILOG_GLEXEC_UID=%u",caller->uid);
    glexec_epilog_printf_env(-1,"GLEXEC_EPILOG_GLEXEC_GID=%u",caller->gid);
    glexec_epilog_printf_env(-1,"GLEXEC_EPILOG_TARGET_UID=%u",target->uid);
    glexec_epilog_printf_env(-1,"GLEXEC_EPILOG_TARGET_GID=%u",target->gid);
    glexec_epilog_printf_env(-1,"GLEXEC_EPILOG_ARGC=%d",argc);

    /* GLEXEC_EPILOG_ARGV<N> */
    for (i=0; i<argc; i++)
	glexec_epilog_printf_env(-1,"GLEXEC_EPILOG_ARGV%d=%s",i,argv[i]);

    /* GLEXEC_EPILOG_GLEXEC_SGIDS */
    if ((string=glexec_sgids2str(caller->nsgid,caller->sgid))==NULL) {
	glexec_log(GLEXEC_LOG_ERROR,
		"Error adding variables to epilogue env\n");
	glexec_exit(GLEXEC_EPILOG_SETUP_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    glexec_epilog_setenv("GLEXEC_EPILOG_GLEXEC_SGIDS",string);
    free(string);

    /* GLEXEC_EPILOG_TARGET_SGIDS */
    if ((string=glexec_sgids2str(target->nsgid,target->sgid))==NULL) {
	glexec_log(GLEXEC_LOG_ERROR,
		"Error adding variables to epilogue env\n");
	glexec_exit(GLEXEC_EPILOG_SETUP_ERROR, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    glexec_epilog_setenv("GLEXEC_EPILOG_TARGET_SGIDS",string);
    free(string);
}

/*!
 * Does last minute tests on child status, sets up remaining env variables
 * (depending on child status), tests again for trustedness, then forks and
 * waits for epilog
 * \return epilog exit value, also 0 when no epilogue runs */
void glexec_run_epilog(void)	{
    pid_t pid,childpid;
    char *t_argv[2];
    char *abs_epilogpath;
    int childrc,epilogrc;

    /* Check we have to run (maybe child failed) */
    if (!glexec_get_epilog_should_run())
	return;

    /* Get payload status */
    childpid=glexec_get_childpid();
    childrc=glexec_get_childrc();

    /* Get realpath: absolute without symlinks */
    if ((abs_epilogpath=cgul_realpath(epilogpath, NULL))==NULL)  {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Cannot get canonical absolute name for program '%s'\n",epilogpath);
	glexec_log(GLEXEC_LOG_ERROR, "  errno: %d (%s)\n", errno,strerror(errno));
	glexec_exit(GLEXEC_PREMATURE_EXIT,childrc);
    }
    /* Set arguments */
    t_argv[0]=abs_epilogpath;
    t_argv[1]=NULL;

    /* Add childpid, child process group and childrc */
    glexec_epilog_printf_env(childrc,"GLEXEC_EPILOG_TARGET_PID=%d", childpid);
    glexec_epilog_printf_env(childrc,"GLEXEC_EPILOG_TARGET_PGID=%d",
	    glexec_get_use_setpgid() ? childpid : getpgrp());
    glexec_epilog_printf_env(childrc,"GLEXEC_EPILOG_TARGET_RC=%d", childrc);

    /* Do last minute test on trustedness */
    if (!glexec_check_epilog_trusted(abs_epilogpath)) {
	free(abs_epilogpath);
	glexec_log(GLEXEC_LOG_ERROR,"Epilogue %s is NOT trusted\n",
	    abs_epilogpath);
	glexec_exit(GLEXEC_PREMATURE_EXIT,childrc);
    }

    /* Log */
    if (glexec_can_switch())
	glexec_log(GLEXEC_LOG_DEBUG,"Starting epilogue %s as uid/gid=%u/%u\n",
	    abs_epilogpath,epilog_uid,epilog_gid);
    else
	glexec_log(GLEXEC_LOG_DEBUG,"Starting epilogue %s\n",
	    abs_epilogpath);

    /* Fork off epilog */
    pid=glexec_forkepilog();
    if (pid==0)	{   /* epilogue script */
	/* Change identity to epilog_uid and epilog_gid, when needed. Note: if
	 * we can switch, we are uid/gid=0/0 now */
	if (glexec_can_switch() &&
	    ( setregid(epilog_gid,epilog_gid) ||
	      setreuid(epilog_uid,epilog_uid) )) {
	    glexec_log(GLEXEC_LOG_ERROR,
		"Unexpected error: "
		"we cannot switch to epilog uid/gid=%u/%u: %s\n",
		epilog_uid,epilog_gid,strerror(errno));
	    exit(GLEXEC_EXIT_INTERNAL_ERROR);
	}
	/* Reopen stdout, stderr, and stdin to /dev/null */
	if (freopen("/dev/null","a",stdout)==NULL ||
	    freopen("/dev/null","a",stderr)==NULL ||
	    freopen("/dev/null","r",stdin)==NULL)  {
	    glexec_log(GLEXEC_LOG_ERROR,
		"Redirecting stdin/stdout/stderr for epilogue %s failed: %s\n",
		abs_epilogpath,strerror(errno));
	    exit(GLEXEC_EXIT_INTERNAL_ERROR);
	}
	/* Now execve() */
	execve(abs_epilogpath,t_argv,glexec_get_epilog_env());
	/* Either NEVER come here or execve() was unsuccessful */
	glexec_log(GLEXEC_LOG_ERROR,"Execve of epilogue %s failed: %s\n",
		abs_epilogpath,strerror(errno));
	free(abs_epilogpath);
        exit(GLEXEC_EXECVE_FAILED_EXIT_CODE);
    }
    /* Lingering for epilogue to finish */
    epilogrc=glexec_waitepilog();

    /* No longer need abs_epilogpath */
    free(abs_epilogpath);

    if (epilogrc!=0)
	glexec_exit(GLEXEC_PREMATURE_EXIT, childrc);
}

/*!
 * returns whether epilogue is trusted or not.
 * \return 1 epilogue is trusted
 * \return 0 epilogue is untrusted or other error occurred
 */
int glexec_check_epilog_trusted(char *epilog_filename) {
    int rc;
    uid_t euid=geteuid();
    char *filename=(epilog_filename ? epilog_filename : epilogpath);
    int switching=(euid && glexec_can_switch());

    /* If glexec can switch, we upgrade to effective root */
    if (switching && seteuid(0))  {
	switching=0;
	glexec_log(GLEXEC_LOG_WARNING,
	    "Warning: cannot seteuid(0) to examine epilogue.\n");
    }
    rc=cgul_executable_trusted(filename, epilog_uid, epilog_gid,
			       epilog_group_present);
    if (switching && seteuid(euid)) {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error: cannot seteuid=%d while examining epilogue.\n",euid);
	return 0;
    }

    /* Parse the return code */
    switch(rc) {
	case 0: /* All is fine: return */
	    return 1;
	case -1:
	    glexec_log(GLEXEC_LOG_ERROR,"Failed to examine epilogue %s.\n",
		filename);
	    break;
	case -2:
	    glexec_log(GLEXEC_LOG_ERROR,
		"Unexpected error: "
		"could not drop priviliges while examining epilogue.\n");
	    break;
	case -3:
	    glexec_log(GLEXEC_LOG_ERROR,
		"Epilogue script %s is untrusted! "
		"Should be writable only for root and "
		"optionally epilog_user and epilog_group.\n",
		filename);
	    break;
	case -4:
	    glexec_log(GLEXEC_LOG_ERROR,
                    "Out of memory while examining epilogue %s.\n",
		    filename);
	    break;
        case -5:    /* unknown or safefile error */
        default:
            glexec_log(GLEXEC_LOG_ERROR,
                    "Unknown error while examining epilogue %s.\n",
		    filename);
	    break;
    }
    return 0;
}

