/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 * glexec_exit.c -- exit handling functions for gLExec
 * \file glexec_error.c
 */

#include "glexec_exit.h"

#include "glexec.h"
#include "glexec_log.h"
#include "glexec_conf.h"
#include "glexec_account.h"
#include "glexec_environ.h"
#include "glexec_util.h"
#include "glexec_ipc.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include <errno.h>

/************************************************************************/
/* DEFINES AND TYPES                                                    */
/************************************************************************/
#define ERR_MAX_BUF ((size_t)1024U)

typedef enum GLEXEC_LOG_TO_WHO_E
{
    LOG_TO_ADMIN          = 1,
    LOG_TO_USER           = 2,
    LOG_TO_ADMIN_AND_USER = 3
} GLEXEC_LOG_TO_WHO_T;


/*************************************************************************/
/* FUNCTION PROTOTYPES                                                   */
/*************************************************************************/
/*!
 *  Display error messages based on the error code.
 *  \param rc  Error code.
 *  \param forwho
 */
static void display_error(int rc, int childrc, GLEXEC_LOG_TO_WHO_T forwho);

/*!
 * Converts int input in a character array representation of maximum
 * size maxlen.
 * \param input int
 * \param output character array of size maxlen (including \0)
 * \return length on success, -1 on error
 */
static int int2charstr(int input, char *output, int maxlen);

/*************************************************************************/
/* FUNCTION DECLARATIONS                                                 */
/*************************************************************************/

/*!
 *  Exit glexec and free allocated resources.
 *  \param glexec_internal_rc	Error code leading to the exit.
 *  \param child_return_code	Error code of the child process
 */
void glexec_exit(int glexec_internal_rc, int child_return_code)
{
    sigset_t oldsigset;

    /* Block signals */
    glexec_block_signals(&oldsigset);

    glexec_conf_close();
    glexec_free_cached_passwds();
    glexec_free_cached_groups();
    glexec_free_env();

    /*
     * Decision from Maarten Litmaath for clean and clear error codes
     *
     * 201 - client error
     * 202 - internal error
     * 203 - authz error
     * 
     * To deal with the application using any of those codes one would add this
     * case:
     * 
     * 204 - application exited with code 201, 202, 203 or 204
     */


    /*
     * When internal error code returns GLEXEC_SUCCESS and when the 
     * child_return_code returns GLEXEC_NO_CHILD_EXIT_CODE then this indicates
     * the master gLExec to have exited nicely
     */
    if ((glexec_internal_rc == GLEXEC_SUCCESS) && 
        (child_return_code == GLEXEC_NO_CHILD_EXIT_CODE))
    {
        glexec_close_log();
        exit (0);
    }


    /*
     * When internal error code returns GLEXEC_SUCCESS and when the 
     * child_return_code returns something else, then this is the return code 
     * of the child GLEXEC_NO_CHILD_EXIT_CODE then this indicates the return 
     * of the child gLExec
     */
    if (glexec_internal_rc == GLEXEC_SUCCESS)
    {
        /* Child return code overlap detection */
        switch (child_return_code)
        {
            /* 201, 202, 203, 204 are the error codes used for gLExec */
            case GLEXEC_EXIT_USER_ERROR               :
            case GLEXEC_EXIT_INTERNAL_ERROR           :
            case GLEXEC_EXIT_AUTHZ_ERROR              :
            case GLEXEC_EXIT_CHILD_PROCESS_RC_OVERLAP :
                glexec_close_log();
		/* 204 - application exited with code 201, 202, 203 or 204 */
                exit (GLEXEC_EXIT_CHILD_PROCESS_RC_OVERLAP);
            default :
                glexec_close_log();
                exit (child_return_code);
        }
    }
    else
    {
	/* Switching on the internal gLExec error codes - Everything is an error
	 * from here. */
        switch (glexec_internal_rc)
        {
            case  GLEXEC_NO_SOURCE_PROXY              :
            case  GLEXEC_SETENV_ERROR                 :
            case  GLEXEC_STAT_ERROR                   :
            case  GLEXEC_STAT_ERROR_EXECUTABLE_CMD    :
            case  GLEXEC_STAT_ERROR_CMD_SYMLINKED     :
            case  GLEXEC_STAT_ERROR_CURRENT_WORK_DIR  :
            case  GLEXEC_CHPWD_ERROR                  :
            case  GLEXEC_STAT_CMD_PERMS_CANT_EXEC     :
            case  GLEXEC_COPY_PROXY_FAILED            :
            case  GLEXEC_PROXY_IS_LINK                :
            case  GLEXEC_PROXY_OPEN_FAILED            :
            case  GLEXEC_SOURCE_PROXY_OPEN_FAILED     :
            case  GLEXEC_PROXY_CLOSE_FAILED           :
            case  GLEXEC_FAILURE_WRITING_PROXY        :
            case  GLEXEC_WRONG_OWNERSHIP              :
            case  GLEXEC_PROG_WRONG_OWNERSHIP         :
            case  GLEXEC_PROXY_TOO_LARGE              :
            case  GLEXEC_PRIVILEGED_USER              :
            case  GLEXEC_FILE_CHANGED                 :
            case  GLEXEC_FAILURE_READING_PROXY_ENV    :
            case  GLEXEC_FAILURE_READING_PROXY_FILE   :
            case  GLEXEC_WRONG_SET_OF_ARGS            :
            case  GLEXEC_NO_CHILD_EXIT_CODE           :
            case  GLEXEC_WRONG_PERMISSION_PROXY       :
            case  GLEXEC_WRONG_PERMISSION_ON_CMD      :
	    case  GLEXEC_LOCK_FAILURE                 :
	    case  GLEXEC_USER_TRIED_V		      :
                display_error (glexec_internal_rc, child_return_code,
			       LOG_TO_ADMIN_AND_USER);
		glexec_close_log();
		/* 201 - client error */
                exit (GLEXEC_EXIT_USER_ERROR);

	    case  GLEXEC_ENV_CLEAN_ERROR	      :
	    case  GLEXEC_TARGET_ENV_ERROR	      :
	    case  GLEXEC_EPILOG_SETUP_ERROR	      :
            case  GLEXEC_LCAS_INIT_FAIL               :
            case  GLEXEC_LCMAPS_INIT_FAIL             :
            case  GLEXEC_INCOMPLETE_MAPPING_RESULT    :
            case  GLEXEC_BACKLOG_FAILED               :
            case  GLEXEC_CONF_VALUE_MALFORMED         :
            case  GLEXEC_CONF_FAIL_USER_ID_SWITCH_BY  :
            case  GLEXEC_CONFIG_READING_ERROR         :
            case  GLEXEC_WRONG_PERMISSION             :
            case  GLEXEC_DIR_WRONG_OWNERSHIP          :
            case  GLEXEC_RSL_FAILED                   :
            case  GLEXEC_OUT_OF_MEMORY                :
            case  GLEXEC_ALLOC_ERROR                  :
            case  GLEXEC_COULD_NOT_DROP_PRIVILEGES    :
            case  GLEXEC_DID_NOT_USERSWITCH	      :
	    case  GLEXEC_STILL_EFFECTIVE_ROOT	      :
            case  GLEXEC_WRONG_USERSWITCH	      :
            case  GLEXEC_DOUBLE_USERSWITCH	      :
	    case  GLEXEC_TOO_MANY_GROUPS	      :
            case  GLEXEC_WRONG_NUMBER_OF_BYTES_READ   :
            case  GLEXEC_REJECTED_ENV_VAR             :
            case  GLEXEC_ERROR_CHDIR		      :
            case  GLEXEC_UNKNOWN_USER                 :
            case  GLEXEC_UNKNOWN_GROUP                :
            case  GLEXEC_FAILED_TO_RESTORE_EFF_UID    :
            case  GLEXEC_MAPPED_TO_FORBIDDEN_USER     :
            case  GLEXEC_MAPPED_TO_FORBIDDEN_GROUP    :
            case  GLEXEC_MEM_ALLOC_FAILURE            :
            case  GLEXEC_INSUFFICIENT_PRIVILEGES      :
            case  GLEXEC_SIGNAL_ERROR		      :
            case  GLEXEC_PREMATURE_EXIT		      :
	    case  GLEXEC_DUNGEON_COLLAPSE             :
            case  GLEXEC_FORK_FAILED                  :
            case  GLEXEC_UNCLASIFIED                  :
            case  GLEXEC_LOG_FILE_CLOSE_FAIL          :
                display_error (glexec_internal_rc, child_return_code,
			       LOG_TO_ADMIN);
                glexec_close_log();
		/* 202 - internal error */
                exit (GLEXEC_EXIT_INTERNAL_ERROR);
                
            case  GLEXEC_LOG_FILE_OPEN_FAIL           :
            case  GLEXEC_LOG_FILE_WRONG_PERMISSION    :
            case  GLEXEC_LCAS_FAILED                  :
            case  GLEXEC_LCMAPS_FAILED                :
            case  GLEXEC_USER_NOT_WHITELISTED         :
                display_error (glexec_internal_rc, child_return_code,
			       LOG_TO_ADMIN_AND_USER);
                glexec_close_log();
		/* 203 - authz error */
                exit (GLEXEC_EXIT_AUTHZ_ERROR);

            default :
                /* Undefined error, thus internal error */
                glexec_close_log();
		/* 202 - undefined internal glexec_exit */
                exit (GLEXEC_EXIT_INTERNAL_ERROR);
        }
    }
}

/*!
 * Called when the execve fails: prints proper error message to stdout and
 * glexec log.
 * \param cmd command which was tried.
 * \param cwd current working directory, ignored when cmd is absolute.
 */
void glexec_execve_exit(const char *cwd, char *fullcmd, char* realcmd)	{
    int len,i;
    char cmdline_error[GLEXEC_MAX_LOG_LINE];
    /* list of NULL terminated char* pointers */
    const char *print_err_fmt[8]={NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};

    /* First log to syslog/logfile */
    glexec_log(GLEXEC_LOG_ERROR,
	    "Error: execve of %s (=%s) in %s failed: %s\n",
	    fullcmd,realcmd,cwd,strerror(errno));
   
    /* Form error message for user */
    len=snprintf(cmdline_error,GLEXEC_MAX_LOG_LINE,
		"  Cannot execute: \"%s\" (=\"%s\").\n",fullcmd,realcmd);
    if (len<0)	{ /* snprintf failed */
	strncpy(cmdline_error,
		"  Cannot execute command.\n",GLEXEC_MAX_LOG_LINE);
	cmdline_error[GLEXEC_MAX_LOG_LINE-1]='\0';
    } else if ((unsigned)len>=GLEXEC_MAX_LOG_LINE)
	sprintf(&(cmdline_error[GLEXEC_MAX_LOG_LINE-5]),"...\n");

    /* Print the first line */
    fprintf(stderr, "[gLExec]:%s", cmdline_error);

    switch (errno) {
	case ENOTDIR      :   
	    print_err_fmt[0]=
		"A component of the path prefix is not a directory.\n";
	    break;
	case ENAMETOOLONG :
	    print_err_fmt[0]=
		"A component of a pathname exceeded {NAME_MAX} characters, "
		"or an entire path name exceeded {PATH_MAX} characters.\n";
	    break;
	case ENOENT       :  
	    print_err_fmt[0]="The new process file does not exist.\n";
	    break;
	case ELOOP        : 
	    print_err_fmt[0]=
		"Too many symbolic links were encountered in translating the pathname.\n";
	    break;
	case EACCES       : 
	    print_err_fmt[0]="No execute permission for payload executable.\n";
	    print_err_fmt[1]="Possible reasons:\n";
	    print_err_fmt[2]="- Search permission is denied for a component of the path prefix,\n";
	    print_err_fmt[3]="- The new process file is not an ordinary file,\n";
	    print_err_fmt[4]="- The new process file mode denies execute permission,\n";
	    print_err_fmt[5]="- The new process file is on a filesystem mounted with execution \n";
	    print_err_fmt[6]="  disabled (MNT_NOEXEC in <sys/mount.h>).\n";
	    break;
	case ENOEXEC      :   
	    print_err_fmt[0]=
		"The new process file has the appropriate access permission, "
		"but has an invalid magic number in its header.\n";
	    break;
	case ETXTBSY      :   
	    print_err_fmt[0]=
		"The new process file is a pure procedure (shared text) file "
		"that is currently open for writing or reading by some process.\n";
	    break;
	case ENOMEM       :  
	    print_err_fmt[0]=
		"The new process requires more virtual memory than is allowed "
		"by the imposed maximum (getrlimit(2)).\n";
	    break;
	case E2BIG        :             
	    print_err_fmt[0]=
		"The number of bytes in the new process's argument list is "
		"larger than the system-imposed limit.  This limit is "
		"specified by the sysctl(3) MIB variable KERN_ARGMAX.\n";
	    break;
	case EFAULT       :  
	    print_err_fmt[0]="The new process file has an unexpected size or an unaccessible address.\n";
	    break;
	case EIO         :  
	    print_err_fmt[0]="An I/O error occurred while reading from the file system.\n";
	    break;
	default	     :
	    print_err_fmt[0]="An unknown error occurred during execve().\n";
	    break;
    }
    
    /* Log and print the different messages */
    for (i=0; print_err_fmt[i]!=NULL; i++)
	fprintf (stderr, "           %s", print_err_fmt[i]);

    /* Now still have to close the logfile... */
    glexec_log(GLEXEC_LOG_DEBUG,"Closing the logfile (after execve has failed).\n");
    glexec_close_log();
    /*
     * Tested with bash, 
     * when an error is triggered, 
     * the execve() passed program returns the exit code of '126' when the
     * command can't be executed.
     * Other errors might also be triggered by it now.
     */
    exit(GLEXEC_EXECVE_FAILED_EXIT_CODE);
}

/*!
 *  Exit glexec from within signal handler: don't try to free allocated
 *  resources or close file descriptors.
 *  \param glexec_internal_rc	Error code leading to the exit.
 *  \param childrc Exit value of child or GLEXEC_NO_CHILD_EXIT_CODE when absent
 *  \param signal Signal causing exit or -1 when none.
 */
void glexec_signal_exit(int glexec_internal_rc, int childrc, int signo)
{
    int rc;
    size_t len,i,lineno;
    const char *line[6];  /* Maximum 5 lines of errors */
    char childrc_str[4]; /* Max 255 + \0 */
    char signo_str[3]; /* Max 64 + \0 */

    /* Flush the signal log if applicable */
    glexec_emergency_flush_signallog();

    lineno=0;
    /* Now call exit with the correct value */
    switch (glexec_internal_rc)	{
	case GLEXEC_SUCCESS:
	    /* 201 user error */
            rc=GLEXEC_EXIT_USER_ERROR;
	    line[lineno++]="[gLExec]:  gLExec interrupted by caller";
	    break;
	case GLEXEC_PREMATURE_EXIT:
	    rc=GLEXEC_EXIT_INTERNAL_ERROR;
	    line[lineno++]="[gLExec]:  Premature exit due to signal";
	    break;
	case GLEXEC_SIGNAL_ERROR:
	    rc=GLEXEC_EXIT_INTERNAL_ERROR;
	    line[lineno++]="[gLExec]:  Could not signal child process";
	    break;
	case GLEXEC_UNCLASIFIED:
	    rc=GLEXEC_EXIT_INTERNAL_ERROR;
	    line[lineno++]="[gLExec]:  Unexpected signal received";
	default:
	    /* 202 internal error */
	    rc=GLEXEC_EXIT_INTERNAL_ERROR;
	    line[lineno++]="[gLExec]:  Unknown error during signal handling";
	    break;
    }

    /* Do we have a valid signal? */
    if (signo>0)	{
	line[lineno++]=", signal: ";
	if (int2charstr(signo,signo_str,3)>0)
	    line[lineno++]=signo_str;
	else
	    line[lineno++]="UNKNOWN";
    }

    /* Do we have a valid payload return value? */ 
    if (childrc!=GLEXEC_NO_CHILD_EXIT_CODE)	{
	line[lineno++]=", payload exited with exit code: ";
	if (int2charstr(childrc,childrc_str,4)>0)
	    line[lineno++]=childrc_str;
	else
	    line[lineno++]="UNKNOWN";
    }
    /* End of line */
    line[lineno++]="\n";
   
    /* Print error lines on stderr */
    for (i=0; i<lineno; i++)	{
	for (len=0; line[i][len]!='\0'; len++);
	/* Write on stderr */
	if (write(STDERR_FILENO, (const void*) line[i], len)<0)
	    break;
    }

    /* Now exit: call _exit() instead of exit() since we MUST bypass any set
     * atexit() functions (such as from VOMS) */
    _exit(rc);
}

/*************************************************************************/
/* INTERNAL FUNCTIONS		                                         */
/*************************************************************************/

/*!
 * Converts int input in a character array representation of maximum
 * size maxlen.
 * \param input int
 * \param output character array of size maxlen (including \0)
 * \return length on success, -1 on error
 */
static int int2charstr(int input, char *output, int maxlen) {
    const char *dig="0123456789";
    int i,pos,negative;
    unsigned digit,rest;

    /* Check if there is enough space for even a single digit */
    if (maxlen<2)   {
        output[0]='\0';
        return -1;
    }

    /* Initialize pos, output and rest */
    pos=maxlen-1;
    output[pos]='\0';
    /* handle negative input */
    if (input<0)    {
	negative=1;
	/* input is smaller than 0, so -input > 0 */
	rest=(unsigned int)(-input);
    } else {
	negative=0;
	/* input is >= 0 */
	rest=(unsigned int)input;
    }

    /* loop until we either have no space or rest==0 */
    do {
        digit=rest%10;
        output[--pos]=dig[digit];
        rest=(rest-digit)/10;
    } while (pos && rest);

    /* if there is rest or pos is 0 and negative it doesn't fit */
    if (rest || (pos==0 && negative) )	{
	output[0]='\0';
	return -1;
    }

    /* Negative: add - */
    if (negative)
	output[--pos]='-';

    /* Now shift to beginning */
    for (i=0; i<maxlen-pos; i++)
        output[i]=output[i+pos];

    /* return length of written string including \0 */
    return maxlen-pos;
}

/*!
 *  Display error messages based on the error code.
 *  \param rc  Error code.
 *  \param forwho
 */
static void display_error(int rc, int childrc, GLEXEC_LOG_TO_WHO_T forwho) 
{
    char *envval1,*envval2,*envval3;
    char first_line[ERR_MAX_BUF];
    const char *err_msgs[3]={NULL,NULL,NULL};
    int i;
    uid_t uid;
    struct passwd *pw;

    memset(first_line, 0, ERR_MAX_BUF);

    /* Switching on the internal gLExec error codes - Everything is an error
     * from here. */
    switch (rc)
    {
        /* User error */
        case  GLEXEC_WRONG_SET_OF_ARGS            :
            snprintf (first_line, ERR_MAX_BUF, "  The wrong number of arguments was specified on the command line. Use '%s -h' for more information\n", 
		    GLEXEC_PACKAGE_NAME);
            break;
        case  GLEXEC_NO_SOURCE_PROXY              :
	    envval1=getenv("GLEXEC_SOURCE_PROXY");
            snprintf (first_line, ERR_MAX_BUF, "  The source proxy was not found at '%s'.\n", envval1 ? envval1 : "(unset variable GLEXEC_SOURCE_PROXY)");
            break;
        case  GLEXEC_SETENV_ERROR              :
            snprintf (first_line, ERR_MAX_BUF, "  Error while trying to set environment variable.\n");
            break;
        case  GLEXEC_STAT_ERROR                   :
            snprintf (first_line, ERR_MAX_BUF, "  The stat syscall returned with an error (not clear for what, please file a bug).\n");
            break;
        case  GLEXEC_STAT_ERROR_EXECUTABLE_CMD    :
            snprintf (first_line, ERR_MAX_BUF, "  Error while examining the executable that is about to be executed.\n");
            break;
        case  GLEXEC_STAT_ERROR_CMD_SYMLINKED     :
            snprintf (first_line, ERR_MAX_BUF, "  The program to execute is symlinked.\n");
            break;
        case  GLEXEC_STAT_ERROR_CURRENT_WORK_DIR  :
            snprintf (first_line, ERR_MAX_BUF, "  The stat syscall returned with an error for the current working directory.\n");
            break;
        case  GLEXEC_CHPWD_ERROR  :
            snprintf (first_line, ERR_MAX_BUF, "  Cannot do a chdir to the target user directory.\n");
            break;
        case  GLEXEC_STAT_CMD_PERMS_CANT_EXEC     :
            snprintf (first_line, ERR_MAX_BUF, "  The program to execute is either not executable or the (mapped) account doesn't have sufficient file permissions to execute it.\n");
            break;
        case  GLEXEC_COPY_PROXY_FAILED            :
            snprintf (first_line, ERR_MAX_BUF, "  Copying of the proxy failed.\n");
            break;
        case  GLEXEC_WRONG_PERMISSION_ON_CMD      :
            snprintf (first_line, ERR_MAX_BUF, "  The permissions on the executable are wrong. It must not be writable for the world (could be a dismissed statement), no setuid or setgid and must be executable for the (mapped) user.\n");
            break;
        case  GLEXEC_WRONG_PERMISSION_PROXY:
            snprintf (first_line, ERR_MAX_BUF, "  The permissions on $GLEXEC_SOURCE_PROXY and/or $GLEXEC_CLIENT_CERT are wrong: they may NOT be readable for anyone but the owner.\n");
            break;
        case  GLEXEC_PROXY_IS_LINK         :
            snprintf (first_line, ERR_MAX_BUF, "  The path to $GLEXEC_SOURCE_PROXY or $GLEXEC_CLIENT_CERT seems to be a symlink. Must use a path to the actual file.\n");
            break;
        case  GLEXEC_PROXY_OPEN_FAILED     :
	    envval1=getenv("GLEXEC_CLIENT_CERT");
	    envval2=getenv("GLEXEC_SOURCE_PROXY");
            snprintf (first_line, ERR_MAX_BUF, "  Failed to open $GLEXEC_CLIENT_CERT=%s or $GLEXEC_SOURCE_PROXY=%s.\n",
		    (envval1 ? envval1 : "(NULL)"),
		    (envval2 ? envval2 : "(NULL)"));
            break;
        case  GLEXEC_SOURCE_PROXY_OPEN_FAILED     :
	    envval1=getenv("GLEXEC_SOURCE_PROXY");
            snprintf (first_line, ERR_MAX_BUF, "  Failed to open $GLEXEC_SOURCE_PROXY=%s.\n",
		    (envval1 ? envval1 : "(NULL)"));
            break;
        case  GLEXEC_PROXY_CLOSE_FAILED           :
	    envval1=getenv("GLEXEC_CLIENT_CERT");
	    envval2=getenv("GLEXEC_SOURCE_PROXY");
            snprintf (first_line, ERR_MAX_BUF, "  Failed to close $GLEXEC_CLIENT_CERT=%s or $GLEXEC_SOURCE_PROXY=%s.\n",
		    (envval1 ? envval1 : "(NULL)"),
		    (envval2 ? envval2 : "(NULL)"));
            break;
        case  GLEXEC_FAILURE_WRITING_PROXY        :
            snprintf (first_line, ERR_MAX_BUF, "  Failure writing proxy file\n");
            break;
        case  GLEXEC_WRONG_OWNERSHIP              :
            snprintf (first_line, ERR_MAX_BUF, "  Expecting a different ownership for directory or file.\n");
            break;
        case  GLEXEC_DIR_WRONG_OWNERSHIP              :
            snprintf (first_line, ERR_MAX_BUF, "  Expecting a different ownership for the directory.\n");
            break;
        case  GLEXEC_PROG_WRONG_OWNERSHIP              :
            snprintf (first_line, ERR_MAX_BUF, "  Expecting a different ownership for the program.\n");
            break;
        case  GLEXEC_PROXY_TOO_LARGE       :
            snprintf (first_line, ERR_MAX_BUF, "  The $GLEXEC_CLIENT_CERT or $GLEXEC_SOURCE_PROXY is too large\n");
            break;
        case  GLEXEC_PRIVILEGED_USER              :
            snprintf (first_line, ERR_MAX_BUF, "  (internal problem) gLExec is still running with privileged credentials where they should have been dropped.\n");
            break;
        case  GLEXEC_FILE_CHANGED                 :
            snprintf (first_line, ERR_MAX_BUF, "  gLExec has detected an input file change during the use of the file. It's unknown if this file-jacking was accidental or intentional.\n");
            break;
        case  GLEXEC_FAILURE_READING_PROXY_ENV    :
            snprintf (first_line, ERR_MAX_BUF, "  environment variable $GLEXEC_CLIENT_CERT is empty.\n");
            break;
        case  GLEXEC_FAILURE_READING_PROXY_FILE   :
	    envval1=getenv("GLEXEC_CLIENT_CERT");
            snprintf (first_line, ERR_MAX_BUF, "  Unable to read the proxy file from $GLEXEC_CLIENT_CERT=%s.\n",
			(envval1 ? envval1 : "(NULL)"));
            break;
        case  GLEXEC_LOCK_FAILURE                 :
	    envval1=getenv("GLEXEC_CLIENT_CERT");
	    envval2=getenv("GLEXEC_SOURCE_PROXY");
	    envval3=getenv("GLEXEC_TARGET_PROXY");
	    if (envval3!=NULL)
		snprintf (first_line, ERR_MAX_BUF, "  Failed to lock $GLEXEC_CLIENT_CERT=%s, $GLEXEC_SOURCE_PROXY=%s or $GLEXEC_TARGET_PROXY=%s.\n",
			(envval1 ? envval1 : "(NULL)"),
			(envval2 ? envval2 : "(NULL)"),
			envval3);
	    else
		snprintf (first_line, ERR_MAX_BUF, "  Failed to lock $GLEXEC_CLIENT_CERT=%s, $GLEXEC_SOURCE_PROXY=%s or destination proxy.\n",
			(envval1 ? envval1 : "(NULL)"),
			(envval2 ? envval2 : "(NULL)"));
            break;
	case  GLEXEC_USER_TRIED_V		  :
	    snprintf (first_line, ERR_MAX_BUF, "  User is not permitted to use -V/--defines option.\n");
	    break;
        /* Administrative / internal failure domain */
        case  GLEXEC_CONF_FAIL_LOCK_MECH          :
            snprintf (first_line, ERR_MAX_BUF, "  The value for one of the locking options \"glexec:*lock_mechanism\" is unexpected.\n");
            break;
        case  GLEXEC_LCAS_INIT_FAIL               :
            snprintf (first_line, ERR_MAX_BUF, "  Initialisation of LCAS failed.\n");
            break;
        case  GLEXEC_LCMAPS_INIT_FAIL             :
            snprintf (first_line, ERR_MAX_BUF, "  Initialisation of LCMAPS failed.\n");
            break;
        case  GLEXEC_INCOMPLETE_MAPPING_RESULT    :
            snprintf (first_line, ERR_MAX_BUF, "  Incomplete mapping result returned (could be missing user or group information).\n");
            break;
        case  GLEXEC_BACKLOG_FAILED               :
            snprintf (first_line, ERR_MAX_BUF, "  The backlog function failed to create a backlog entry.\n");
            break;
        case  GLEXEC_CONF_FAIL_USER_ID_SWITCH_BY  :
            snprintf (first_line, ERR_MAX_BUF, "  The value for the option \"user_identity_switch_by\" is unexpected.\n");
            break;
        case  GLEXEC_CONF_VALUE_MALFORMED         :
            snprintf (first_line, ERR_MAX_BUF, "  Configuration file contains values that are not understood.\n");
            break;
        case  GLEXEC_CONFIG_READING_ERROR         :
            snprintf (first_line, ERR_MAX_BUF, "  Could not parse or correctly read the configuration.\n");
            break;
        case  GLEXEC_WRONG_PERMISSION             :
            /* To do, must be split to be specific to the file that has wrong permission bits */
            snprintf (first_line, ERR_MAX_BUF, "  Wrong permissions on the executable that is to be executed or source proxy\n");
            break;
        case  GLEXEC_RSL_FAILED                   :
            snprintf (first_line, ERR_MAX_BUF, "  Generating the RSL string failed.\n");
            break;
        case  GLEXEC_OUT_OF_MEMORY                :
            snprintf (first_line, ERR_MAX_BUF, "  Out of memory\n");
            break;
        case  GLEXEC_ALLOC_ERROR                 :
	    snprintf (first_line, ERR_MAX_BUF, "  Memory allocation error (not out-of-memory)\n");
            break;
        case  GLEXEC_COULD_NOT_DROP_PRIVILEGES    :
	    snprintf (first_line, ERR_MAX_BUF, "  Couldn't drop privileges, gLExec doesn't have sufficient privileges to drop.\n");
            break;
        case  GLEXEC_DID_NOT_USERSWITCH		  :
	    snprintf (first_line, ERR_MAX_BUF, "  LCMAPS should have changed identity, but lcmaps_posix_enf did not run.\n");
            break;
        case  GLEXEC_STILL_EFFECTIVE_ROOT	  :
	    snprintf (first_line, ERR_MAX_BUF, "  Identity has changed, but without dropping effective root bit.\n");
            break;
        case  GLEXEC_WRONG_USERSWITCH		  :
	    snprintf (first_line, ERR_MAX_BUF, "  Identity has been changed to wrong account.\n");
            break;
        case  GLEXEC_DOUBLE_USERSWITCH		  :
	    snprintf (first_line, ERR_MAX_BUF, "  gLExec is configured to do the userswitch, but LCMAPS already changed identity.\n");
            break;
	case  GLEXEC_TOO_MANY_GROUPS		  :
	    snprintf (first_line, ERR_MAX_BUF, "  Too many groups for target user.\n");
	    break;
        case  GLEXEC_WRONG_NUMBER_OF_BYTES_READ   :
            snprintf (first_line, ERR_MAX_BUF, "  Wrong number of bytes read.\n");
            break;
        case  GLEXEC_REJECTED_ENV_VAR             :
            snprintf (first_line, ERR_MAX_BUF, "  Rejected environment variables.\n");
            break;
	case  GLEXEC_ERROR_CHDIR		  :
            snprintf (first_line, ERR_MAX_BUF, "  Chdir failed.\n");
            break;
        case  GLEXEC_UNKNOWN_USER                 :
            snprintf (first_line, ERR_MAX_BUF, "  Unknown Unix user on this system\n");
            break;
        case  GLEXEC_UNKNOWN_GROUP                :
            snprintf (first_line, ERR_MAX_BUF, "  Unknown Unix group on this system\n");
            break;
        case  GLEXEC_FAILED_TO_RESTORE_EFF_UID    :
            snprintf (first_line, ERR_MAX_BUF, "  Failed to restore effective UID in the process\n");
            break;
        case  GLEXEC_MAPPED_TO_FORBIDDEN_USER     :
            snprintf (first_line, ERR_MAX_BUF, "  Failed mapping, mapped to forbidden userid (like a system account)\n");
            break;
        case  GLEXEC_MAPPED_TO_FORBIDDEN_GROUP    :
            snprintf (first_line, ERR_MAX_BUF, "  Failed mapping, mapped to forbidden groupid (like a system account)\n");
            break;
        case  GLEXEC_MEM_ALLOC_FAILURE            :
            snprintf (first_line, ERR_MAX_BUF, "  Memory allocation failure\n");
            break;
        case  GLEXEC_INSUFFICIENT_PRIVILEGES      :
            snprintf (first_line, ERR_MAX_BUF, "  Insufficient privileges (no setuid bit on root for setuid-enabled gLExec?)\n");
            break;
        case  GLEXEC_LOG_FILE_OPEN_FAIL           :
            snprintf (first_line, ERR_MAX_BUF, "  Error: Could not open log file for gLExec.\n");
            break;
        case  GLEXEC_LOG_FILE_WRONG_PERMISSION    :
            snprintf (first_line, ERR_MAX_BUF, "  Permissions or ownership of the log file or its path were found to be wrong.\n");
            break;
        case  GLEXEC_LOG_FILE_CLOSE_FAIL          :
            snprintf (first_line, ERR_MAX_BUF, "  Can't close log file for gLExec?\n");
            break;
        case  GLEXEC_ENV_CLEAN_ERROR              :
            snprintf (first_line, ERR_MAX_BUF, "  Error during cleaning of environment.\n");
            break;
        case  GLEXEC_TARGET_ENV_ERROR             :
            snprintf (first_line, ERR_MAX_BUF, "  Error during creation of target environment.\n");
            break;
        case  GLEXEC_EPILOG_SETUP_ERROR             :
            snprintf (first_line, ERR_MAX_BUF, "  Error during setup of epilog.\n");
            break;
        case  GLEXEC_SIGNAL_ERROR		  :
	    snprintf (first_line, ERR_MAX_BUF, "  Fatal error during signal handling.\n");
	    break;
        case  GLEXEC_PREMATURE_EXIT		  :
	    snprintf (first_line, ERR_MAX_BUF, "  Payload ran (exit value %d) but gLExec could not properly finalize.\n",childrc);
	    break;
        case  GLEXEC_FORK_FAILED                  :
	    if (childrc==GLEXEC_NO_CHILD_EXIT_CODE)
		snprintf (first_line, ERR_MAX_BUF, "  Forking payload process failed.\n");
	    else
		snprintf (first_line, ERR_MAX_BUF, "  Forking epilogue process failed.\n");
            break;
        case  GLEXEC_DUNGEON_COLLAPSE             :
	    snprintf (first_line, ERR_MAX_BUF, "  Some POSIX function did not behave according to the specification.\n");
	    break;
        case  GLEXEC_UNCLASIFIED                  :
            snprintf (first_line, ERR_MAX_BUF, "  Unclassified error found. Please file a bug.\n");
            break;

        /* Authorisation failure */
        case  GLEXEC_LCAS_FAILED                  :
	    snprintf (first_line, ERR_MAX_BUF, "  LCAS failed.\n");
            if (glexec_get_log_to_file())
                err_msgs[0]="The reason can be found in the logfile.\n";
	    else
                err_msgs[0]="The reason can be found in the syslog.\n";
            break;
        case  GLEXEC_LCMAPS_FAILED                :
            snprintf (first_line, ERR_MAX_BUF, "  LCMAPS failed.\n");
            if (glexec_get_log_to_file())
                err_msgs[0]="The reason can be found in the logfile.\n";
	    else
                err_msgs[0]="The reason can be found in the syslog.\n";
            break;
        case  GLEXEC_USER_NOT_WHITELISTED         :
	    uid=getuid();
	    pw=getpwuid(uid);
		snprintf (first_line, ERR_MAX_BUF,
		    "  User %s (uid=%d) is not whitelisted, i.e. may not invoke gLExec.\n",
		    (pw!=NULL && pw->pw_name!=NULL) ? pw->pw_name : "???", uid);
	    if (glexec_get_log_to_file())
		err_msgs[0]="If this is unexpected, ask the sysadmin to check the glexec log \n";
	    else
		err_msgs[0]="If this is unexpected, ask the sysadmin to check the syslog and the\n";
	    err_msgs[1]="permissions of the config file.\n";
            break;

        default:
            snprintf (first_line, ERR_MAX_BUF, "  Unknown error (%d) encountered!\n", rc);

    }

    /*
     *Everything went to the GLEXEC_LOG_ERROR anyway, 
     * now I can state more easily to which stderr or the logfile/syslog
     * And... the output is buffered and limited by the buffersize
     */
    switch (forwho)
    {
        case LOG_TO_ADMIN :
            glexec_log(GLEXEC_LOG_ERROR, "%s", first_line);
            break;
        case LOG_TO_USER :
            fprintf(stderr, "[gLExec]:%s", first_line);
	    for (i=0; err_msgs[i]!=NULL; i++)
		fprintf(stderr, "         %s", err_msgs[i]);
            break;
        case LOG_TO_ADMIN_AND_USER :
            glexec_log(GLEXEC_LOG_ERROR, "%s", first_line);
            fprintf(stderr, "[gLExec]:%s", first_line);
	    for (i=0; err_msgs[i]!=NULL; i++)
		fprintf(stderr, "           %s", err_msgs[i]);
            break;
        default :
            glexec_log(GLEXEC_LOG_ERROR, "%s", first_line);
    }
}

