/**
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors:
 * 2009-
 *      Oscar Koeroo <okoeroo@nikhef.nl>
 *      Mischa Sall\'e <msalle@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *      Gerben Venekamp <venekamp@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 *  \file glexec_ipc.c
 *
 *  Provides functionality for ipc handling: payload and epilogue, signals
 *  handling etc.
 */

#include "glexec_ipc.h"
#include "glexec_exit.h"
#include "glexec_log.h"
#include "glexec_util.h"    /* for glexec_is_switching and umask */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/times.h>

#include <sys/stat.h> /* For open() */
#include <fcntl.h> /* For open() */ 

/************************************************************************/
/* DEFINES AND TYPES                                                    */
/************************************************************************/

/* Child process stages, used for both payload and epilogue */
typedef enum CHLD_STAGE_E
{
    CHLD_WONT_RUN	    = -4,
    CHLD_HAS_ENDED	    = -3,
    CHLD_HAS_NOT_YET_RUN    = -2,
    CHLD_FORK_FAILED	    = -1,
    CHLD_PROC_IS_CHILD	    = 0
} CHLD_STAGE_T;

typedef enum STDIN_OPTS_E
{
    PAYLOAD_STDINCLOSE	    = (unsigned)(1<<0),
    PAYLOAD_TAKETTY	    = (unsigned)(1<<1),
    LINGER_STDINCLOSE	    = (unsigned)(1<<2)
} STDIN_OPTS_T;

typedef enum ALARM_TYPE_E
{
    ALARM_RESET		    = 0,
    ALARM_TERMDELAY	    = 1,
    ALARM_KILLDELAY	    = 2,
    ALARM_EPILOGTIMEOUT	    = 3,
    ALARM_PAUSE		    = 4
} ALARM_TYPE_T;

/************************************************************************/
/* GLOBAL VARIABLES                                                     */
/************************************************************************/

/* Stage of child process (eventual payload)
 * -3	child has ended
 * -2	child has not yet run
 * -1	fork failed
 * 0	process is child
 * >0	pid of child
 */
static pid_t global_ipc_childstage=CHLD_HAS_NOT_YET_RUN;

/* payload child pid, -1 no child, 0 child itself. */
static pid_t global_ipc_childpid=-1;

/* Exit code of child, only valid when global_ipc_childstage==-3 */
static int global_ipc_childrc=GLEXEC_NO_CHILD_EXIT_CODE;

/* Stage of epilogue process
 * -4	epilogue won't run
 * -3	epilogue has ended
 * -2	epilogue has not yet run (but will)
 * -1	fork failed
 * 0	process is epilogue
 * >0	pid of epilogue
 */
static pid_t global_ipc_epilogstage=CHLD_WONT_RUN;

/* epilogue child pid, -1 no child, 0 child itself. */
static pid_t global_ipc_epilogpid=-1;

/* Exit code of epilogue, only valid when epiloguestage==-3 */
static int global_ipc_epilogrc=GLEXEC_NO_CHILD_EXIT_CODE;

/* Whether we can switch: this might change after e.g. lcmaps posix_enf */
static int global_ipc_canswitch=-1;

/* Number of TERM/KILL signals sent so far to the payload.
 * 0 nothing yet, 1 TERM, 2 KILL, 3 kill didn't succeed: premature exit */
static int global_ipc_numsigsent=0;

/* Current alarm(), being updated by glexec_alarm() */
static ALARM_TYPE_T global_ipc_active_alarm=ALARM_RESET;

/* When alarm() updates global_ipc_active_alarm while there is one pending, this
 * flag will save the reason for the pending SIGALRM */
static ALARM_TYPE_T global_ipc_pending_alarm=ALARM_RESET;

/* Once the SIGALRM has been handled and the global_ipc_caught_signal has been
 * set to SIGALRM,  the global_ipc_caught_alarm will contain the reason for the
 * alarm */
static ALARM_TYPE_T global_ipc_caught_alarm=ALARM_RESET;

/* signal caught, determines the action to be taken in the main wait loops. */
static int global_ipc_caught_signal=0;

/* Grace period before which to test for surviving children */
static unsigned global_ipc_termdelay=GLEXEC_TERM_GRACE_TIME;
static unsigned global_ipc_killdelay=GLEXEC_KILL_GRACE_TIME;
static unsigned global_ipc_epilogtimeout=GLEXEC_EPILOG_TIMEOUT;

/* Whether gLExec should setpgid() on its children */
static int global_ipc_usesetpgid=1;

/* Whether gLExec should install handlers for SIGTRAP, SIGPOLL, SIGPROF etc.:
 * -1 unset, 1 set handlers for signals, 0 don't set handlers */
static int global_ipc_extrasighandlers=-1;

/* Whether gLExec will force the payload to run in the background, i.e. whether
 * the lingering gLExec keeps control of the tty. Only relevant
 * when payload has different process group than gLExec and stdin is a tty:
 * either payload or gLExec will have the stdin tty. By default gLExec transfers
 * to payload, this flag keeps it at lingering gLExec and gives payload
 * /dev/null */
static int global_ipc_forcepayloadbg=0;

/* Options concerning payload stdin: whether it should take tty, or whether is
 * should close (reopen to /dev/null) stdin.
 * Values are a combination of PAYLOAD_STDINCLOSE, PAYLOAD_TAKETTY and
 * LINGER_STDINCLOSE */
static unsigned int global_ipc_stdinopts=0;

/* UID of user calling gLExec: if root (s)he may still send signals in
 * post-payload stages */
static uid_t global_ipc_calleruid;

/*************************************************************************/
/* FUNCTION PROTOTYPES                                                   */
/*************************************************************************/

/*!
 * Sends a signal to child pid, only when it's a child process
 * Temporarily blocks all signals (except KILL and STOP) to prevent other signal
 * handlers from calling us a second time.
 * \param pid process ID or process group (when negative) to signal
 * \param signo signal to send
 * \param stage either global_ipc_childstage or global_ipc_epilogstage, used to
 *        test whether that childprocess matches abs(pid) and updated when that
 *        pid isn't there any longer.
 * \return 0 signal delivered
 * \return -1 signal failed
 * \return -2 invalid child pid, cannot kill
 */
static int glexec_send_signal(pid_t pid, int signo, pid_t *stage);

/*!
 * Sends final signal to child(group) from within terminating signal handler.
 * The only potentially signal unsafe call is seteuid()
 * \param pid process ID or process group (when negative) to signal
 * \param signo signal to send
 * \param stage either global_ipc_childstage or global_ipc_epilogstage, used to
 *        test whether that childprocess matches abs(pid) and updated when that
 *        pid isn't there any longer.
 * \return 0 signal delivered
 * \return -1 signal failed
 * \return -2 invalid child pid, cannot kill
 */
static int glexec_emergency_send_signal(pid_t pid, int signo, pid_t *stage);

/*!
 * Handler for ignore signals: log and ignore, e.g. SIGPIPE and SIGTSTP
 * (typically ctrl-z)
 */
static void glexec_ignore_action(int, siginfo_t *, void *);

/*!
 * Handler for SIGCHLD, updates global_ipc_{child,epilog}stage and sets an alarm
 * when a running payload has ended.
 */
static void glexec_child_action(int signo, siginfo_t *info, void *context);

/*!
 * Handler for SIGINT, SIGTERM, SIGXCPU, SIGXFSZ: send children a TERM and set
 * alarm. */
static void glexec_term_action(int, siginfo_t *, void *);

/*!
 * Handler for SIGALRM:
 * - glexec_term_action sets alarm for grace period after which to send a
 *   SIGKILL
 * - glexec_alarm_action sets alarm when a SIGKILL has been sent for a final
 *   graceperiod, after which it ends without further sending signals
 */
static void glexec_alarm_action(int, siginfo_t *, void *);

/*!
 * Handler for other signals: forward to child (not children!)
 */
static void glexec_other_action(int, siginfo_t *, void *);

/*!
 * Handler for SIGBUS, SIGFPE, SIGILL, SIGSEGV when received from kernel.

 * when something is hopelessly wrong (e.g. failed malloc).
 * Sends a SIGTERM to children, without setting alarm and exits.
 */
static void glexec_fatal_action(int, siginfo_t *, void *);

/*!
 * Changes effective and real identity (except secondary groups) to root to
 * prevent receiving of signals. This should be called either when receiving a
 * term-type signal, while there is a payload or when payload has finished.
 */
static void glexec_continue_as_root(void);

/*!
 * Looks for processes in process group pid (when using process groups) and
 * logs if there are processes remaining.
 * \param pid
 */
static void glexec_cleanup_procesgroup(pid_t pid);

/*!
 * Starts (started==0) and stops+logs (started==1) accounting info. When called
 * a second time with started==1, nothing happens.
 * \param started 0 for starting the timing records, 1 for stopping and
 *  printing.
 * \param use_signalversion 0 for normal logging, 1 for glexec_signal_log
 */
static void glexec_accounting(int started, int use_signalversion);

/*!
 * substitute for fprintf(stderr,string) that can be used within signal handlers
 * \param text to be printed on stderr STDERR_FILENO
 * \return return code of write, 0 when text is NULL
 */
static ssize_t glexec_write_stderr(const char *text);

/*!
 * Handles forwarding of signals to payload process. The first line signal
 * handling is done in the signal handlers (_term_action, _alarm_action,
 * _other_action) which set a flag and log using glexec_signal_log().
 * \param signo signal received
 */
static void glexec_forward_to_payload(int signo);

/*!
 * Handles forwarding of signals to epilogue process. The first line signal
 * handling is done in the signal handlers (_term_action, _alarm_action,
 * _other_action) which set a flag and log using glexec_signal_log().
 * \param signo signal received
 */
static void glexec_forward_to_epilog(int signo);

/*!
 * Wrapper around alarm() function, needed to provide SIGALRM handler with
 * information about the cause of the alarm. Note that the handler might be
 * handling a SIGALRM while there is already a new alarm() set.
 * It updates the global_ipc_active_alarm with the newly set alarm(). When there
 * is a pending alarm and global_ipc_pending_alarm is not yet set, that flag is
 * also updated. Finally global_ipc_numsigsent is updated when we set any of the
 * timeout alarms.
 * \param type
 * \return number of seconds left for previous alarm, 0 when unknown type.
 */
static unsigned glexec_alarm(ALARM_TYPE_T type);

/*************************************************************************/
/* FUNCTION DECLARATIONS                                                 */
/*************************************************************************/

/*!
 * Sets caller id into global static.
 * \param global_ipc_calleruid;
 */
void glexec_set_caller_uid(uid_t glexec_global_ipc_calleruid)	{
    global_ipc_calleruid=glexec_global_ipc_calleruid;
}

/*!
 * Sets whether gLExec should call setpgid(0,0) in a child process and hence
 * send TERM and KILL signals to the process group or just the process id.
 * Default is to use setpgid()
 * \param value 1 to use setpgid 0 not to.
 */
void glexec_set_use_setpgid(int value)  {
    global_ipc_usesetpgid=value;
}

/*!
 * Getter function for glexec_set_use_setpgid
 * \return 0 don't use process groups
 * \return 1 use process groups
 */
int glexec_get_use_setpgid(void)  {
    return global_ipc_usesetpgid;
}

/*
 * Sets whether gLExec will force the payload to run in the background, only
 * relevant when payload has different process group than gLExec: either payload
 * or gLExec will have the stdin tty. By default gLExec transfers to payload,
 * this flag keeps it at lingering gLExec and gives payload /dev/null */
void glexec_set_payload_should_run_in_background(int value) {
    global_ipc_forcepayloadbg=value;
}

/*!
 * Sets grace time after sending TERM to child process, before sending KILL.
 * Default is GLEXEC_TERM_GRACE_TIME, 0 doesn't time-out
 * \param delay in seconds
 */
void glexec_set_term_delay(unsigned delay)  {
    global_ipc_termdelay=delay;
}

/*!
 * Sets grace time after sending KILL to child process, before exiting.
 * Default is GLEXEC_KILL_GRACE_TIME, 0 doesn't time-out (don't do that!)
 * \param delay in seconds
 */
void glexec_set_kill_delay(unsigned delay)  {
    global_ipc_killdelay=delay;
}

/*!
 * Sets grace time after start of epilogue script, before sending it a SIGTERM.
 * Default is GLEXEC_EPILOG_TIMEOUT, 0 doesn't time-out (not advisable)
 * \param delay in seconds
 */
void glexec_set_epilog_timeout(unsigned timeout)  {
    global_ipc_epilogtimeout=timeout;
}

/*!
 * Sets flag to make sure epilogue will run, if already called, or child is
 * already being forked, it will be ignored.
 */
void glexec_set_epilog_should_run(void)	{
    if (global_ipc_epilogstage==CHLD_WONT_RUN &&
	global_ipc_childstage==CHLD_HAS_NOT_YET_RUN)
	global_ipc_epilogstage=CHLD_HAS_NOT_YET_RUN;
}

/*!
 * Getter function for epilogue running...
 * \return 0 epilogue will not run
 * \return 1 other cases, including failed, finished etc.
 */
int glexec_get_epilog_should_run(void)    {
    return (global_ipc_epilogstage==CHLD_WONT_RUN ? 0 : 1);
}

/*!
 * Sets whether gLExec should call set handlers for signals used by debuggers
 * and the like, such as SIGTRAP.
 * \param value 1 to install custom handlers, 0 not to set them.
 */
void glexec_set_use_extra_sighandlers(int value)  {
    global_ipc_extrasighandlers=value;
}

/*!
 * Getter whether glexec_ipc thinks glexec can switch. This is only reliable
 * after glexec_install_sig_handlers() has been called
 * \return 1 when glexec can switch 0 otherwise */
int glexec_can_switch(void) {
    return global_ipc_canswitch;
}

/*!
 * Getter function for child PID
 * \return pid_t of child process
 */
pid_t glexec_get_childpid(void)	{
    return global_ipc_childpid;
}

/*!
 * Getter function for epilogue PID
 * \return pid_t of child process
 */
pid_t glexec_get_epilogpid(void)	{
    return global_ipc_epilogpid;
}

/*!
 * Getter function for child exit code
 * \return exit code of payload process
 */
int glexec_get_childrc(void)	{
    return global_ipc_childrc;
}


/*!
 * Installs signal handlers
 */
void glexec_install_sig_handlers(void)	{
    struct sigaction act;

    glexec_log(GLEXEC_LOG_DEBUG,"Installing signal handlers\n");

    /* Make sure there are no previous alarms */
    glexec_alarm(ALARM_RESET);

    /* When not yet set, set the global_ipc_canswitch flag. We are normally
     * pre-lcmaps here, so switching mode is reliable. */
    if (global_ipc_canswitch==-1)
	global_ipc_canswitch=glexec_is_switching();

    /* Setup general blocking mask: all */
    if (sigfillset(&act.sa_mask))    {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }
    act.sa_flags=SA_SIGINFO | SA_RESTART; /* Make sure to restart after
					     interrupt for e.g. waitpid() */

    /* Setup different handlers:
     * - Start with segv type handlers
     * - Remove the core type from the mask set and install the term handlers
     * - install the handlers only needed once we have a payload
     */

    /* SIGCHLD type */
    act.sa_sigaction=glexec_child_action;
    if (sigaction(SIGCHLD,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }

    /* ABORT type: SIGABRT, SIGSYS
     * CORE type: SIGBUS, SIGFPE, SIGILL, SIGSEGV and if present SIGSTKFLT */
    act.sa_sigaction=glexec_fatal_action;
    if (sigaction(SIGABRT,&act,NULL) ||
	sigaction(SIGSYS,&act,NULL) ||
	sigaction(SIGBUS,&act,NULL) ||
	sigaction(SIGFPE,&act,NULL) ||
#ifdef SIGSTKFLT
	sigaction(SIGSTKFLT,&act,NULL) ||
#endif
	sigaction(SIGILL,&act,NULL) ||
	sigaction(SIGSEGV,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }

    /* TERM type: SIGINT, SIGTERM, SIGXCPU, SIGXFSZ. */
    act.sa_sigaction=glexec_term_action;
    if (sigaction(SIGINT,&act,NULL) ||
	sigaction(SIGTERM,&act,NULL) ||
	sigaction(SIGXCPU,&act,NULL) ||
	sigaction(SIGXFSZ,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }

    /* ALARM type: SIGALRM */
    act.sa_sigaction=glexec_alarm_action;
    if (sigaction(SIGALRM,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }

    /* IGNORE type */
    act.sa_sigaction=glexec_ignore_action;
    if (sigaction(SIGPIPE,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }

    /* OTHER type: include SIGTSTP (ctrl-z in terminal) and SIGCONT, also
     * I/O related ones such as SIGPOLL and SIGURG */
    act.sa_sigaction=glexec_other_action;
    if (sigaction(SIGHUP,&act,NULL) ||
	sigaction(SIGQUIT,&act,NULL) ||
	sigaction(SIGTSTP,&act,NULL) ||
	sigaction(SIGCONT,&act,NULL) ||
	sigaction(SIGURG,&act,NULL) ||
#ifdef SIGPOLL
	/* Note: SIGPOLL though POSIX is synonym for SIGIO on Linux and does not
	 * exist on MacOS */
	sigaction(SIGPOLL,&act,NULL) ||
#endif
#ifdef SIGIO
	/* Note: SIGPOLL and SIGIO on Linux are the same */
	sigaction(SIGIO,&act,NULL) ||
#endif
#ifdef SIGPWR
	/* Set a handler for SIGPWR when available, since it's it could lead to
	 * terminate on Linux: log and forward when possible */
	sigaction(SIGPWR,&act,NULL) ||
#endif
#ifdef SIGWINCH
	/* Set a handler for SIGWINCH when available, since it's it might be
	 * useful for the payload: log and forward when possible */
	sigaction(SIGWINCH,&act,NULL) ||
#endif
	sigaction(SIGUSR1,&act,NULL) ||
	sigaction(SIGUSR2,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }
    
    /* Make sure to really ignore SIGTTIN and SIGTTOU or we can't do jobcontrol
     * on the payload */
    act.sa_flags=0;
    act.sa_handler=SIG_IGN;
    if (sigaction(SIGTTIN,&act,NULL) ||
	sigaction(SIGTTOU,&act,NULL))	{
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }

}

/*!
 * Install optional extra signal handlers. We only can install these after
 * having set the flag global_ipc_extrasighandlers using the
 * glexec_set_use_extra_sighandlers setter function.
 */
void glexec_install_extra_sig_handlers(void) {
    struct sigaction act;
#if defined(SIGRTMIN) && defined(SIGRTMAX)
    int signo;
#endif

    if (global_ipc_extrasighandlers!=1)
	return;

    /* Setup general blocking mask: all but non-blockable ones */
    if (sigfillset(&act.sa_mask))   {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing extra signal handlers: %s\n",
	    strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }
    act.sa_flags=SA_SIGINFO | SA_RESTART; /* Make sure to restart after
					     interrupt for e.g. waitpid() */
    /* OTHER type: include SIGTSTP (ctrl-z in terminal) and SIGCONT, also
     * I/O related ones such as SIGPOLL and SIGURG */
    act.sa_sigaction=glexec_other_action;

    /* Debug type handlers: SIGTRAP etc. These might prevents the use of a
     * debugger: log and forward when possible */
#ifdef SIGEMT
    glexec_log(GLEXEC_LOG_DEBUG,
	"Installing signal handlers for SIGTRAP, SIGEMT, SIGVTALRM, SIGPROF\n");
#else
    glexec_log(GLEXEC_LOG_DEBUG,
	"Installing signal handlers for SIGTRAP, SIGVTALRM, SIGPROF\n");
#endif
    if (sigaction(SIGTRAP,&act,NULL) ||
#ifdef SIGEMT
	sigaction(SIGEMT,&act,NULL) ||
#endif
	sigaction(SIGVTALRM,&act,NULL) ||
	sigaction(SIGPROF,&act,NULL))   {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error while initializing signal handlers: %s\n",
	    strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }

    /* Real-time signal handlers. These might interfere with profilers and tools
     * such as valgrind */
#if defined(SIGRTMIN) && defined(SIGRTMAX)
    glexec_log(GLEXEC_LOG_DEBUG,
	"Installing realtime signal handlers (SIGRTMIN=%d, SIGRTMAX=%d).\n",
	SIGRTMIN,SIGRTMAX);
    for (signo=SIGRTMIN; signo<=SIGRTMAX; signo++)	{
	/* Install Real-time signal handler, but don't exit on failure */
	if (sigaction(signo,&act,NULL))
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error while initializing real-time sighandler for %d: %s\n",
		signo,strerror(errno));
    }
#endif
}

/*!
 * Blocks (in addition to whatever is already blocked) all signals except for:
 * - unblockable: SIGKILL, SIGSTOP
 * - SIGABRT 
 * - core-type signals: SIGBUS, SIGFPE, SIGILL, SIGSEGV, SIGSYS
 * \param oldsigset previous signal mask
 */
void glexec_block_signals(sigset_t *old_blocked_signals_set)	{
    static sigset_t blocked_signals_set;
    static int blocked_init=0;

    if (!blocked_init)	{ /* Whether blocked set is initialized */
	/* Setup blocked set */
	if (sigfillset(&blocked_signals_set) ||
	    sigdelset(&blocked_signals_set,SIGKILL) ||
	    sigdelset(&blocked_signals_set,SIGSTOP) ||
	    sigdelset(&blocked_signals_set,SIGABRT) ||
	    sigdelset(&blocked_signals_set,SIGBUS) ||
	    sigdelset(&blocked_signals_set,SIGFPE) ||
	    sigdelset(&blocked_signals_set,SIGILL) ||
	    sigdelset(&blocked_signals_set,SIGSEGV) ||
	    sigdelset(&blocked_signals_set,SIGSYS))
	{
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error: cannot setup blocked signal set: %s",
		strerror(errno));
	    glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
	}
	blocked_init=1;
    }
    /* Now block the set and save the old set, note use SIG_BLOCK instead of
     * SIG_SETMASK as we might have blocked some which would become unblocked */
    if (sigprocmask(SIG_BLOCK,&blocked_signals_set,old_blocked_signals_set)) {
	glexec_log(GLEXEC_LOG_ERROR,
		"Error: cannot block signal set: %s", strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }
}

/*!
 * Blocks (in addition to whatever is already blocked) all signals:
 * \param oldsigset previous signal mask
 */
void glexec_block_allsignals(sigset_t *old_blocked_signals_set)	{
    static sigset_t blocked_signals_set;
    static int blocked_init=0;

    if (!blocked_init)	{ /* Whether blocked set is initialized */
	/* Setup blocked set */
	if (sigfillset(&blocked_signals_set))   {
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error: cannot setup blocked signal set: %s",
		strerror(errno));
	    glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
	}
	blocked_init=1;
    }
    /* Now block the set and save the old set */
    if (sigprocmask(SIG_BLOCK,&blocked_signals_set,old_blocked_signals_set)) {
	glexec_log(GLEXEC_LOG_ERROR,
		"Error: cannot block signal set: %s", strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }
}

/*!
 * Set blocked signal mask to sigset, exit when fails
 * \param sigset signal set to block
 */
void glexec_sigprocmask(const sigset_t *masksigset) {
    if (sigprocmask(SIG_SETMASK,masksigset,NULL))  {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Error: cannot set signal mask: %s", strerror(errno));
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }
}

/*!
 * Forks of the child process, sets the global_ipc_childpid and start the
 * accounting.
 * Failure is handled internally.
 * \return 0 child
 * \return pid parent
 */
pid_t glexec_forkchild(void)	{
    sigset_t oldset;
    pid_t glexec_pgrp, stdin_pgrp;

    /* First figure out what to do with STDIN, both for gLExec and payload */

    /* If stdin is a tty:
     * - glexec should keep it open.
     * - payload: when sharing pgids, do nothing special, when different pgids:
     *   take over the controlling tty, unless force_payload_background */
    if (isatty(STDIN_FILENO))	{
	/* Has payload different process group? */
	if (global_ipc_usesetpgid)    {
	    /* -b cmdline or force_payload_background in glexec.conf? */
	    if (global_ipc_forcepayloadbg) {
		/* backgrounding forced: lingering gLExec will keep control */
		global_ipc_stdinopts=PAYLOAD_STDINCLOSE;
		glexec_log(GLEXEC_LOG_DEBUG,
		    "STDIN is a tty, but payload should run in background, "
		    "payload will have /dev/null\n");
	    } else { /* If glexec owns STDIN tty: set it on child */
		glexec_pgrp=getpgrp();
		stdin_pgrp=tcgetpgrp(STDIN_FILENO);
		if (stdin_pgrp==glexec_pgrp) { /* stdin is tty, but not ours */
		    global_ipc_stdinopts=PAYLOAD_TAKETTY;
		    glexec_log(GLEXEC_LOG_DEBUG,
			"Payload will take over stdin, which is a tty owned "
			"by gLExec's pgid %d\n",glexec_pgrp);
		} else {
		    global_ipc_stdinopts=PAYLOAD_STDINCLOSE;
		    glexec_log(GLEXEC_LOG_NOTICE,
			"STDIN is a tty owned by pgid %d "
			"while gLExec's pgid=%d, payload will have /dev/null\n",
			stdin_pgrp,glexec_pgrp);
		}
	    }
	} else	/* No separate process group: no job control, keep stdin open */
	    global_ipc_stdinopts=0;
    } else  {
	/* No tty: child keeps open stdin, lingering gLExec closes stdin */
	global_ipc_stdinopts=LINGER_STDINCLOSE;
    }

    /* Block signals for duration of fork() */
    glexec_block_signals(&oldset);

    /* Reset all signal flags */
    glexec_alarm(ALARM_RESET);
    global_ipc_caught_alarm=global_ipc_pending_alarm=ALARM_RESET;
    global_ipc_caught_signal=global_ipc_numsigsent=0;

    /* Actual fork */
    global_ipc_childstage=fork();

    if (global_ipc_childstage<0)	{ /* error */
	glexec_log(GLEXEC_LOG_ERROR ,"Fork failed: %s.\n",strerror(errno));
	glexec_exit(GLEXEC_FORK_FAILED, global_ipc_childrc);
    }

    /* fork() succeeded: set childpid */
    global_ipc_childpid=global_ipc_childstage;

    if (global_ipc_childstage==CHLD_PROC_IS_CHILD)  {   /* actual child */
	/* Child should do a setpgid: do before unblocking signal */
	if (global_ipc_usesetpgid && setpgid(0,0))   {
	    glexec_log(GLEXEC_LOG_ERROR ,"Error: setpgid() failed: %s.\n",
		    strerror(errno));
	    /* Now exit with a GLEXEC_EXIT_INTERNAL_ERROR, which will trigger a
	     * GLEXEC_EXIT_CHILD_PROCESS_RC_OVERLAP, but it's the only
	     * reasonable value. */
	    exit(GLEXEC_EXIT_INTERNAL_ERROR);
	}
	/* Handle the stdin of the payload process */
	if (global_ipc_stdinopts & PAYLOAD_TAKETTY)	{
	    /* Take tty, failure not considered fatal, but unexpected: warn */
	    if (tcsetpgrp(STDIN_FILENO, getpgrp()))
		glexec_log(GLEXEC_LOG_WARNING,
		    "Payload cannot grab control of tty: %s\n",strerror(errno));
	    else
		glexec_log(GLEXEC_LOG_DEBUG,
		    "Payload has taken over tty from gLExec process.\n");
	} else	{
	    if (global_ipc_stdinopts & PAYLOAD_STDINCLOSE)
		/* Reopen /dev/null as stdin, failure not considered fatal but
		 * unexpected: warn */
		if (freopen("/dev/null","r",stdin)==NULL)
		    glexec_log(GLEXEC_LOG_WARNING,
			"Redirecting stdin for payload failed.\n");
	}
    } else {	/* Lingering gLExec */
	/* Initialize accounting time in parent */ 
	glexec_accounting(0,0);
	
	/* Now also close stdin (reopen from /dev/null). Do this here, so that
	 * we can easily send a signal to just created child in case of failure
	 * */
	if (global_ipc_stdinopts & LINGER_STDINCLOSE)	{
	    /* Reopen /dev/null as stdin, failure is considered fatal. */
	    if (freopen("/dev/null","r",stdin)==NULL)  {
		glexec_log(GLEXEC_LOG_ERROR,
		    "Redirecting stdin for lingering gLExec failed, sending "
		    "SIGTERM to payload process %s: %s\n",
		    global_ipc_usesetpgid ? "group" : "id",strerror(errno));
		/* Send signal */
		glexec_send_signal(
		    global_ipc_usesetpgid ? -global_ipc_childpid :
					    global_ipc_childpid, SIGTERM,
					    &global_ipc_childstage);
		/* exit */
		glexec_exit(GLEXEC_PREMATURE_EXIT,global_ipc_childrc);
	    }
	}
    }
    /* unblock signals in both parent and child */
    glexec_sigprocmask(&oldset);

    return global_ipc_childpid;
}

/*!
 * Forks of the epilog process, sets the global_ipc_epilogpid and start the
 * accounting.
 * Failure is handles internally.
 * \return 0 epilog
 * \return pid parent
 */
pid_t glexec_forkepilog(void)	{
    sigset_t oldset;

    /* Block signals for duration of fork() */
    glexec_block_signals(&oldset);

    /* Reset all signal flags */
    glexec_alarm(ALARM_RESET);
    global_ipc_caught_alarm=global_ipc_pending_alarm=ALARM_RESET;
    global_ipc_caught_signal=global_ipc_numsigsent=0;

    /* Actual fork */
    global_ipc_epilogstage=fork();

    if (global_ipc_epilogstage<0)	{ /* error */
	glexec_log(GLEXEC_LOG_ERROR ,"Fork failed: %s.\n",strerror(errno));
	glexec_exit(GLEXEC_FORK_FAILED, global_ipc_childrc);
    }

    /* fork() succeeded: set epilogpid */
    global_ipc_epilogpid=global_ipc_epilogstage;

    if (global_ipc_epilogstage==CHLD_PROC_IS_CHILD)	{  /* epilogue itself */
	/* Child should do a setpgid: do before unblocking signal */
	if (global_ipc_usesetpgid && setpgid(0,0))   {
	    glexec_log(GLEXEC_LOG_ERROR ,"Error: setpgid() failed: %s.\n",
		    strerror(errno));
	    /* No reasonable exit value other than just
	     * GLEXEC_EXIT_INTERNAL_ERROR. Note: this will trigger a
	     * 204/GLEXEC_EXIT_CHILD_PROCESS_RC_OVERLAP, but that is actually
	     * correct since both child and parent should exit with an internal
	     * error. */
	    exit(GLEXEC_EXIT_INTERNAL_ERROR); 
	}
    } else {
	/* Set epilogue timeout: ONLY IN PARENT! */
	glexec_alarm(ALARM_EPILOGTIMEOUT);
    }
    /* unblock signals in both parent and child */
    glexec_sigprocmask(&oldset);

    return global_ipc_epilogpid;
}

/*!
 * Waits for the child/payload process (in linger mode) to finish and correctly
 * handles, logs etc. the reason (normal exit, signal etc.). Timing statistics
 * are also logged.
 * \return exit code for child process, following (ba)sh convention: when due to
 * a signal: 128+<signal>, in case of unsignalled but abnormal program end, we
 * return 1. See http://tldp.org/LDP/abs/html/exitcodes.html for details.
 */ 
int glexec_waitchild(void)  {
    pid_t glexec_pgid=getpgrp(),wpid;
    int status,sig,werrno;
    int wait_continue=1;
    sigset_t oldset;

    /* Log at least to debug what we are doing */
    glexec_log(GLEXEC_LOG_DEBUG,"Waiting for payload process %d\n",
		global_ipc_childpid);
    
    /* Wait for pid to end. Also wait for stopped and continued processes to
     * give feedback. */
    do	{
	/* Flush signal bufferlog */
	glexec_flush_signallog();

	/* Block all signals until we have parsed the wait status, make sure to
	 * do as little as possible as we also block SEGV etc. */
	glexec_block_allsignals(&oldset);

	/* Reset alarm when it's not due to a sigterm etc. */
	if (global_ipc_active_alarm==ALARM_PAUSE)
	    glexec_alarm(ALARM_RESET);

	/* Do waitpid(), only here */
	wpid=waitpid(global_ipc_childpid,&status,WCONTINUED|WUNTRACED|WNOHANG);
	werrno=errno;
	if (wpid>0) { /* status changed, parse it */
	    /* Reason for state change */
	    if (WIFSTOPPED(status)) {
		/* Unblock signals */
		glexec_sigprocmask(&oldset);

		/* payload suspended */
		sig=WSTOPSIG(status);
		glexec_log(GLEXEC_LOG_NOTICE,
			"Payload %d has suspended due to signal %d.\n",
			(int)wpid,sig);
		fprintf(stderr, "[gLExec]: Payload %d suspended.\n",
			(int)wpid);
		/* Take controlling terminal if needed, warn if fails, it's
		 * unexpected */
		if ( global_ipc_stdinopts & PAYLOAD_TAKETTY) {
		    if (tcsetpgrp(STDIN_FILENO, glexec_pgid))   {
			global_ipc_stdinopts&=~(unsigned)PAYLOAD_TAKETTY;
			glexec_log(GLEXEC_LOG_NOTICE,
			    "Cannot grab control of tty from payload: %s\n",
			    strerror(errno));
		    } else  
			glexec_log(GLEXEC_LOG_DEBUG,
			    "Payload suspended, taking back tty.\n");
		}
#ifdef WIFCONTINUED
	    } else if (WIFCONTINUED(status))	{
		/* Unblock signals */
		glexec_sigprocmask(&oldset);

		/* payload continued */
		glexec_log(GLEXEC_LOG_NOTICE,
			"Payload %d has continued.\n",(int)wpid);
		fprintf(stderr, "[gLExec]: Payload %d continued.\n", (int)wpid);
		/* Return controlling terminal if needed, warn if fails, it's
		 * unexpected */
		if ( global_ipc_stdinopts & PAYLOAD_TAKETTY)  {
		    if (tcsetpgrp(STDIN_FILENO, wpid))  {
			global_ipc_stdinopts&=~(unsigned)PAYLOAD_TAKETTY;
			glexec_log(GLEXEC_LOG_NOTICE,
			    "Cannot give control of tty to payload: %s\n",
			    strerror(errno));
		    } else
			glexec_log(GLEXEC_LOG_DEBUG,
			    "Payload continued, giving back tty.\n");
		}
#endif
	    } else  {
		/* Not suspended or continued: exit. Don't reset signal block
		 * until AFTER we have set the return code of the payload. This
		 * guarantees that we do not interrupt between the waitpid and
		 * setting of the child return code. */ 
		glexec_continue_as_root();
		/* Update other fields */
		global_ipc_childstage=CHLD_HAS_ENDED;
		/* We're done now. */
		wait_continue=0;
		/* Reset alarm and clear pending or caught alarm */
		glexec_alarm(ALARM_RESET);
		global_ipc_caught_alarm=global_ipc_pending_alarm=ALARM_RESET;
		global_ipc_caught_signal=global_ipc_numsigsent=0;
		/* Parse status */
		if (WIFEXITED(status))   {
		    /* payload exited normally */
		    global_ipc_childrc=WEXITSTATUS(status);
		    /* Unblock signals before logging */
		    glexec_sigprocmask(&oldset);
		    /* Log reason */
		    glexec_log(global_ipc_childrc ? GLEXEC_LOG_ERROR
						  : GLEXEC_LOG_INFO,
			    "Child process %d exited with return code %d.\n",
			    global_ipc_childpid, global_ipc_childrc);
		} else if (WIFSIGNALED(status)) {
		    /* Signal ended program: use WTERMSIG() */
		    /* Bash convention:
		     * http://tldp.org/LDP/abs/html/exitcodes.html */
		    sig=WTERMSIG(status);
		    global_ipc_childrc=GLEXEC_EXECVE_SIGEXIT_BASE+sig;
		    /* Unblock signals before logging */
		    glexec_sigprocmask(&oldset);
		    /* Log reason */
		    glexec_log(GLEXEC_LOG_ERROR,
			    "Child process %d exited with return code %d "
			    "due to signal %d.\n",
			    global_ipc_childpid, global_ipc_childrc, sig);
		} else {
		    /* Unknown reason: set global_ipc_childrc to 1 */
		    global_ipc_childrc=1;
		    /* Unblock signals before logging */
		    glexec_sigprocmask(&oldset);
		    /* Log reason */
		    glexec_log(GLEXEC_LOG_ERROR,
			"Child process %d changed to unknown status. "
			"Setting return code to %d (status=%d).\n",
			global_ipc_childpid, global_ipc_childrc, status);
		}
	    }
	} else { /* wpid <0 or wpid== 0 */
	    /* Unblock signals */
	    glexec_sigprocmask(&oldset);

	    if (wpid<0)	{
		/* Warning: none of ECHILD, EINTR and EINVAL should occur */
		glexec_log(GLEXEC_LOG_WARNING,
			"Warning: waitpid() failed: %s\n",strerror(werrno));
	    } else { /* i.e. wpid==0 */
		/* Only act when the child is still running */
		if (global_ipc_childstage>0)	{
		    /* No wait change, child is running */
		    if (global_ipc_caught_signal)   {
			/* Signal has been caught: handle it */
			if (global_ipc_caught_signal!=SIGCHLD)
			    glexec_forward_to_payload(global_ipc_caught_signal);
			else
			    global_ipc_caught_signal=0;
		    } else  {
			/* NOTE: This is the only blocking statement in the
			 * loop, the SIGCHLD and SIGCONT handlers set a timer
			 * such that if they arrive just before the following
			 * pause(), the pause itself will eventually be
			 * interrupted */
			pause();
		    }
		}
	    }
	}
    } while (wait_continue);

    /* Look for remaining processes in pgid if we use them */
    glexec_cleanup_procesgroup(global_ipc_childpid);

    /* Take back controlling terminal when needed. */
    if ( global_ipc_stdinopts & PAYLOAD_TAKETTY)    {
	if (tcsetpgrp(STDIN_FILENO, glexec_pgid))
	    glexec_log(GLEXEC_LOG_NOTICE,
		"Cannot grab control of tty from payload: %s\n",
		strerror(errno));
	else
	    glexec_log(GLEXEC_LOG_DEBUG,"Payload finished, taking back tty.\n");
    }

    /* Dump accounting info */
    glexec_accounting(1,0);
    
    return global_ipc_childrc;
}

/*!
 * Waits for the epilogue process to finish and correctly handles, logs etc. the
 * reason (normal exit, signal etc.).
 * \return exit code for epilogue process, following (ba)sh convention: when due
 * to a signal: 128+<signal>, in case of unsignalled but abnormal program end,
 * we return 1. See http://tldp.org/LDP/abs/html/exitcodes.html for details.
 */ 
int glexec_waitepilog(void)  {
    pid_t wpid;
    int status,sig,werrno;
    int wait_continue=1;
    sigset_t oldset;

    /* Log at least to debug what we are doing */
    glexec_log(GLEXEC_LOG_DEBUG,"Waiting for epilogue process %d\n",
		global_ipc_epilogpid);

    /* Wait for pid to end. Also wait for stopped and continued processes to
     * give feedback. */
    do	{
	/* Flush signal bufferlog */
	glexec_flush_signallog();

	/* Block all signals until we have parsed the wait status, make sure to
	 * do as little as possible as we also block SEGV etc. */
	glexec_block_allsignals(&oldset);

	/* Reset alarm when it's not due to a sigterm or walltime */
	if (global_ipc_active_alarm==ALARM_PAUSE)
	    glexec_alarm(ALARM_RESET);

	/* Do waitpid(), only here */
	wpid=waitpid(global_ipc_epilogpid,&status,WCONTINUED|WUNTRACED|WNOHANG);
	werrno=errno;

	if (wpid>0) { /* status changed, parse it */
	    if (WIFSTOPPED(status)) {
		/* Unblock signals */
		glexec_sigprocmask(&oldset);

		/* epilogue suspended */
		sig=WSTOPSIG(status);
		glexec_log(GLEXEC_LOG_NOTICE,
			"Epilogue %d has suspended due to signal %d.\n",
			(int)wpid,sig);
#ifdef WIFCONTINUED
	    } else if (WIFCONTINUED(status))	{
		/* Unblock signals */
		glexec_sigprocmask(&oldset);

		/* epilogue continued */
		glexec_log(GLEXEC_LOG_NOTICE,
			"epilogue %d has continued.\n",(int)wpid);
#endif
	    } else  {
		/* Not suspended or continued: exit. Don't reset signal block
		 * until AFTER we have set the return code of the epilogue. This
		 * guarantees that we do not interrupt between the waitpid and
		 * setting of the epilog return code. */ 
		global_ipc_epilogstage=CHLD_HAS_ENDED;
		/* We're done now. */
		wait_continue=0;
		/* Reset alarm and clear pending or caught alarm */
		glexec_alarm(ALARM_RESET);
		global_ipc_caught_alarm=global_ipc_pending_alarm=ALARM_RESET;
		global_ipc_caught_signal=global_ipc_numsigsent=0;
		/* Parse status */
		if (WIFEXITED(status))   {
		    /* epilogue exited normally */
		    global_ipc_epilogrc=WEXITSTATUS(status);
		    /* Unblock signals before logging */
		    glexec_sigprocmask(&oldset);
		    /* Log reason */
		    glexec_log(global_ipc_epilogrc ? GLEXEC_LOG_ERROR
						  : GLEXEC_LOG_INFO,
			    "Epilogue process %d exited with return code %d.\n",
			    global_ipc_epilogpid, global_ipc_epilogrc);
		} else if (WIFSIGNALED(status)) {
		    /* Signal ended program: use WTERMSIG() */
		    /* Bash convention:
		     * http://tldp.org/LDP/abs/html/exitcodes.html */
		    sig=WTERMSIG(status);
		    global_ipc_epilogrc=GLEXEC_EXECVE_SIGEXIT_BASE+sig;
		    /* Unblock signals before logging */
		    glexec_sigprocmask(&oldset);
		    /* Log reason */
		    glexec_log(GLEXEC_LOG_ERROR,
			    "Epilogue process %d exited with return code %d "
			    "due to signal %d.\n",
			    global_ipc_epilogpid, global_ipc_epilogrc, sig);
		} else {
		    /* Unknown reason: set global_ipc_epilogrc to 1 */
		    global_ipc_epilogrc=1;
		    /* Unblock signals before logging */
		    glexec_sigprocmask(&oldset);
		    /* Log reason */
		    glexec_log(GLEXEC_LOG_ERROR,
			"Epilogue process %d changed to unknown status. "
			"Setting return code to %d (status=%d).\n",
			global_ipc_epilogpid, global_ipc_epilogrc, status);
		}
	    }
	} else { /* wpid <0 or wpid== 0 */
	    /* Unblock signals */
	    glexec_sigprocmask(&oldset);

	    if (wpid<0)	{
		/* Warning: none of ECHILD, EINTR and EINVAL should occur */
		glexec_log(GLEXEC_LOG_WARNING,
			"Warning: waitpid() failed: %s\n",strerror(werrno));
	    } else { /* i.e. wpid==0 */
		/* Only act when the epilog is still running */
		if (global_ipc_epilogstage>0)	{
		    /* No waitpid() change, epilog is running */
		    if (global_ipc_caught_signal)   {
			/* Signal has been caught: handle it */
			if (global_ipc_caught_signal!=SIGCHLD)
			    glexec_forward_to_epilog(global_ipc_caught_signal);
			else
			    global_ipc_caught_signal=0;
		    } else  {
			/* NOTE: This is the only blocking statement in the
			 * loop, the SIGCHLD and SIGCONT handlers set a timer
			 * such that if they arrive just before the following
			 * pause(), the pause itself will eventually be
			 * interrupted */
			pause();
		    }
		}
	    }
	}
    } while (wait_continue);

    /* Look for remaining processes in pgid if we use them */
    glexec_cleanup_procesgroup(global_ipc_epilogpid);

    return global_ipc_epilogrc;
}


/************************************************************************/
/* INTERNAL FUNCTIONS                                                   */
/************************************************************************/

/*!
 * Sends a signal to child pid, only when it's a child process
 * Temporarily blocks all signals (except KILL and STOP) to prevent other signal
 * handlers from calling us a second time.
 * \param pid process ID or process group (when negative) to signal
 * \param signo signal to send
 * \param stage either global_ipc_childstage or global_ipc_epilogstage, used to
 *        test whether that childprocess matches abs(pid) and updated when that
 *        pid isn't there any longer.
 * \return 0 signal delivered
 * \return -1 signal failed
 * \return -2 invalid child pid, cannot kill
 */
static int glexec_send_signal(pid_t pid, int signo, pid_t *stage)	{
    uid_t euid;
    sigset_t oldset;
    int kerrno,rerrno,rc,rc_seteuid1=0,rc_seteuid2=0;
    pid_t procpid=pid<0 ? -pid : pid;

     /* Block all signals, including critical ones */
    glexec_block_allsignals(&oldset);

    /* Active child: forward signal, test stage is actually equal to abs(pid) */
    if (procpid==*stage)  {
	/* do geteuid here instead of inside following 'if' block to prevent
	 * compiler warning */
	euid=geteuid();
	/* Switch to effective root if possible, don't log anything now */
	if (global_ipc_canswitch && ((rc_seteuid1=seteuid(0))!=0) )
	    global_ipc_canswitch=0;
	rc=kill(pid,signo);
	/* save errno of kill, and update rc and stage if it's due to
	 * non-existing child */
	if ( (kerrno=errno) == ESRCH)	{
	    /* Update stage: we don't send before it started, so it has finished
	     * */
	    *stage=CHLD_HAS_ENDED;
	    /* Better log with signal_log, as we are blocking all signals */
	    glexec_signal_log(GLEXEC_LOG_NOTICE,"Process %s %d does not exist.\n",
		    pid<0 ? "group" : "id", procpid);
	    rc=-2;
	}
	if (global_ipc_canswitch && ((rc_seteuid2=seteuid(euid))!=0) )
	    global_ipc_canswitch=0;
	/* Log if userswitching went wrong */
	if (rc_seteuid1 || rc_seteuid2)
	    /* Better log with signal_log, as we are blocking all signals */
	    glexec_signal_log(GLEXEC_LOG_ERROR,
		"This should not happen: could not userswitch, "
		"now lingering as uid/euid=%d/%d\n",getuid(),geteuid());
	rerrno=kerrno;
    } else {
	rerrno=0;
	rc=-2;
    }

    /* Reset procmask */
    glexec_sigprocmask(&oldset);
    /* Reset errno to correct one */
    errno=rerrno;
    return rc;
}

/*!
 * Sends final signal to child(group) from within terminating signal handler.
 * The only potentially signal unsafe call is seteuid()
 * \param pid process ID or process group (when negative) to signal
 * \param signo signal to send
 * \param stage either global_ipc_childstage or global_ipc_epilogstage, used to
 *        test whether that childprocess matches abs(pid) and updated when that
 *        pid isn't there any longer.
 * \return 0 signal delivered
 * \return -1 signal failed
 * \return -2 invalid child pid, cannot kill
 */
static int glexec_emergency_send_signal(pid_t pid, int signo, pid_t *stage) {
    uid_t euid;
    int rc,rc_seteuid=0;
    pid_t procpid=pid<0 ? -pid : pid;

    /* Active child, then send signal. Note that we only call wait*() in the
     * wait{child,epilog} functions inside a block-all. Hence when either stage
     * says they are running, they can be at most zombie. */
    if (procpid==*stage)  {
	/* Switch to effective root if possible */
	euid=geteuid();
	if (global_ipc_canswitch) 
	    rc_seteuid=seteuid(0);
	rc=kill(pid,signo);
	/* save errno of kill, and update rc and stage if it's due to
	 * non-existing child */
	if ( errno==ESRCH)	{
	    /* Update stage: we don't send before it started, so it has
	     * finished. Don't log it, as it's not important enough */
	    *stage=CHLD_HAS_ENDED;
	    rc=-2;
	}
	/* if previous seteuid succeeded, try to return */
	if (rc_seteuid==0)
	    rc_seteuid=seteuid(euid);
	/* Log if either userswitch failed */
	if (rc_seteuid)
	    glexec_signal_log(GLEXEC_LOG_ERROR,
		"Error: This should not happen: userswitching failed.\n");
	/* Log if kill failed, only if 'real' error */
	if (rc==-1)
	    glexec_signal_log(GLEXEC_LOG_ERROR,
		"Error: could not send signal %d to %d.\n",signo,pid);
    } else
	rc=-2;

    return rc;
}

/*!
 * Changes effective and real identity (except secondary groups) to root to
 * prevent receiving of signals. This should be called either when receiving a
 * term-type signal, while there is a payload or when payload has finished.
 */
static void glexec_continue_as_root(void)	{
    /* Don't do double, first check real uid, as that makes us already immune
     * for further signals */
    if (getuid()==0 && geteuid()==0 && getgid()==0)
	return;

    if (global_ipc_canswitch)	{
	if (seteuid(0) || setgid(0) || setreuid(0,0))  {
	    /* Update the global_ipc_canswitch flag, aparently we lost it's
	     * capability */
	    global_ipc_canswitch=0;
	    glexec_signal_log(GLEXEC_LOG_WARNING,
		"Warning: Failed to setreuid(0,0) or setgid(0) "
		"will continue to linger as uid/euid=%d/%d: %s\n",
		getuid(),geteuid(),strerror(errno));
	} else 
	    glexec_signal_log(GLEXEC_LOG_DEBUG,
		"gLExec now lingers as uid/euid=0/0\n");
    }
}

/*!
 * Looks for processes in process group pid (when using process groups) and
 * logs if there are processes remaining.
 * \param pid
 */
static void glexec_cleanup_procesgroup(pid_t pid)    {
    /* If we're using process groups, we can still send a signal to the process
     * group in case there are some daemonized processes left. */
    if (global_ipc_usesetpgid)	{
	/* If pid is a pgid, test if there are grandchild processes in the same
	 * pgid, to be able to log the information */
	if (kill(-pid,0)==0)	{
	    glexec_log(GLEXEC_LOG_NOTICE,
		"Child %d finished, but there are processes in "
		"group %d, there might be daemonized grandchildren\n",
		pid,pid);
	}
	else if (errno!=ESRCH)
	    glexec_log(GLEXEC_LOG_WARNING,
		"Could not get status for processgroup %d while looking for "
		"daemonized grandchildren: %s.\n",
		pid,strerror(errno));
    }
}

/*!
 * Handles forwarding of signals to payload process. The first line signal
 * handling is done in the signal handlers (_term_action, _alarm_action,
 * _other_action) which set a flag and log using glexec_signal_log().
 * \param signo signal received
 */
static void glexec_forward_to_payload(int signo)	{
    int sendsignal;
    ALARM_TYPE_T caught_alarm;
    sigset_t oldset;

    /* Block all signals */
    glexec_block_signals(&oldset);

    /* reset caught signal */
    global_ipc_caught_signal=0;

    /* handle different signals */
    switch(signo)   {
	case SIGALRM:
	    caught_alarm=global_ipc_caught_alarm;
	    global_ipc_caught_alarm=ALARM_RESET;
	    switch (caught_alarm)  {
		case ALARM_TERMDELAY:
		    /* TERM has been sent: send KILL */
		    glexec_log(GLEXEC_LOG_WARNING,
			"Grace time of %d sec after SIGTERM expired, "
			"sending SIGKILL to payload %s %d...\n",
			global_ipc_termdelay,
			global_ipc_usesetpgid ? "group" : "id",
			global_ipc_childpid);
		    sendsignal=SIGKILL;
		    glexec_alarm(ALARM_KILLDELAY);
		    break;
		case ALARM_KILLDELAY:
		    /* KILL has been sent, but didn't lead to child ending after
		     * global_ipc_killdelay: something is horribly broken: exit
		     * */
		    glexec_log(GLEXEC_LOG_ERROR,
			"Error: payload %d did not die from a SIGKILL )-:\n",
			global_ipc_childpid);
		    glexec_accounting(1,0);
		    glexec_exit(GLEXEC_PREMATURE_EXIT,global_ipc_childrc);
		case ALARM_RESET: /* This should not be possible */
		case ALARM_EPILOGTIMEOUT: /* This should not be possible */
		case ALARM_PAUSE:
		default: /* probably expired pause-alarm */
		    glexec_log(GLEXEC_LOG_INFO,
			"Ignoring ALARM signal.\n");
		    return;
	    }
	    break;
	case SIGINT:
	case SIGTERM:
	case SIGXCPU:
	case SIGXFSZ:
	    if (signo==SIGINT)
		fprintf(stderr,
			"[gLExec]: Caught SIGINT, sending SIGTERM to payload.\n");
	    glexec_log(GLEXEC_LOG_WARNING,
		"Lingering gLExec caught signal %d, "
		"sending SIGTERM to payload process %s %d...\n",
		signo, global_ipc_usesetpgid ? "group" : "id",
		global_ipc_childpid);
	    sendsignal=SIGTERM;
	    glexec_alarm(ALARM_TERMDELAY);
	    break;
	default:
	    /* Log CTRL-z on cmdline such that glexec user knows what's
	     * happening */
	    if (signo==SIGTSTP)
		fprintf(stderr,"[gLExec]: Forwarding caught SIGTSTP to payload.\n");
	    glexec_log(GLEXEC_LOG_NOTICE,
		"Forwarding signal %d to payload process %s %d...\n",
		signo, global_ipc_usesetpgid ? "group" : "id",
		global_ipc_childpid);
	    sendsignal=signo;
	    break;
    }

    /* Reset signal mask */
    glexec_sigprocmask(&oldset);
    
    /* Forward signal to payload, ignore when it has already finished */
    if (glexec_send_signal(
	    global_ipc_usesetpgid ? -global_ipc_childpid : global_ipc_childpid,
	    sendsignal, &global_ipc_childstage)==-1)    {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Could not send signal %d to payload %s %d.\n",
	    sendsignal,
	    global_ipc_usesetpgid ? "group" : "id",
	    global_ipc_childpid);
	glexec_accounting(1,0);
	glexec_exit(GLEXEC_SIGNAL_ERROR,global_ipc_childrc);
    }
}

/*!
 * Handles forwarding of signals to epilogue process. The first line signal
 * handling is done in the signal handlers (_term_action, _alarm_action,
 * _other_action) which set a flag and log using glexec_signal_log().
 * \param signo signal received
 */
static void glexec_forward_to_epilog(int signo)	{
    int sendsignal;
    ALARM_TYPE_T caught_alarm;
    sigset_t oldset;

    /* Block all signals */
    glexec_block_signals(&oldset);

    /* reset caught signal */
    global_ipc_caught_signal=0;

    /* handle different signals */
    switch(signo)   {
	case SIGALRM:
	    caught_alarm=global_ipc_caught_alarm;
	    global_ipc_caught_alarm=ALARM_RESET;
	    switch (caught_alarm) {
		case ALARM_EPILOGTIMEOUT:
		    /* total runtime has expired: send TERM */
		    glexec_log(GLEXEC_LOG_WARNING,
			"Maxmimum runtime of %d sec expired for epilogue, "
			"sending SIGTERM to its process %s %d...\n",
			global_ipc_epilogtimeout,
			global_ipc_usesetpgid ? "group" : "id",
			global_ipc_epilogpid);
		    sendsignal=SIGTERM;
		    glexec_alarm(ALARM_TERMDELAY);
		    break;
		case ALARM_TERMDELAY:
		    /* TERM has been sent: send KILL */
		    glexec_log(GLEXEC_LOG_WARNING,
			"Grace time of %d sec after SIGTERM expired for epilogue, "
			"sending SIGKILL to its process %s %d...\n",
			global_ipc_termdelay,
			global_ipc_usesetpgid ? "group" : "id",
			global_ipc_epilogpid);
		    sendsignal=SIGKILL;
		    glexec_alarm(ALARM_KILLDELAY);
		    break;
		case ALARM_KILLDELAY:
		    /* KILL has been sent, but didn't lead to child ending after
		     * global_ipc_killdelay: something is horribly broken: exit
		     * */
		    glexec_log(GLEXEC_LOG_ERROR,
			"Error: epilogue %d did not die from a SIGKILL, "
			"payload finished with rc=%d.\n",
			global_ipc_epilogpid,global_ipc_childrc);
		    glexec_exit(GLEXEC_PREMATURE_EXIT,global_ipc_childrc);
		case ALARM_RESET: /* This should not be possible */
		case ALARM_PAUSE: /* This should not be possible */
		default: /* unexpected */
		    glexec_log(GLEXEC_LOG_WARNING,
			"Warning: ignoring unexpected ALARM signal.\n");
		    return;
	    }
	    break;
	case SIGINT:
	case SIGTERM:
	case SIGXCPU:
	case SIGXFSZ:
	    if (signo==SIGINT)
		fprintf(stderr,
			"[gLExec]: Caught SIGINT, sending SIGTERM to epilogue.\n");
	    glexec_log(GLEXEC_LOG_WARNING,
		"Warning: lingering gLExec caught signal %d, "
		"sending SIGTERM to epilogue process %s %d...\n",
		signo, global_ipc_usesetpgid ? "group" : "id",
		global_ipc_epilogpid);
	    sendsignal=SIGTERM;
	    glexec_alarm(ALARM_TERMDELAY);
	    break;
	default:
	    /* Log CTRL-z on cmdline such that glexec user knows what's
	     * happening */
	    if (signo==SIGTSTP)
		fprintf(stderr,"[gLExec]: Forwarding caught SIGTSTP to epilogue.\n");
	    glexec_log(GLEXEC_LOG_NOTICE,
		"Forwarding signal %d to epilogue process %s %d...\n",
		signo, global_ipc_usesetpgid ? "group" : "id",
		global_ipc_epilogpid);
	    sendsignal=signo;
	    break;
    }

    /* Reset signal mask */
    glexec_sigprocmask(&oldset);

    /* Forward signal to epilogue, ignore when it has already finished */
    if (glexec_send_signal(
	    global_ipc_usesetpgid ? -global_ipc_epilogpid : global_ipc_epilogpid,
	    sendsignal, &global_ipc_epilogstage)==-1)    {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Could not send signal %d to epilogue %s %d, "
	    "payload finished with rc=%d.\n",
	    sendsignal,
	    global_ipc_usesetpgid ? "group" : "id",
	    global_ipc_epilogpid, global_ipc_childrc);
	glexec_accounting(1,0);
	glexec_exit(GLEXEC_PREMATURE_EXIT,global_ipc_childrc);
    }
}

/*!
 * Starts (started==0) and stops+logs (started==1) accounting info. When called
 * a second time with started==1, nothing happens.
 * \param started 0 for starting the timing records, 1 for stopping and
 * printing.
 */
static void glexec_accounting(int started, int use_signalversion)	{
    static clock_t real_time_start;
    static struct tms tms_start;
    struct tms tms_stop;
    clock_t clock_ticks;
    clock_t real_time_stop;
    static int logged=0;
    int r,u,s;

    if (started==0) {
	real_time_start=times(&tms_start);
    } else {
	if (logged) return;

	/* Store endtime */
	real_time_stop=times(&tms_stop);

	/* Print out timing statistics */
	if ((clock_ticks = sysconf(_SC_CLK_TCK)) <= 0 ) {
	    if (use_signalversion)
		glexec_signal_log(GLEXEC_LOG_WARNING,
		    "Clockticks/second could not be determined, "
		    "using clockticks instead of seconds.\n");
	    else
		glexec_log(GLEXEC_LOG_WARNING,
		    "Clockticks/second could not be determined, "
		    "using clockticks instead of seconds.\n");
	    clock_ticks = 1;
	}
	if (use_signalversion)	{
	    /* Better not use doubles in vsnprintf in signal handler */
	    r=(int)( ((double)real_time_stop-(double)real_time_start)/(double)clock_ticks );
	    u=(int)( ((double)tms_stop.tms_cutime-(double)tms_start.tms_cutime)/(double)clock_ticks );
	    s=(int)( ((double)tms_stop.tms_cstime-(double)tms_start.tms_cstime)/(double)clock_ticks );
	    glexec_signal_log(GLEXEC_LOG_NOTICE,
		    "times: real/user/sys (%d/%d/%d)\n",r,u,s);
	} else
	    glexec_log(GLEXEC_LOG_NOTICE,
		    "times: real/user/sys (%f/%f/%f)\n",
		    ((double)real_time_stop - (double)real_time_start)/(double)clock_ticks,
		    ((double)tms_stop.tms_cutime-(double)tms_start.tms_cutime)/(double)clock_ticks,
		    ((double)tms_stop.tms_cstime-(double)tms_start.tms_cstime)/(double)clock_ticks);
   
	/* Prevent double logging */
	logged=1;
    }
}

/*!
 * Substitute for fprintf(stderr,string) that can be used within signal handlers
 * \param text to be printed on stderr STDERR_FILENO
 * \return return code of write, 0 when text is NULL
 */
static ssize_t glexec_write_stderr(const char *text)	{
    ssize_t rc;
    size_t len;

    if (text==NULL) return 0;

    for (len=0; text[len]!='\0'; len++);
    
    rc=write(STDERR_FILENO, (const void *)text, len);

    return (int)rc;
}

/*!
 * Wrapper around alarm() function, needed to provide SIGALRM handler with
 * information about the cause of the alarm. Note that the handler might be
 * handling a SIGALRM while there is already a new alarm() set.
 * It updates the global_ipc_active_alarm with the newly set alarm(). When there
 * is a pending alarm and global_ipc_pending_alarm is not yet set, that flag is
 * also updated. Finally global_ipc_numsigsent is updated when we set any of the
 * timeout alarms.
 * \param type 
 * \return number of seconds left for previous alarm, 0 when unknown type.
 */
static unsigned glexec_alarm(ALARM_TYPE_T type)  {
    unsigned ret;
    sigset_t oldset,pending;

    /* Block all signals, likelihood of a SEGV or the like in minute... */
    glexec_block_allsignals(&oldset);

    /* If we have a pending SIGALRM, store the alarm type */
    sigpending(&pending);
    if (sigismember(&pending, SIGALRM) && global_ipc_pending_alarm==ALARM_RESET)
	global_ipc_pending_alarm=global_ipc_active_alarm;

    /* Set time based on type */
    switch (type)   {
	case ALARM_RESET:
	    ret=alarm(0);
	    break;
	case ALARM_PAUSE:
	    ret=alarm(1);
	    break;
	case ALARM_EPILOGTIMEOUT:
	    global_ipc_numsigsent=0;
	    ret=alarm(global_ipc_epilogtimeout);
	    break;
	case ALARM_TERMDELAY:
	    global_ipc_numsigsent=1;
	    ret=alarm(global_ipc_termdelay);
	    break;
	case ALARM_KILLDELAY:
	    global_ipc_numsigsent=2;
	    ret=alarm(global_ipc_killdelay);
	    break;
	default:
	    glexec_sigprocmask(&oldset);
	    return 0;
	    break;
    }

    /* Set alarm and new alarm type */
    global_ipc_active_alarm=type;

    /* Unblock signals */
    glexec_sigprocmask(&oldset);

    return ret;
}

/************************************************************************/
/* SIGNAL HANDLERS                                                      */
/************************************************************************/

/*!
 * Handler for ignore signals: log and ignore, e.g. SIGPIPE
 */
static void glexec_ignore_action(int signo, siginfo_t *info, void *context)  {
    glexec_signal_log(GLEXEC_LOG_NOTICE,
	    "Ignoring signal %d (si_code=%d)\n",
	    signo,info->si_code);
}

/*!
 * Handler for SIGCHLD, updates global_ipc_{child,epilog}stage and sets an alarm
 * when a running payload has ended.
 */
static void glexec_child_action(int signo, siginfo_t *info, void *context)  {
    /* Now parse code */ 
    switch(info->si_code)   {
	/* Terminated states: */
	case CLD_EXITED:
	case CLD_KILLED:
	case CLD_DUMPED:
	    if (info->si_pid==global_ipc_childstage)	{
		/* save childpid (normally unneeded, SIGCHLD is blocked from
		 * fork() until it is already set) */
		global_ipc_childpid=global_ipc_childstage;
		global_ipc_childstage=CHLD_HAS_ENDED;
		/* Store caught signal: child has finished so nothing to forward or
		 * kill */
		global_ipc_caught_signal=signo;
		/* Even if we have a running alarm from a numsigsent it's now
		 * irrelevant, as the child has already finished, so we can
		 * safely set a new one */
		glexec_alarm(ALARM_PAUSE);
		glexec_signal_log(GLEXEC_LOG_DEBUG,
			"Will handle SIGCHLD due to end of payload %d\n",
			global_ipc_childpid);
	    } else if (info->si_pid==global_ipc_epilogstage)	{
		/* save childpid (normally unneeded, SIGCHLD is blocked from
		 * fork() until it is already set) */
		global_ipc_epilogpid=global_ipc_epilogstage;
		global_ipc_epilogstage=CHLD_HAS_ENDED;
		/* Store caught signal: child has finished so nothing to forward or
		 * kill */
		global_ipc_caught_signal=signo;
		/* Epilogue has finished, so even walltime is irrelevant */
		glexec_alarm(ALARM_PAUSE);
		glexec_signal_log(GLEXEC_LOG_DEBUG,
			"Will handle SIGCHLD due to end of epilogue %d\n",
			global_ipc_epilogpid);
	    } else if (info->si_pid!=global_ipc_childpid &&
		       info->si_pid!=global_ipc_epilogpid) 
		glexec_signal_log(GLEXEC_LOG_NOTICE,
		    "Ignoring SIGCHLD from unknown finished child %d.\n",
		    info->si_pid);
	    break;
	/* Stopped/continued states: don't act here */
	case CLD_TRAPPED:
	case CLD_STOPPED:
	case CLD_CONTINUED:
	    /* Store caught signal: only if we haven't yet caught another
	     * signal. */
	    if (global_ipc_caught_signal==0)	{
		global_ipc_caught_signal=signo;
		/* If we're in running payload stage and have no other running
		 * alarm, set one here, to force interrupt on pause() */
		if (info->si_pid==global_ipc_childstage &&
		    global_ipc_active_alarm==ALARM_RESET)
		    glexec_alarm(ALARM_PAUSE);
		/* Log on debug */
		glexec_signal_log(GLEXEC_LOG_DEBUG,
		    "Will handle SIGCHLD for change in pid=%d/uid=%u\n",
		    info->si_pid,info->si_uid);
	    }
	    break;
	default:
	    glexec_signal_log(GLEXEC_LOG_NOTICE,
		"Ignoring SIGCHLD from pid=%d/uid=%u\n",
		info->si_pid,info->si_uid);
	    break;
    }
}

/*!
 * Handler for SIGINT, SIGTERM, SIGXCPU, SIGXFSZ: send children a TERM and set
 * alarm. Behaviour depends on stage of payload and epilogue.
 */
static void glexec_term_action(int signo, siginfo_t *info, void *context)   {
    /* Action depends on the stage were in */
    switch (global_ipc_childstage)	{
	case CHLD_HAS_NOT_YET_RUN:
	    /* pre-child: log and exit */
	    /* Only accept process signals from caller or root */
	    if (info->si_code<=0 &&
		    info->si_uid!=global_ipc_calleruid && info->si_uid!=0)   {
		glexec_signal_log(GLEXEC_LOG_NOTICE,
		    "Ignoring signal %d from pid=%d/uid=%u\n",
		    signo,info->si_pid,info->si_uid);
		return;
	    }
	    /* Origin is ok: log and exit */
	    glexec_signal_log(GLEXEC_LOG_NOTICE,
		"Caught signal %d, exiting...\n", signo);
	    /* This will exit with a 201, since we can blame the user... Note
	     * that we shouldn't call syslog() as we might be inside e.g. lcmaps */
	    if (signo==SIGINT || signo==SIGTERM)
		glexec_signal_exit(GLEXEC_SUCCESS,global_ipc_childrc,signo);
	    else
		glexec_signal_exit(GLEXEC_UNCLASIFIED,global_ipc_childrc,signo);
	case CHLD_FORK_FAILED:
	    /* fork failed: we received it during the exit: this should not
	     * happen because glexec_exit blocks SIGTERM etc. */
	    glexec_signal_log(GLEXEC_LOG_NOTICE,
		"Caught signal %d but fork failed so we are already exiting.\n",
		signo);
	    return;
	case CHLD_PROC_IS_CHILD:
	    /* child itself before execve: log and exit */
	    glexec_signal_log(GLEXEC_LOG_NOTICE,
		"Payload: caught signal %d, exiting...\n", signo);
	    glexec_emergency_flush_signallog();
	    exit(GLEXEC_EXECVE_SIGEXIT_BASE+signo);
	case CHLD_HAS_ENDED:
	    /* post-child: action depends on epilogue: first handle
	     * non-authorized. On Linux a CTRL-c sends a non-process code
	     * (code>0) and we should check for global_ipc_calleruid, on MacOS
	     * it looks like a signal from a different process, si_code>=0. */
	    if ( (info->si_code<=0 && info->si_uid!=0) ||
		 (info->si_code>0  && global_ipc_calleruid !=0
				   && signo==SIGINT) ) {
		/* warn about CTRL-C also on stderr */
		if (signo==SIGINT)
		    glexec_write_stderr(
			"[gLExec]: Ignoring caught SIGINT from non-root since "
			"payload has finished.\n");
		/* log message depends on type of signal */
		if (info->si_code<=0)
		    glexec_signal_log(GLEXEC_LOG_NOTICE,
			"Ignoring signal %d from pid=%d/uid=%u, "
			"payload has already finished.\n",
			signo,info->si_pid,info->si_uid);
		else
		    glexec_signal_log(GLEXEC_LOG_NOTICE,
			"Ignoring signal %d (si_code=%d), "
			"payload has already finished.\n",
			signo,info->si_code);
		return;
	    }
	    
	    /* Remainder is authorized signal: either kernel or root
	     * action depends on stage of epilogue. */
	    switch (global_ipc_epilogstage) {
		case CHLD_WONT_RUN: /* epilogue won't run */
		case CHLD_HAS_ENDED: /* epilogue finished */
		case CHLD_FORK_FAILED: /* epilogue fork failed */
		    /* Final shutdown, this should not take long so don't
		     * interrupt */
		    glexec_signal_log(GLEXEC_LOG_NOTICE,
			"Ignoring signal %d during gLExec shutdown.\n",signo);
		    return;
		case CHLD_HAS_NOT_YET_RUN:
		    /* epilogue hasn't run yet: accounting might not yet have
		     * been printed */ 
		    glexec_signal_log(GLEXEC_LOG_WARNING,
			    "Warning: Caught signal %d before epilogue ran!, "
			    "payload finished with childrc=%d.\n",
			    signo,global_ipc_childrc);
		    glexec_accounting(1,1);
		    /* It's unsafe to call syslog here, so we probably lose log
		     * messages here */
		    glexec_signal_exit(GLEXEC_PREMATURE_EXIT,
			               global_ipc_childrc,signo);
		case CHLD_PROC_IS_CHILD:
		    /* epilogue itself, before execve() */
		    glexec_signal_log(GLEXEC_LOG_WARNING,
			    "Epilogue: Caught signal %d before execve()'ed\n",
			    signo);
		    glexec_emergency_flush_signallog();
		    exit(GLEXEC_EXECVE_SIGEXIT_BASE+signo);
		default:
		    /* active epilogue is running: need to forward, but only
		     * forward something when we haven't yet sent a SIGTERM */
		    if (global_ipc_numsigsent)  {
			/* Log the warning here */
			glexec_signal_log(GLEXEC_LOG_NOTICE,
			    "Ignoring signal %d since epilogue group %d has "
			    "already been sent SIGTERM...\n",
			    signo,global_ipc_epilogpid);
		    } else {
			/* Origin is ok: log here only on DEBUG and set flag */
			glexec_signal_log(GLEXEC_LOG_DEBUG,
			    "Caught signal %d, will send SIGTERM to epilogue.\n",
			    signo);
			/* set signal flag */
			global_ipc_caught_signal=signo;
			/* Since we are going to send a SIGTERM we no longer
			 * need the walltime or other timers and can set one to
			 * interrupt pause() where needed */
			glexec_alarm(ALARM_PAUSE);
		    }
		    return;
	    }
	    break;
	default:    /* global_ipc_childstage>0 i.e. there is a valid child */
	    /* Only do something when we haven't yet sent a TERM */
	    if (global_ipc_numsigsent)   {
		/* Log the warning here */
		glexec_signal_log(GLEXEC_LOG_NOTICE,
			"Caught signal %d but payload group %d has "
			"already been sent SIGTERM...\n",
			signo,global_ipc_childpid);
	    } else {
		/* become root to be immune for SIGKILL and SIGSTOP, this is not
		 * entirely signal safe, but we need to do it before doing
		 * anything else (including leaving the sighandler) */
		glexec_continue_as_root();

		/* Origin is ok: log here only on DEBUG and set flag */
		glexec_signal_log(GLEXEC_LOG_DEBUG,
		    "Caught signal %d, will send SIGTERM to payload.\n",
		    signo);

		/* set signal flag */
		global_ipc_caught_signal=signo;
		/* we have a child: set alarm to prevent missing it just before
		 * pause(), forward will set a new alarm to TERMDELAY */
		glexec_alarm(ALARM_PAUSE);
	    }
	    return;
    }
}

/*!
 * Handler for SIGALRM:
 * - glexec_term_action sets alarm for grace period after which to send a
 *   SIGKILL (global_ipc_active_alarm==ALARM_TERMDELAY)
 * - glexec_alarm_action sets alarm when a SIGKILL has been sent for a final
 *   graceperiod, after which it ends without further sending signals
 *   (global_ipc_active_alarm==ALARM_KILLDELAY)
 * - startup of epilogue triggers a alarm after which it will be terminated
 *   (global_ipc_active_alarm==ALARM_EPILOGTIMEOUT, only for
 *   global_ipc_epilogstage>0 or ==-3)
 */
static void glexec_alarm_action(int signo, siginfo_t *info, void *context)  {
    /* reason is either a pending one: new alarm was set during blocked signals
     * which had a pending SIGALRM, otherwise, it's from the current alarm */
    unsigned int reason=global_ipc_pending_alarm ? global_ipc_pending_alarm
				                 : global_ipc_active_alarm;

    /* Always reset pending_alarm, as we are now processing it. Don't reset
     * alarm() as there might already be a new one */
    global_ipc_pending_alarm=ALARM_RESET;

    /* Check the origin: should be from a timer (i.e. kernel) */
    if (info->si_code <= 0) {
	glexec_signal_log(GLEXEC_LOG_NOTICE,
		"Ignoring ALARM signal (%d) since it is not sent by kernel "
		"(si_code=%d).\n",
		signo,info->si_code);
	return;
    }

    /* Act depending on reason */
    switch(reason)  {
	case ALARM_PAUSE:   /* Can ignore: was needed to interrupt pause() */
	    glexec_signal_log(GLEXEC_LOG_DEBUG,
		    "Caught ALARM meant to interrupt pause(), "
		    "can further ignore it.\n");
	    return;
	    break;
	case ALARM_TERMDELAY:
	case ALARM_KILLDELAY:
	case ALARM_EPILOGTIMEOUT:
	    /* maximum time expired */
	    /* running payload or epilogue? */
	    if ( global_ipc_childstage>0 || global_ipc_epilogstage>0)	{
		glexec_signal_log(GLEXEC_LOG_DEBUG,
			"Will handle SIGALRM from expired timeout.\n");
		/* since stage is running, we haven't yet handled a SIGCHLD */
		global_ipc_caught_signal=signo;
		global_ipc_caught_alarm=reason;
		/* We can set a new pause alarm here, as the forward function
		 * will set a new one based on number of signals sent so far */
		glexec_alarm(ALARM_PAUSE);
		return;
	    }
	    /* Old alarm for either payload or epilogue. Whoever set stage to
	     * ended either has already set either a pause-alarm or prevents
	     * pause() altogether */
	    if (global_ipc_childstage==CHLD_HAS_ENDED)	{
		glexec_signal_log(GLEXEC_LOG_NOTICE,
			"Ignoring ALARM signal (%d) (si_code=%d): "
			"payload has already exited.\n",
			signo,info->si_code);
		return;
	    }
	    /* Epilogue already finished, but alarm still running */
	    if (global_ipc_epilogstage==CHLD_HAS_ENDED)	{
		glexec_signal_log(GLEXEC_LOG_NOTICE,
			"Ignoring ALARM signal (%d) (si_code=%d): "
			"epilogue has already exited.\n",
			signo,info->si_code);
		return;
	    }
	default:
	    /* Any other reason is unexpected and should not happen */
	    break;
    }
    
    /* Any other reason is unexpected and should not happen */
    glexec_signal_log(GLEXEC_LOG_WARNING,
	    "Warning: ignoring unexpected ALARM signal (%d) (si_code=%d).\n",
	    signo,info->si_code);
    
    return;
}

/*!
 * Handler for other signals: forward to child (not children!). For SIGCONT also
 * handle special case that there is a ALARM_PAUSE active, which we might have
 * missed if the SIGCONT woke us up.
 */
static void glexec_other_action(int signo, siginfo_t *info, void *context)  {
    int fd;

    /* For SIGCONT: it might be that we woke up just before pause() and that any
     * already set timer has just expired. For termdelay, killdelay and
     * epilogtimeout, a new one will be set in the SIGALRM that will run in that
     * case, but for an expired ALARM_PAUSE that's not the case. */
    if (signo==SIGCONT && global_ipc_active_alarm==ALARM_PAUSE)	{
	glexec_alarm(ALARM_PAUSE);
	glexec_signal_log(GLEXEC_LOG_DEBUG,
		"Caught SIGCONT while ALARM_PAUSE was active, "
		"restarting alarm as might have missed it.\n");
    }

    /* Remainder is primarily for forwarding */

    /* Only forward the signal to payload when there is one and it hasn't been
     * send a TERM yet. Also forward when it's not from a process, it might be
     * CTRL-Z for example. */
    if (global_ipc_childstage>0 && global_ipc_numsigsent==0)	{
	/* Close stdin when receiving a SIGHUP (e.g. close shell) */
	if (signo==SIGHUP &&
	    ( (info->si_code<=0 && info->si_uid==global_ipc_calleruid) ||
	      (info->si_code>0) ) && 
	      (global_ipc_stdinopts & LINGER_STDINCLOSE)==0 )
	{
	    /* SIGHUP is typically sent when the terminal is closed: no longer
	     * tty, so reopen stdin as /dev/null and payload no longer takes tty */
	    glexec_signal_log(GLEXEC_LOG_NOTICE,
		"Caught SIGHUP, redirecting stdin for lingering gLExec "
		"to /dev/null.\n");
	    /* Reopen /dev/null as stdin, failure is NOT considered fatal. */
	    if ( (fd=open("/dev/null", O_RDONLY))==-1 ||
		 (fd!=STDIN_FILENO &&
		  (dup2(fd, STDIN_FILENO)==-1 || close(fd)==-1) ) )
		glexec_signal_log(GLEXEC_LOG_WARNING,
		    "Redirecting stdin for lingering gLExec failed.\n");
	    else    {
		global_ipc_stdinopts|=LINGER_STDINCLOSE;
		global_ipc_stdinopts&=~(unsigned int)PAYLOAD_TAKETTY;
	    }
	}

	/* Forwarding of signals to payload */

	/* Make sure not to reset the global_ipc_caught_signal, as we might
	 * prevent a sigterm or sigalarm. Also don't set alarm, because whoever
	 * set global_ipc_caught_signal also should have set alarm */
	if (global_ipc_caught_signal!=0)    {
	    glexec_signal_log(GLEXEC_LOG_NOTICE,
		"Not forwarding signal %d, as we already have a signal waiting.\n",
		signo);
	} else { /* global_ipc_caught_signal==0 for payload */
	    /* Active child: forward */
	    if (info->si_code<=0)
		glexec_signal_log(GLEXEC_LOG_DEBUG,
			"Will forward signal %d from pid=%d/uid=%u to pid %d.\n",
			signo,info->si_pid,info->si_uid,global_ipc_childpid);
	    else
		glexec_signal_log(GLEXEC_LOG_DEBUG,
			"Will forward signal %d (si_code=%d) to pid %d.\n",
			signo,info->si_code,global_ipc_childpid);

	    global_ipc_caught_signal=signo;
	    glexec_alarm(ALARM_PAUSE);
	}
    } else if (global_ipc_epilogstage>0 && global_ipc_numsigsent==0)	{
	/* Only forward signals SIGTSPT (in special cases) or SIGCONT during
	 * global_ipc_epilogstage */
	if (signo==SIGTSTP || signo==SIGCONT)	{
	    /* when we were called by root (Linux) or signalled by root (MacOS)
	     * forward the signal to payload */
	    if ( (info->si_code>0 && global_ipc_calleruid==0) ||
		 (info->si_code<=0 && info->si_uid==0) )    {
		/* Make sure not to reset the global_ipc_caught_signal, as we
		 * might prevent a sigterm or sigalarm */
		if (global_ipc_caught_signal!=0)    {
		    glexec_signal_log(GLEXEC_LOG_NOTICE,
			"Not forwarding signal %d, "
			"as we have already a signal waiting to be forwarded.\n",
			signo);
		} else {
		    /* Forward to epilogue: either SIGTSTP or SIGCONT */
		    if ( info->si_code<=0)
			glexec_signal_log(GLEXEC_LOG_DEBUG,
			    "Will forward signal %d from pid=%d/uid=%u to pid %d.\n",
			    signo,info->si_pid,info->si_uid,global_ipc_epilogpid);
		    else
			glexec_signal_log(GLEXEC_LOG_DEBUG,
			    "Will forward signal %d (si_code=%d) to pid %d.\n",
			    signo,info->si_code,global_ipc_epilogpid);

		    global_ipc_caught_signal=signo;
		    /* Cannot set alarm, have already walltime */
		}
	    } else {
		/* Signal comes from either a non-root glexec caller (e.g.
		 * Linux) or non-root process (probably ctrl-z on MacOS) */
		if ( info->si_code<=0)
		    glexec_signal_log(GLEXEC_LOG_NOTICE,
			"Ignoring signal %d from pid=%d/uid=%u, "
			"payload has already finished.\n",
			signo,info->si_pid,info->si_uid);
		else
		    glexec_signal_log(GLEXEC_LOG_NOTICE,
			"Ignoring signal %d: "
			"payload has already finished.\n",signo);
		if (signo==SIGTSTP) {
		    /* Log on cmdline such that user knows what's happening */
		    glexec_write_stderr(
			"[gLExec]: Ignoring caught SIGTSTP from non-root since "
			"payload has finished.\n");
		}
	    }
	} else {
	    /* all other cases should be ignored */
	    glexec_signal_log(GLEXEC_LOG_NOTICE,
		    "Ignoring signal %d (si_code=%d)\n",
		    signo,info->si_code);
	}
    }
}

/*!
 * Handler for SIGBUS, SIGFPE, SIGILL, SIGSEGV when received from kernel and
 * SIGABRT and SIGSYS: typically not from kernel, but still called when
 * something is hopelessly wrong (e.g. failed malloc).
 * Sends a SIGTERM to children, without setting alarm and exits.
 * The signal forwarding is not fully signal safe, since we need to call
 * seteuid() but we are doing damage control in any case.
 * Note that most of the logs are lost when logging to syslog as we cannot
 * reliably log to syslog from a sighandler.
 */
static void glexec_fatal_action(int signo, siginfo_t *info, void *context)  {
    /* Check the signal is not send from another user process: when si_code<=0
     * it's a process: that's only acceptable if pid is ours and signal is
     * either SIGABRT or SIGSYS */
    if ( info->si_code<=0 && 
	 (info->si_pid!=getpid() || (signo!=SIGABRT && signo!=SIGSYS)) ) {
	glexec_signal_log(GLEXEC_LOG_NOTICE,
		"Ignoring signal %d from pid=%d/uid=%u "
		"(only acceptable from kernel).\n",
		signo,info->si_pid,info->si_uid);
	return;
    }

    /* Handle according to stage */
    if (global_ipc_childstage>0)	{ /* Active child: terminate */
	if (signo==SIGABRT || signo==SIGSYS)
	    /* SIGABRT and SIGSYS don't set address */
	    glexec_signal_log(GLEXEC_LOG_ERROR,
		    "GLExec caught signal %d, this looks like a BUG, "
		    "sending TERM to payload process %s %d and exiting.\n",
		    signo,global_ipc_usesetpgid ? "group" : "id",
		    global_ipc_childpid);
	else
	    glexec_signal_log(GLEXEC_LOG_ERROR,
		    "GLExec caught signal %d, address 0x%x, "
		    "this looks like a BUG, "
		    "sending TERM to payload process %s %d and exiting.\n",
		    signo,info->si_addr, global_ipc_usesetpgid ? "group" : "id",
		    global_ipc_childpid);

	/* Kill payload: use global_ipc_childstage instead of
	 * global_ipc_childpid as the latter might not yet be initialized, while
	 * the former is positive and hence a valid pid */
	glexec_emergency_send_signal(
		global_ipc_usesetpgid ? -global_ipc_childstage
		                      : global_ipc_childstage,
		SIGTERM, &global_ipc_childstage);

	/* Log accounting information */
	glexec_accounting(1,1);
    } else if (global_ipc_epilogstage>0)	{ /* Active epilog: terminate */
	if (signo==SIGABRT || signo==SIGSYS)
	    /* SIGABRT and SIGSYS don't set address */
	    glexec_signal_log(GLEXEC_LOG_ERROR,
		    "GLExec caught signal %d, this looks like a BUG, "
		    "sending TERM to epilogue process %s %d and exiting.\n",
		    signo,global_ipc_usesetpgid ? "group" : "id",
		    global_ipc_epilogpid);
	else
	    glexec_signal_log(GLEXEC_LOG_ERROR,
		    "GLExec caught signal %d, address 0x%x, "
		    "this looks like a BUG, "
		    "sending TERM to epilogue process %s %d and exiting.\n",
		    signo,info->si_addr, global_ipc_usesetpgid ? "group" : "id",
		    global_ipc_epilogpid);

	/* Kill epilogue: use global_ipc_epilogstage instead of
	 * global_ipc_epilogpid as the latter might not yet be initialized, while
	 * the former is positive and hence a valid pid  */
	glexec_emergency_send_signal(
		global_ipc_usesetpgid ? -global_ipc_epilogstage
		                      : global_ipc_epilogstage,
		SIGTERM, &global_ipc_epilogstage);
    } else  {
	if (signo==SIGABRT || signo==SIGSYS)
	    /* SIGABRT and SIGSYS don't set address */
	    glexec_signal_log(GLEXEC_LOG_ERROR,
		    "GLExec caught signal %d, this looks like a BUG.\n",
		    signo);
	else
	    glexec_signal_log(GLEXEC_LOG_ERROR,
		    "GLExec caught signal %d, address 0x%x, "
		    "this looks like a BUG.\n",
		    signo,info->si_addr);
    }

    /* When child has finished log its exit code since we'll lose it */
    if (global_ipc_childrc!=GLEXEC_NO_CHILD_EXIT_CODE)	{
	glexec_signal_log(global_ipc_childrc ? GLEXEC_LOG_ERROR
		                             : GLEXEC_LOG_NOTICE,
		"Payload %d exited with exit code %d\n",
		global_ipc_childpid,global_ipc_childrc);
    }

    /* Can safely call with global_ipc_childrc as it is initialized with
     * GLEXEC_NO_CHILD_EXIT_CODE */
    glexec_signal_exit(GLEXEC_PREMATURE_EXIT,global_ipc_childrc,signo);
}
