/**
 * Copyright (c) Members of the EGEE Collaboration. 2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 * Authors: Oscar Koeroo, Mischa Sall\'e
 *      NIKHEF Amsterdam, the Netherlands
 *      <grid-mw-security@nikhef.nl>  
 */

/*
 * glexec_lcmaps provides the functionality to dynamically open lcmaps and its
 * API
 */

#include "glexec_lcmaps.h"

#include "glexec.h"
#include "glexec_conf.h"
#include "glexec_log.h"
#include "glexec_environ.h"
#include "glexec_util.h"
#include "glexec_account.h"

#include "fileutil/fileutil.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h> /* For GLEXEC_LOG_FILEMASK / GLEXEC_LOG_DIRMASK */
#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>

#define LCMAPS_USE_DLOPEN
#define LCMAPS_UGLY_VERSION_BUG_HACK
#include <lcmaps/lcmaps_return_account_from_pem.h>

/*************************************************************************/
/* INTERNAL FUNCTION DECLARATIONS                                        */
/*************************************************************************/

/*
 * Loads the LCMAPS library and its functions using dlopen() and dlsym().
 * For older versions of the LCMAPS library, it is necessary to provide the
 * functions getMajorVersion() getMinorVersion and getPatchVersion(), which can
 * be provided by LCAS. This function automatically loads LCAS when needed.
 * returns 0 on success and -1 on error.
 */
static int glexec_lcmaps_load(const char *libdir,
			      LCMAPS_HANDLE_RETURN_ACCOUNT_FROM_PEM *lcmaps);

/*
 * Closes the LCMAPS library.
 * returns 0 on success and -1 on error.
 */
static int glexec_lcmaps_close(LCMAPS_HANDLE_RETURN_ACCOUNT_FROM_PEM *lcmaps);

/*
 * Setup LCMAPS environment variables
 * sets libdir (might be "") where to find lcmaps*.so
 */
static void glexec_lcmaps_env_setup(const char **libdir);

/*
 * Open LCMAPS logging.
 * returns pointer to glexec_lcmaps_log file (might be NULL)
 */
static FILE *glexec_lcmaps_log_setup(LCMAPS_HANDLE_RETURN_ACCOUNT_FROM_PEM *lcmaps);

/*
 * Converts an lcmaps_account_info_t into a glexec_account_t for the
 * target user.
 * returns GLEXEC_SUCCESS on success
 */
static int glexec_lcmaps_get_account_info(lcmaps_account_info_t *account_info,
					  glexec_account_t *target);

/*************************************************************************/
/* FUNCTIONS			                                         */
/*************************************************************************/

/*
 * Setup, run and close LCMAPS, returning a target account
 * returns GLEXEC_SUCCESS when successful
 */
int glexec_lcmaps(char *pem_string, glexec_account_t *target_account)	{
    const char *lcmaps_libdir;
    FILE *lcmaps_log;
    LCMAPS_HANDLE_RETURN_ACCOUNT_FROM_PEM lcmaps;
    /* Alternatively, can use following macro WITHOUT trailing ; */
    /*    LCMAPS_DECL_HANDLE_RETURN_ACCOUNT_FROM_PEM(lcmaps) */
    lcmaps_account_info_t paccount_info;
    int rc;

    /* Set up the LCMAPS handle */
    LCMAPS_INIT_HANDLE_RETURN_ACCOUNT_FROM_PEM(&lcmaps);

    /* Setup LCMAPS environment */
    glexec_lcmaps_env_setup(&lcmaps_libdir);

    /* Load LCMAPS library */
    if (glexec_lcmaps_load(lcmaps_libdir,&lcmaps))
	return GLEXEC_LCMAPS_INIT_FAIL;

    /* Setup LCMAPS log */
    lcmaps_log=glexec_lcmaps_log_setup(&lcmaps);

    /* Determine if we want to skip voms attr checking (only new API) */
    if (LCMAPS_MAJOR_VERSION(&lcmaps) > 0) {
	/* Default is yes (1) */
	if (glexec_conf_boolean("glexec:lcmaps_voms_verification", 1))
	    LCMAPS_CALL(&lcmaps,lcmaps_enable_voms_attributes_verification)();
	else {
	    glexec_log(GLEXEC_LOG_INFO,
		       "Disabling LCMAPS VOMS attribute verification.\n");
	    LCMAPS_CALL(&lcmaps,
			lcmaps_disable_voms_attributes_verification)();
	}
    }

    /* Init will return 0 on success or otherwise on failure, note that
     * glexec_lcmaps_log can be NULL in which case syslog is used.  */
    if (LCMAPS_CALL(&lcmaps,lcmaps_init)(lcmaps_log)) {
	rc=GLEXEC_LCMAPS_INIT_FAIL;
	goto lcmaps_finalize_noterm;
    }

    glexec_log(GLEXEC_LOG_DEBUG, "Initialisation of LCMAPS has finished.\n");

    /* LCMAPS mapping call */
    if (LCMAPS_CALL(&lcmaps,lcmaps_account_info_init)(&paccount_info) != 0) {
	glexec_log(GLEXEC_LOG_ERROR, "Error initialising LCMAPS account\n");
	rc=GLEXEC_LCMAPS_INIT_FAIL;
	goto lcmaps_finalize;
    }

    /* -1 is the mapcounter */
    if (LCMAPS_CALL(&lcmaps,lcmaps_return_account_from_pem)(pem_string,-1,&paccount_info)!=0) 
    {
	LCMAPS_CALL(&lcmaps,lcmaps_account_info_clean)(&paccount_info);
	rc=GLEXEC_LCMAPS_FAILED;
	goto lcmaps_finalize;
    }
    /* Mapping is successful: fill the target_account struct */
    rc=glexec_lcmaps_get_account_info(&paccount_info,target_account);

    /* Clean up lcmaps */
    LCMAPS_CALL(&lcmaps,lcmaps_account_info_clean)(&paccount_info);

lcmaps_finalize:
    /* terminate LCMAPS */
    LCMAPS_CALL(&lcmaps,lcmaps_term)();

lcmaps_finalize_noterm:
    /* Close lcmaps library */
    glexec_lcmaps_close(&lcmaps);

    /* Close the log file only if one is actually open (like in syslog usage
     * mode) */
    if (lcmaps_log) {
	if (fclose(lcmaps_log))
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error while closing lcmaps log file: %s\n",
		    strerror(errno));
	lcmaps_log = NULL;
    }

    /* Now check whether LCMAPS succeeded */
    if (rc==GLEXEC_SUCCESS)
	glexec_log(GLEXEC_LOG_DEBUG, "LCMAPS is finished\n");

    return rc;
}

/*
 * Print the known LCMAPS defines to file out.
 */
void glexec_lcmaps_defines(FILE *out)	{
#ifdef LCMAPS_API_MAJOR_VERSION
    fprintf(out," -D LCMAPS_API_MAJOR_VERSION=\"%d\"\n", LCMAPS_API_MAJOR_VERSION);
#endif
#ifdef LCMAPS_API_MINOR_VERSION
    fprintf(out," -D LCMAPS_API_MINOR_VERSION=\"%d\"\n", LCMAPS_API_MINOR_VERSION);
#endif
#ifdef LCMAPS_API_PATCH_VERSION
    fprintf(out," -D LCMAPS_API_PATCH_VERSION=\"%d\"\n", LCMAPS_API_PATCH_VERSION);
#endif
#ifdef LCMAPS_DB_FILE
    fprintf(out," -D LCMAPS_DB_FILE=\"%s\"\n", LCMAPS_DB_FILE);
#endif
#ifdef LCMAPS_LOG_FILE
    fprintf(out," -D LCMAPS_LOG_FILE=\"%s\"\n", LCMAPS_LOG_FILE);
#endif
#ifdef LCMAPS_DEBUG_LEVEL
    fprintf(out," -D LCMAPS_DEBUG_LEVEL=\"%s\"\n", LCMAPS_DEBUG_LEVEL);
#endif
#ifdef LCMAPS_MODULEDIR_SFX
    fprintf(out," -D LCMAPS_MODULEDIR_SFX=\"%s\"\n", LCMAPS_MODULEDIR_SFX);
#endif
#ifdef LCMAPS_LIBDIR
    fprintf(out," -D LCMAPS_LIBDIR=\"%s\"\n", LCMAPS_LIBDIR);
#endif
#ifdef LCMAPS_POLICY_NAME
    fprintf(out," -D LCMAPS_POLICY_NAME=\"%s\"\n", LCMAPS_POLICY_NAME);
#endif
}

/*************************************************************************/
/* INTERNAL FUNCTIONS			                                 */
/*************************************************************************/

/*
 * Loads the LCMAPS library and its functions using dlopen() and dlsym().
 * For older versions of the LCMAPS library, it is necessary to provide the
 * functions getMajorVersion() getMinorVersion and getPatchVersion(), which can
 * be provided by LCAS. This function automatically loads LCAS when needed.
 * returns 0 on success and -1 on error.
 */
static int glexec_lcmaps_load(const char *libdir,
			      LCMAPS_HANDLE_RETURN_ACCOUNT_FROM_PEM *lcmaps)   {
    /* char *error=dlerror(); /\* Make sure dlerror() is cleared *\/ */
    const char *lcmaps_so="liblcmaps_return_account_from_pem" LIBSUFF;
    const char *helper_so="liblcas" LIBSUFF;
    char lcmaps_lib_path[AP_MAXPATH],helper_lib_path[AP_MAXPATH];
    const char *lcmaps_lib,*helper_lib;
    int rc_sn;

    /* prefix lcmaps library name with libdir when needed */
    if (strlen(libdir)>0)   {
	rc_sn=snprintf(lcmaps_lib_path,(size_t)AP_MAXPATH,"%s/%s",
		       libdir,lcmaps_so);
	if (rc_sn<0)	{
    	    glexec_log(GLEXEC_LOG_ERROR,"Cannot create path for liblcmaps\n");
    	    return -1;
    	}
    	if (rc_sn>=AP_MAXPATH)    {
    	    glexec_log(GLEXEC_LOG_ERROR,"Full path %s/%s is too long\n",
    		    libdir,lcmaps_so);
    	    return -1;
    	}
	rc_sn=snprintf(helper_lib_path,(size_t)AP_MAXPATH,"%s/%s",
		       libdir,helper_so);
	if (rc_sn<0)	{
    	    glexec_log(GLEXEC_LOG_ERROR,"Cannot create path for helper lib\n");
    	    return -1;
    	}
	if (rc_sn>=AP_MAXPATH)    {
	    glexec_log(GLEXEC_LOG_ERROR,"Full path %s/%s is too long\n",
		       libdir,helper_so);
	    return -1;
	}
    	lcmaps_lib=lcmaps_lib_path;
	helper_lib=helper_lib_path;
    } else {
    	lcmaps_lib=lcmaps_so;
	helper_lib=helper_so;
    }


    /* Load the library lazily: this allows checking for missing symbols */
    glexec_log(GLEXEC_LOG_DEBUG, "Opening LCMAPS lib: %s\n",lcmaps_lib);

    LCMAPS_SET_LIBFILE_PATH(lcmaps,lcmaps_lib);
    LCMAPS_SET_HELPER_PATH(lcmaps,helper_lib);
    if(!LCMAPS_LOAD_RETURN_ACCOUNT_FROM_PEM(lcmaps)) {
	glexec_log(GLEXEC_LOG_ERROR,"Error opening LCMAPS lib: %s\n",
		   LCMAPS_ERRMSG(lcmaps));
        if (!LCMAPS_CLOSE_HANDLE_RETURN_ACCOUNT_FROM_PEM(lcmaps)) {
	    glexec_log(GLEXEC_LOG_ERROR, "Error closing LCMAPS library: %s\n",
		       LCMAPS_ERRMSG(lcmaps));
	}
	return -1;
    }

    /* New API functions */
    if (LCMAPS_MAJOR_VERSION(lcmaps) > 0) {
	glexec_log(GLEXEC_LOG_INFO,"LCMAPS version is %d.%d.%d\n",
		LCMAPS_MAJOR_VERSION(lcmaps),
		LCMAPS_MINOR_VERSION(lcmaps),
		LCMAPS_PATCH_VERSION(lcmaps));

    	/* Now load the 'OSG functionality': this has to exist since
	 * LCMAPS_MAJOR_VERSION is non-zero */
    	if (! LCMAPS_REQUIRE_FUNC(lcmaps,
		    lcmaps_disable_voms_attributes_verification) ||
	    ! LCMAPS_REQUIRE_FUNC(lcmaps,
		    lcmaps_enable_voms_attributes_verification) /* ||
	    ! LCMAPS_REQUIRE_FUNC(lcmaps,
		    lcmaps_is_set_to_verify_voms_attributes)*/)	{
		glexec_log(GLEXEC_LOG_ERROR,
			   "New LCMAPS API found, but could not load API "\
			   "dis-/enabling voms attr checking NOT found: %s\n",
			   LCMAPS_ERRMSG(lcmaps));
		return -1;
	}
	/* OSG functionality is present */
	/* lcmaps->voms_attr_check=1; */
	glexec_log(GLEXEC_LOG_DEBUG,
		"LCMAPS API for dis-/enabling voms attr checking loaded.\n");
    } else {
	glexec_log(GLEXEC_LOG_INFO,"LCMAPS version is pre 1.4.21\n",
		LCMAPS_MAJOR_VERSION(lcmaps),
		LCMAPS_MINOR_VERSION(lcmaps),
		LCMAPS_PATCH_VERSION(lcmaps));
    }
    return 0;
}

/*
 * Closes the LCMAPS library.
 * returns 0 on success and -1 on error.
 */
static int glexec_lcmaps_close(LCMAPS_HANDLE_RETURN_ACCOUNT_FROM_PEM *lcmaps)	{

    glexec_log(GLEXEC_LOG_DEBUG,"Closing LCMAPS library.\n");
    if(!LCMAPS_CLOSE_HANDLE_RETURN_ACCOUNT_FROM_PEM(lcmaps)) {
	glexec_log(GLEXEC_LOG_ERROR,"Error closing LCMAPS library: %s\n",
		   LCMAPS_ERRMSG(lcmaps));
	return -1;
    }
    return 0;

}

/*
 * Setup LCMAPS environment variables, except LCMAPS_LOG_STRING which will be
 * setup in the lcmaps_log function, since we need it to be dlopened beforehand.
 * set libdir (might be "") where to find lcmaps*.so
 */
static void glexec_lcmaps_env_setup(const char **lcmaps_libdir) {
    const char *conflibdir,*libdir;
    const char *moddir_suffix;
    char modules_dir[AP_MAXPATH];
    int rc_sn;

    /* setenv("LCMAPS_VERIFY_TYPE", "uid"); */
    /* Disabling the blocking factor to use multiple LCMAPS policies */
    setenv("LCMAPS_VERIFY_TYPE", "uid_pgid", 1);

    glexec_setenv_regex("LCMAPS_DEBUG_LEVEL",
	glexec_conf_value_default("glexec:lcmaps_debug_level",
				  LCMAPS_DEBUG_LEVEL),
	"^[0-5]$");
    glexec_setenv_regex("LCMAPS_DB_FILE",
	glexec_conf_value_default("glexec:lcmaps_db_file",LCMAPS_DB_FILE),
	"\\/(.*\\/)+[^/]*");

    /* Setup LCMAPS policy */
    setenv( "LCMAPS_POLICY_NAME",
	    glexec_conf_value_default("glexec:lcmaps_get_account_policy",
					LCMAPS_POLICY_NAME),
	    1 );

    /* Setup LCMAPS libdir, which will be used for dlopen-ing LCMAPS */

    /* First see if there is a config file value */
    conflibdir=glexec_conf_value("glexec:lcmaps_libdir");
    if ( conflibdir )	{
	if (conflibdir[0]=='\0' || glexec_dir_is_abs(conflibdir))
	    libdir=conflibdir;
	else	{
	    glexec_log(GLEXEC_LOG_WARNING,
		"Warning: ignoring lcmaps_libdir config option: "
		"\"%s\" is not an absolute path to a valid directory.\n",
		conflibdir);
	    libdir=LCMAPS_LIBDIR;
	}
    } else
	libdir=LCMAPS_LIBDIR;

    /* If conflibdir didn't do the trick, try LCMAPS_LIBDIR */
    if (libdir!=conflibdir && libdir[0]!='\0' && !glexec_dir_is_abs(libdir))   {
	glexec_log(GLEXEC_LOG_WARNING,
	    "Warning: ignoring built-in LCMAPS_LIBDIR: "
	    "\"%s\" is not an absolute path to a valid directory.\n",
	    libdir);
	libdir="";
    }
    *lcmaps_libdir=libdir;

    /* If it's non-empty, also setup LCMAPS_MODULES_DIR. */
    if ( libdir[0]!='\0')    { /* No need for strlen... */
	glexec_log(GLEXEC_LOG_DEBUG,"Will look for LCMAPS in '%s'\n",libdir);

	/* Get suffix where to look for LCMAPS modules, used to be /modules,
	 * changed to /lcmaps to satisfy Fedora guidelines */
	moddir_suffix=glexec_conf_value_default("glexec:lcmaps_moduledir_sfx",
						LCMAPS_MODULEDIR_SFX);

	rc_sn=snprintf(modules_dir,(size_t)AP_MAXPATH,"%s%s",
		       libdir, moddir_suffix);
	if (rc_sn < 0)
	    glexec_log(GLEXEC_LOG_WARNING,
		"Cannot create modulespath, not setting LCMAPS_MODULES_DIR\n");
	else if (rc_sn >= AP_MAXPATH)
	    glexec_log(GLEXEC_LOG_WARNING,
		"Full modulespath '%s%s' would be too long, " \
		"not setting LCMAPS_MODULES_DIR\n",
		libdir,moddir_suffix);
	else	{
	    if (!glexec_dir_is_abs(modules_dir))
		glexec_log(GLEXEC_LOG_WARNING,
		    "Warning: modulespath '%s' does not point to absolute "
		    "directory, not setting LCMAPS_MODULES_DIR\n",modules_dir);
	    else    {
		glexec_log(GLEXEC_LOG_DEBUG,
		    "Setting LCMAPS_MODULES_DIR to '%s'\n", modules_dir);
		setenv("LCMAPS_MODULES_DIR", modules_dir, 1);
	    }
	}
    }

}

/*
 * Open LCMAPS logging.
 * returns pointer to glexec_lcmaps_log file (might be NULL)
 */
static FILE *glexec_lcmaps_log_setup(LCMAPS_HANDLE_RETURN_ACCOUNT_FROM_PEM *lcmaps) {
    const mode_t file_mask=GLEXEC_LOG_FILEMASK; /* File permissions */
    const mode_t dir_mask=GLEXEC_LOG_DIRMASK; /* Dir permissions */
    int rc;
    gid_t oldgid=getegid(),lcmaps_loggid;
    char *jobrepoid;
    char *lcmaps_log_name;
    FILE *lcmaps_log=NULL;

    /* Optionally open logfile */
    if (glexec_get_log_to_syslog()) { /* Log to SysLog */
	/*  Make sure the environment variables are never set when glexec has to
	 *  use syslog.*/
        unsetenv("LCMAPS_LOG_FILE");
    } else { /* logfile */
	/* LCMAPS_LOG_STRING for new version is based on executable name and
	 * LCMAPS_LOG_STRING based on JOB_REPOSITORY_ID for old version */
	if ( (LCMAPS_MAJOR_VERSION(lcmaps) == 1 &&
		    LCMAPS_MINOR_VERSION(lcmaps) > 4) ||
	     LCMAPS_MAJOR_VERSION(lcmaps) > 1 )
	    setenv("LCMAPS_LOG_IDENT", GLEXEC_SYSLOG_IDENT , 1);
	else    {
	    /* Get jobrepo ID for old style behaviour */
	    jobrepoid=getenv("JOB_REPOSITORY_ID");
	    setenv("LCMAPS_LOG_STRING", (jobrepoid ? jobrepoid : "NULL"), 1);
	}

	/* LCMAPS_LOG_FILE */
	glexec_setenv_regex("LCMAPS_LOG_FILE",
	    glexec_conf_value_default("glexec:lcmaps_log_file",LCMAPS_LOG_FILE),
	    "\\/(.*\\/)+[^/]*");
        if ( (lcmaps_log_name = getenv("LCMAPS_LOG_FILE"))==NULL )   {
            glexec_log(GLEXEC_LOG_ERROR,
		"Log file for LCMAPS could not be opened: "\
		"No log file name was specified.\n");
            glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	}
	/* Get logging GID as set by glexec */
	lcmaps_loggid=glexec_get_loggid();
	/* Set the effective gid, so that file and dir will be created with the
	 * right ownership */
	if (setegid(lcmaps_loggid))
	    glexec_log(GLEXEC_LOG_WARNING,
		"Setting effective gid to %d failed.\n", (int)lcmaps_loggid);
	/* Now try to open the logfile with file permission file_mask and
	 * directory permissions dir_mask. Don't force it to be trusted. */
	rc=cgul_open_logfile(lcmaps_log_name,file_mask,dir_mask,&lcmaps_log,0);
	/* reset effective gid */
	if (setegid(oldgid))
	    glexec_log(GLEXEC_LOG_WARNING,
		"Setting effective gid to %d failed.\n", (int)oldgid);
	/* Parse return code of cgul_open_logfile */
	switch (rc)	{
	    case 0: break;
	    case -1: 
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: error creating dir for LCMAPS logfile '%s'.\n",
		    lcmaps_log_name);
		glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	    case -2:
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: I/O error during opening LCMAPS logfile: %s.\n",
		    strerror(errno));
		glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	    case -3:
		glexec_log(GLEXEC_LOG_WARNING,
		    "Warning: Permissions for LCMAPS logfile are wrong "
		    "(should be: user root, "
		    "permissions: file at most 750, dir at most 755). "
		    "In next version this will be a fatal error.\n");
		break;
	    case -4:
		glexec_exit(GLEXEC_OUT_OF_MEMORY, GLEXEC_NO_CHILD_EXIT_CODE);
	    case -5:
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: LCMAPS logfile '%s' is not an absolute path.\n",
		    lcmaps_log_name);
		glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	    case -6:
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: LCMAPS logfile is no longer the "
		    "one that was opened!.\n");
		glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	    case -7:
		glexec_log(GLEXEC_LOG_ERROR,
		    "Error: Specified LCMAPS logfile is not a regular file.\n");
		glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	    case -8:
	    default:
		glexec_exit(GLEXEC_LCMAPS_INIT_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	}

	/* Log the log filename */
	glexec_log(GLEXEC_LOG_DEBUG,
	    "LCMAPS will be initialised with log file: %s\n", lcmaps_log_name);
    }

    return lcmaps_log;
}

/*
 * Converts an lcmaps_account_info_t into a glexec_account_t for the
 * target user.
 * returns GLEXEC_SUCCESS on success
 */
static int glexec_lcmaps_get_account_info(lcmaps_account_info_t *account_info,
					  glexec_account_t *target) {
    struct passwd *pw;
    struct group *pgr,*sgr;
    int i;

    /* Initialize target account structure */
    target->uid=GLEXEC_UID_UNDEFINED;
    target->gid=GLEXEC_GID_UNDEFINED;
    target->uname=target->dir=target->gname=NULL;
    target->nsgid=0;
    target->sgid=NULL;

    /* UID */
    if ( (pw = glexec_getpwuid(account_info->uid))==NULL )	{
	glexec_log(GLEXEC_LOG_ERROR,
	      "Error looking up uid (%d)\n", account_info->uid);
	return GLEXEC_UNKNOWN_USER;
    }
    target->uid = account_info->uid;
    target->uname = pw->pw_name;
    /* homedir */
    target->dir = pw->pw_dir;
    if (target->dir==NULL)
	glexec_log(GLEXEC_LOG_DEBUG,
		"User %s(%s) has no home directory.\n",
		target->uid,
		target->uname ? target->uname : "unknown");

    /* primary GID */
    if (account_info->npgid == 0) {
	glexec_log(GLEXEC_LOG_ERROR, "Need at least 1 gid!\n");
	return GLEXEC_INCOMPLETE_MAPPING_RESULT;
    }
    if ( (pgr = glexec_getgrgid(account_info->pgid_list[0]))==NULL ) {
	glexec_log(GLEXEC_LOG_ERROR,
		"Error looking up gid (%d)\n", account_info->pgid_list[0]);
	return GLEXEC_UNKNOWN_GROUP;
    }
    target->gid = account_info->pgid_list[0];
    target->gname = pgr->gr_name;

    /* Print out of uid and primary GID */
    glexec_log(GLEXEC_LOG_INFO,
		"Target uid = %d(%s), gid = %d(%s)\n",
		target->uid, target->uname, target->gid, target->gname);

    /* secondary GIDs */
    if ( (target->nsgid = account_info->nsgid) > 0) {
	/* target->nsgid is quaranteed bigger than 0 here */
	target->sgid = (gid_t *)gmalloc(sizeof(gid_t)*(size_t)target->nsgid);

	for (i = 0; i < target->nsgid; i++) {
	    /* First copy the GID itself */
	    target->sgid[i] = account_info->sgid_list[i];
	    /* Get account information if available */
	    if ( (sgr=glexec_getgrgid(target->sgid[i]))==NULL )
		/* Print notice, but include gid */
		glexec_log(GLEXEC_LOG_NOTICE,
		    "Could not look up name for target secondary gid[%d] = %d\n",
		    i,target->sgid[i]);
	    else
		/* Print found secondary GID and GROUP */
		glexec_log(GLEXEC_LOG_INFO,
		    "Target secondary gid[%d] = %d(%s)\n",
		    i,target->sgid[i], sgr->gr_name);
	    
	}
    }
    /* Protect against strange values in target->nsgid */
    if (target->nsgid < 0)
	target->nsgid=0;

    return GLEXEC_SUCCESS;
}

