/**
 * Copyright (c) Members of the EGEE Collaboration. 2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 * Authors: Oscar Koeroo, Mischa Sall\'e
 *      NIKHEF Amsterdam, the Netherlands
 *      <grid-mw-security@nikhef.nl>  
 */

/*!
 * \file glexec_log.c provides the logging functionality.
 */

#include "glexec_log.h"

#include "glexec.h"
#include "glexec_exit.h"
#include "glexec_ipc.h"
#include "glexec_conf.h"
#include "glexec_util.h"
#include "glexec_account.h"
#include "fileutil/fileutil.h"

#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <grp.h>

/*************************************************************************/
/* DEFINES								 */
/*************************************************************************/

/* Size of 33 is good enough for PID length of 5 and name glexec, just make it
 * somewhat bigger */
#define GLEXEC_LOGPREFIX_LENGTH	50

/* Options to use when opening syslog */
#ifndef GLEXEC_SYSLOG_OPTS
#define GLEXEC_SYSLOG_OPTS  (LOG_CONS|LOG_PID)
#endif

/* Default syslog priority  */
#ifndef GLEXEC_SYSLOG_DEF_PRIORITY
#define GLEXEC_SYSLOG_DEF_PRIORITY LOG_ERR
#endif

/*************************************************************************/
/* GLOBAL VARIABLES							 */
/*************************************************************************/

/* Default log level NOTICE, note: this cutoff is ignored in case of logging to
 * syslog */
static glexec_log_level_t global_log_level = GLEXEC_LOG_INFO;

/* Default destination is known and syslog */
static int global_log_to_file	= 0;  /*! Default is NOT to log to a file
					directly, \see global_log_to_syslog */
static int global_log_to_syslog = 1;  /*! Default is to log to syslog. \see
					global_log_to_file */
static int global_syslog_opened	= 0;

/* Whether glexec's internal loglevels correspond to different syslog levels.
 * Default is 1, i.e. map internal loglevels to those of syslog */
static int global_diff_syslog_levels = 1;

/* Which type of syslog facility to use, default is GLEXEC_SYSLOG_FACILITY */
static int global_syslog_facility = GLEXEC_SYSLOG_FACILITY;

/* glexec log file */
static FILE *global_log_file    = NULL;
static int global_log_fd = -1;

/* glexec log gid: also used for LCAS & LCMAPS, can be obtained using getter
 * glexec_get_loggid() */
static gid_t global_log_gid=0;

/* Mapping from gLExec log levels to syslog levels */
static const int global_syslog_level[GLEXEC_LOG_LEVEL_HIGH]={
    	/* LOG_EMERG unused */
	/* LOG_ALERT unused */ 
	LOG_CRIT,	/* GLEXEC_LOG_LEVEL_LOW */
	LOG_ERR,	/* GLEXEC_LOG_ERROR */
	LOG_WARNING,	/* GLEXEC_LOG_WARNING */
	LOG_NOTICE,	/* GLEXEC_LOG_NOTICE */
	LOG_INFO,	/* GLEXEC_LOG_INFO */
	LOG_DEBUG};	/* GLEXEC_LOG_DEBUG */

/* logging buffer for signal logging function */
typedef struct {
    char body[GLEXEC_MAX_LOG_LINE];
    int loglevel;
} log_entry_t;
static log_entry_t global_log_buffer[GLEXEC_MAX_BUFFER_LINES];
static int global_log_idx=0;

/* log prefix when logging to file. Make this global so that we don't have to
 * set it up when called from a signal handler */
static char global_log_prefix[GLEXEC_LOGPREFIX_LENGTH];
/* Length of global_log_prefix */
static size_t global_log_prefix_length=0;

/*************************************************************************/
/* FUNCTION PROTOTYPES							 */
/*************************************************************************/

/* Open logfile */
static void glexec_open_log_file(void);

/* Open syslog */
static void glexec_open_syslog(void);

/* Create logging prefix when logging to file 
 * \return 0 upon success, -1 on error */
static int glexec_setup_log_prefix(void);

/*************************************************************************/
/* FUNCTIONS								 */
/*************************************************************************/

/*!
 * Set logging to syslog (unsets to file)
 */
void glexec_set_log_to_syslog(void) {
    global_log_to_syslog=1;
    global_log_to_file=0;
}

/* Get whether logging to syslog is enabled
 * \return 1 when logging to syslog, otherwise 0*/
int glexec_get_log_to_syslog(void)  {
    return global_log_to_syslog;
}

/*!
 * Set logging to file (unsets to syslog)
 */
void glexec_set_log_to_file(void)   {
    global_log_to_syslog=0;
    global_log_to_file=1;
}

/*!
 * Get whether logging to file is enabled
 * \return 1 when logging to file, otherwise 0
 * */
int glexec_get_log_to_file(void)  {
    return global_log_to_file;
}

/*!
 * Getter function for glexec logfile gid
 * \return glexec log gid
 */
gid_t glexec_get_loggid(void)	{
    return global_log_gid;
}

/*!
 *  Set the log level for glexec taking into account the lowest and
 *  highest possible log level.
 *  \param loglevel  New log level.
 *  \param diff_syslog_levels  when 1 glexec_log_levels will be translated into
 *         different syslog levels (only in case logging to syslog of course).
 */
void glexec_set_log_level(glexec_log_level_t loglevel, int diff_syslog_levels) {
    /* Set whether to use different syslog levels */
    global_diff_syslog_levels=diff_syslog_levels;

    /* Notice: no need to truncate a too small loglevel: glexec_loglevel_t is an
     * enum, which is unsigned, and GLEXEC_LOG_LEVEL_LOW is 0 */
    if (loglevel>=GLEXEC_LOG_LEVEL_HIGH) {
	/* truncate a too large loglevel */
	global_log_level = GLEXEC_LOG_LEVEL_HIGH-1;
	glexec_log(GLEXEC_LOG_WARNING,
	    "Specified loglevel is too large, using max. log level %d.\n",
	    global_log_level);
    } else {
	global_log_level = loglevel;
	glexec_log(GLEXEC_LOG_DEBUG, "Setting log level to: %d\n",
		   global_log_level);
    }

    /* Only now log whether we differentiate, so that we take the loglevel into
     * account */
    if (diff_syslog_levels)
	glexec_log(GLEXEC_LOG_DEBUG, "Using different syslog loglevels.\n");
}

/*!
 * Set the correct syslog facility using specified facility, and reopen when
 * it was already open. When facility is NULL return the current one. Default is
 * GLEXEC_SYSLOG_FACILITY.
 * \param facility *char of facility to use, when NULL don't change.
 * \return facility new integer facility or current when facility == NULL.
 */
int glexec_set_syslog_facility(const char *facility) {
    /* nothing to do */
    if (facility==NULL)
	return global_syslog_facility;

    /* Now go over all the possibilities */
    if (strcmp(facility,"LOG_KERN")==0)
	global_syslog_facility=LOG_KERN;
    else if (strcmp(facility,"LOG_USER")==0)
	global_syslog_facility=LOG_USER;
    else if (strcmp(facility,"LOG_MAIL")==0)
	global_syslog_facility=LOG_MAIL;
    else if (strcmp(facility,"LOG_NEWS")==0)
	global_syslog_facility=LOG_NEWS;
    else if (strcmp(facility,"LOG_UUCP")==0)
	global_syslog_facility=LOG_UUCP;
    else if (strcmp(facility,"LOG_DAEMON")==0)
	global_syslog_facility=LOG_DAEMON;
    else if (strcmp(facility,"LOG_AUTH")==0)
	global_syslog_facility=LOG_AUTH;
    else if (strcmp(facility,"LOG_CRON")==0)
	global_syslog_facility=LOG_CRON;
    else if (strcmp(facility,"LOG_LPR")==0)
	global_syslog_facility=LOG_LPR;
    else if (strcmp(facility,"LOG_LOCAL0")==0)
	global_syslog_facility=LOG_LOCAL0;
    else if (strcmp(facility,"LOG_LOCAL1")==0)
	global_syslog_facility=LOG_LOCAL1;
    else if (strcmp(facility,"LOG_LOCAL2")==0)
	global_syslog_facility=LOG_LOCAL2;
    else if (strcmp(facility,"LOG_LOCAL3")==0)
	global_syslog_facility=LOG_LOCAL3;
    else if (strcmp(facility,"LOG_LOCAL4")==0)
	global_syslog_facility=LOG_LOCAL4;
    else if (strcmp(facility,"LOG_LOCAL5")==0)
	global_syslog_facility=LOG_LOCAL5;
    else if (strcmp(facility,"LOG_LOCAL6")==0)
	global_syslog_facility=LOG_LOCAL6;
    else if (strcmp(facility,"LOG_LOCAL7")==0)
	global_syslog_facility=LOG_LOCAL7;
    /* Now the non-standard ones... */
#ifdef LOG_SYSLOG
    else if (strcmp(facility,"LOG_SYSLOG")==0)
	global_syslog_facility=LOG_SYSLOG;
#endif
#ifdef LOG_FTP
    else if (strcmp(facility,"LOG_FTP")==0)
	global_syslog_facility=LOG_FTP;
#endif
#ifdef LOG_AUTHPRIV
    else if (strcmp(facility,"LOG_AUTHPRIV")==0)
	global_syslog_facility=LOG_AUTHPRIV;
#endif
#ifdef LOG_AUDIT
    else if (strcmp(facility,"LOG_AUDIT")==0)
	global_syslog_facility=LOG_AUDIT;
#endif
    else
	global_syslog_facility=-1;

    /* Any need to change ? */
    if (global_syslog_facility==GLEXEC_SYSLOG_FACILITY)
	return global_syslog_facility;
   
    /* Is it valid? */
    if (global_syslog_facility==-1)  {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Unknown syslog global_syslog_facility '%s' specified.\n",facility);
	glexec_exit(GLEXEC_CONF_VALUE_MALFORMED, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Reopen with new global_syslog_facility */
    if (global_syslog_opened)	{
	closelog();
	openlog(GLEXEC_SYSLOG_IDENT,GLEXEC_SYSLOG_OPTS,global_syslog_facility);
	glexec_log(GLEXEC_LOG_NOTICE,
	    "Syslog has been reopened with new facility '%s'\n",facility);
    }
    return global_syslog_facility;
}

/*!
 * Wrapper function which will try to open the correct log destination
 */
void glexec_open_log(void)  {
    if (global_log_to_file) { /* We want log to file */
	/* Make sure that we change to syslog for logging-only mode as soon as
	 * possible, to prevent lcas/lcmaps opening their logs.*/
	if (!glexec_is_switching())	{
	    /* Logging only mode: CANNOT log to file */
	    glexec_open_syslog();
	    glexec_log(GLEXEC_LOG_ERROR,
			"Error: Cannot log to file in logging-only mode, " \
			"logging to syslog from now on.\n");
	    return;
	}
	/* Now try opening the logfile */
	glexec_open_log_file();
	/* It might fail (e.g. no longer effective root), then keep syslog */
	if (global_log_file==NULL)  {
	    glexec_open_syslog();
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error: Cannot open logfile, logging to syslog from now on.\n");
	}
    } else /* Log to syslog */
	glexec_open_syslog();
}

/*!
 *  Close the log file and syslog if it had been opened.
 */
void glexec_close_log(void) {
    /* Make sure to flush the signal log first, even is the loglevel below
     * threshold */
    if (global_log_idx)
	glexec_flush_signallog();

    /* syslog */
    if (global_syslog_opened) {
        closelog();
        global_syslog_opened = 0;
    }

    /* file */
    if (global_log_file) {
	/* if it fails, we can (hopefully) still log... */
        if (fclose(global_log_file))
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error while closing glexec log file: %s\n", strerror(errno));
        global_log_file = NULL;
    } 
}

 
/*!
 * Glexec log function, automatically opens the log when needed.
 * When logging to file is specified but fails, it will change to syslog.
 * \param loglevel  The log level of this message.
 * \param fmt       The log message itself
 */
void glexec_log(glexec_log_level_t loglevel, const char* fmt, ...) {
    char body[GLEXEC_MAX_LOG_LINE];
    int i, body_length, prefix_too_long=0, log_failed=0, syslog_level;
    sigset_t old_sigset;
    va_list ap;

    /* Make sure to flush the signal log first, even is the loglevel below
     * threshold */
    if (global_log_idx)
	glexec_flush_signallog();

    /* Check the log level of the message. Test the loglevel is not higher than
     * the predefined cutoff. Also make sure that we do not log any null
     * messages.
     */
    if ( loglevel > global_log_level || fmt == NULL || *fmt == '\0' )	{
	/* Even if we don't log here, still try to set up the prefix. If it
	 * fails, we ignore it here. */
	if (global_log_to_file && global_log_prefix[0]=='\0')
	    glexec_setup_log_prefix();
	return;
    }

    /* Block signals for duration of the function, don't block SEGV etc */
    glexec_block_signals(&old_sigset);
    
    /* Real message: Put in buffer for later use */
    va_start(ap, fmt);
    body_length = vsnprintf(body, GLEXEC_MAX_LOG_LINE, fmt, ap);
    if (body_length<0)	{ /* Naive protection against error */
	body[0]='\n';
	body_length=1;
    }
    if ((unsigned)body_length>=GLEXEC_MAX_LOG_LINE) {
	/* Logline too long: truncate */
	sprintf(&(body[GLEXEC_MAX_LOG_LINE-5]),"...\n");
	body_length=GLEXEC_MAX_LOG_LINE-1;
    }
    va_end(ap);

    /* Replace '?' for any non-printable character */
    for (i=0; i<body_length-1; i++)   {
	if (!isprint(body[i]))
	    body[i]='?';
    }
    /* Force a newline at the end. */
    body[body_length-1]='\n';

    /* In case we are logging to a file, prefix the message by a time stamp.
     * The syslog facility does this automatically and hence we should not
     * prefix it in that case. */
    if (global_log_to_file) { /* Log to file */
	/* If prefix doesn't fit: switch to syslog. Print warning afterwards */
	if (glexec_setup_log_prefix())
	    prefix_too_long=1;
	else	{
	    /* Log the actual message: if it fails switch to syslog */
	    if (fprintf(global_log_file, "%s%s", global_log_prefix, body)<0)
		log_failed=1;
	    else    {
		/* Succesfully logged to file */
		fflush(global_log_file);
		/* revert proc mask */
		glexec_sigprocmask(&old_sigset);
		return;
	    }
	}
	/* Do the switch to syslog and log the reason */
	fclose(global_log_file); global_log_file=NULL; global_log_fd=-1;
	glexec_open_syslog(); /* This also runs glexec_set_log_to_syslog() */
	/* Log reason for changing */
	if (prefix_too_long)
	    syslog(LOG_ERR,
		"Error: Logging to file failed (prefixlength=%lu is too long), "
		"logging to syslog from now on.\n",
		(unsigned long)global_log_prefix_length);
	else if (log_failed)
	    syslog(LOG_ERR,
		"Error: Logging to file failed, "
		"logging to syslog from now on.\n");
    } else
	/* Make sure syslog is opened, we might be logging before it's opened*/
	glexec_open_syslog();

    /* Log the actual message, note: when glexec_ */
    syslog_level= global_diff_syslog_levels ? global_syslog_level[loglevel]
					    : GLEXEC_SYSLOG_DEF_PRIORITY;
    syslog(syslog_level, "%s", body);
    
    /* revert proc mask */
    glexec_sigprocmask(&old_sigset);
}

/*!
 * \param loglevel  The log level of this message.
 * \param fmt       The log message itself
 */
void glexec_signal_log(glexec_log_level_t loglevel, const char* fmt, ...) {
    int i, length; 
    sigset_t old_sigset;
    va_list ap;
    char *body;

    /* Check the log level of the message. Test the loglevel is not higher than
     * the predefined cutoff. Also make sure that we do not log any null
     * messages.
     */
    if ( loglevel > global_log_level || fmt == NULL || *fmt == '\0' )
	return;

    /* Block ALL signals for duration of the function as we might be called from
     * a e.g segv handler */
    glexec_block_allsignals(&old_sigset);
  
    /* Create logline and store */
    if (global_log_idx<GLEXEC_MAX_BUFFER_LINES)  {
	/* Store loglevel */
	global_log_buffer[global_log_idx].loglevel=loglevel;

	/* Real message: Put in buffer for later use */
	body=global_log_buffer[global_log_idx].body;
	va_start(ap, fmt);
	length = vsnprintf(body, GLEXEC_MAX_LOG_LINE, fmt, ap);
	if (length<0)	{ /* Naive protection against error */
	    body[0]='\n';
	    body[1]='\0';
	    length=1;
	}
	if ((unsigned)length>=GLEXEC_MAX_LOG_LINE) {
	    /* Logline too long: truncate */
	    body[GLEXEC_MAX_LOG_LINE-5]='.';
	    body[GLEXEC_MAX_LOG_LINE-4]='.';
	    body[GLEXEC_MAX_LOG_LINE-3]='.';
	    body[GLEXEC_MAX_LOG_LINE-2]='\n';
	    body[GLEXEC_MAX_LOG_LINE-1]='\0';
	    length=GLEXEC_MAX_LOG_LINE-1;
	}
	va_end(ap);

	/* Replace '?' for any non-printable character */
	for (i=0; i<length-1; i++)   {
	    if (!isprint(body[i]))
		body[i]='?';
	}
	/* Force a newline at the end. */
	body[length-1]='\n';
    }
    global_log_idx++;

    /* revert proc mask */
    glexec_sigprocmask(&old_sigset);
}

/*!
 * Flushes the log buffer from glexec_signal_log(), should NOT be called from a
 * signal handler, \see glexec_emergency_flush_signallog()
 */
void glexec_flush_signallog(void) {
    int i, prefix_too_long=0, log_failed=0;
    sigset_t unblockset, normalblockset;

    /* if the queue is empty return */
    if (global_log_idx==0) return;

    /* Block signals for duration of the function */
    glexec_block_signals(&unblockset);
    
    /* In case we are logging to a file, prefix the message by a time stamp.
     * The syslog facility does this automatically and hence we should not
     * prefix it in that case. */
    if (global_log_to_file) { /* Log to file */
	if (glexec_setup_log_prefix())    {
	    prefix_too_long=1;
	    goto glexec_flush_signallog_switchtosyslog;
	}

	/* Now block all signals for actual flushing to prevent corrupting the
	 * global_log_idx */
	glexec_block_allsignals(&normalblockset);
    
	/* Flush buffer: if it fails switch to syslog */
	for (i=0; i<global_log_idx && i<GLEXEC_MAX_BUFFER_LINES; i++) {
	    if (fprintf(global_log_file, "%s%s",
			global_log_prefix, global_log_buffer[i].body)<0)	{
		log_failed=1;
		/* reset signal mask to normal blocked set */
		glexec_sigprocmask(&normalblockset);
		goto glexec_flush_signallog_switchtosyslog;
	    }
	}
	/* Also log warning when we ran out of buffered space */
	if (global_log_idx>GLEXEC_MAX_BUFFER_LINES)  {
	    if (fprintf(global_log_file,
		    "%sWarning: logbuffer too small, "
		    "%d logged messages have been dropped.\n",
		    global_log_prefix,
		    global_log_idx-GLEXEC_MAX_BUFFER_LINES)<0)   {
		log_failed=1;
		/* reset signal mask to normal blocked set */
		glexec_sigprocmask(&normalblockset);
		goto glexec_flush_signallog_switchtosyslog;
	    }
	}
	/* Succesfully logged to file: reset counter */
	global_log_idx=0;
	/* reset signal mask to normal blocked set */
	glexec_sigprocmask(&normalblockset);
	/* flush file */
	fflush(global_log_file);
	/* reset signal mask to starting set */
	glexec_sigprocmask(&unblockset);
	return;

glexec_flush_signallog_switchtosyslog:
	/* Logging to file should switch to syslog */
	/* Do the switch to syslog and log the reason */
	fclose(global_log_file); global_log_file=NULL; global_log_fd=-1;
	glexec_open_syslog(); /* This also runs glexec_set_log_to_syslog() */

	/* Log reason for changing */
	if (prefix_too_long)
	    syslog(global_diff_syslog_levels ? global_syslog_level[GLEXEC_LOG_ERROR] :
					       GLEXEC_SYSLOG_DEF_PRIORITY,
		"Error: Logging to file failed (prefixlength=%lu is too long), "
		"logging to syslog from now on.\n",
		(unsigned long)global_log_prefix_length);
	else if (log_failed)
	    syslog(global_diff_syslog_levels ? global_syslog_level[GLEXEC_LOG_ERROR] :
					       GLEXEC_SYSLOG_DEF_PRIORITY,
		"Error: Logging to file failed, "
		"logging to syslog from now on.\n");
    } else
	/* Make sure syslog is opened, we might be logging before it's opened*/
	glexec_open_syslog();

    /* Now block all signals for actual flushing */
    glexec_block_allsignals(&normalblockset);
    
    /* Log the actual messages */
    if (global_diff_syslog_levels)  {
	for (i=0; i<global_log_idx && i<GLEXEC_MAX_BUFFER_LINES; i++)
	    syslog(global_syslog_level[global_log_buffer[i].loglevel],
		   "%s", global_log_buffer[i].body);
    } else  {
	for (i=0; i<global_log_idx && i<GLEXEC_MAX_BUFFER_LINES; i++)
	    syslog(GLEXEC_SYSLOG_DEF_PRIORITY,"%s", global_log_buffer[i].body);
    }

    /* Do we miss any messages? */
    if (global_log_idx>GLEXEC_MAX_BUFFER_LINES)
	syslog(LOG_ERR,"Error: logbuffer too small, "
		       "%d logged messages have been dropped.\n",
		global_log_idx-GLEXEC_MAX_BUFFER_LINES);

    /* Reset counter */
    global_log_idx=0;
    
    /* revert proc mask */
    glexec_sigprocmask(&unblockset);
}

/*!
 * Can be called when we do an exit from a signal handler. As we can no longer
 * log afterwards we need to flush the buffer. This is only safe when we know
 * we're not inside syslog. It also will fsync and close the filedescriptor.
 */
void glexec_emergency_flush_signallog(void) {
    int i;
    size_t len;
    const char *bufline="Warning: logbuffer too small, some messages have been dropped.\n";
    const char *prefix;

    prefix=(global_log_prefix[0]=='\0' ? "glexec[unknown]: "
				       : global_log_prefix);

    /* if the queue is empty or we're logging to syslog, return */
    if (global_log_idx==0 || !global_log_to_file || global_log_fd==-1)
	return;

    /* Flush buffer: if it fails we give up */
    for (i=0; i<global_log_idx && i<GLEXEC_MAX_BUFFER_LINES; i++) {
	/* find buffer length */
	for (len=0; global_log_buffer[i].body[len]!='\0'; len++);
	/* dump buffer */
	if (write(global_log_fd,prefix,global_log_prefix_length)<0 ||
	    write(global_log_fd,global_log_buffer[i].body,len)<0)
	    return; /* write failed */
    }

    /* Also log warning when we ran out of buffered space */
    if (global_log_idx>GLEXEC_MAX_BUFFER_LINES)  {
	for (len=0; bufline[len]!='\0'; len++);
	if (write(global_log_fd,prefix,global_log_prefix_length)<0 ||
	    write(global_log_fd,bufline,len)<0)
	    return; /* write failed */
    }

    /* Call fsync explicitly and then close the file descriptor */
    if (fsync(global_log_fd) && close(global_log_fd))
	global_log_fd=-1;

    /* Succesfully logged */
    global_log_idx=0;	/* Reset counter */
    return;
    
}

/*************************************************************************/
/* INTERNAL FUNCTIONS							 */
/*************************************************************************/

/*!
 * Open the syslog facility for gLExec.
 */
static void glexec_open_syslog(void)	{
    /* Open syslog in case it wasn't open */
    if (!global_syslog_opened)	{
	openlog(GLEXEC_SYSLOG_IDENT,GLEXEC_SYSLOG_OPTS,global_syslog_facility);
	global_syslog_opened = 1;
    }
    /* Force setting logging to syslog. */
    glexec_set_log_to_syslog();
}

/*!
 *  Open the glexec logfile. During opening logging destination is forced to
 *  syslog, in order to record errors. Upon success global_log_file will be
 *  non-NULL, upon failure, global_log_file is still NULL, and in case it is not
 *  due to no-longer-effective-root, glexec will exit. 
 */
static void glexec_open_log_file(void)	{
    const mode_t file_mask=GLEXEC_LOG_FILEMASK; /* File permissions */
    const mode_t dir_mask=GLEXEC_LOG_DIRMASK; /* Dir permissions */
    const uid_t glexec_loguid=0;
    const char *log_file,*group;
    int rc;
    struct group *gr;
    uid_t oldeuid=geteuid();
    gid_t oldegid=getegid();

    /* Temporarily set logging to default syslog */
    glexec_set_log_to_syslog();

    /* Check if there is an alternative defined for the default log file
     * for glexec. */
    log_file=glexec_conf_value_default("glexec:log_file", GLEXEC_LOG_FILE);

    /* Check if we specified a group for creation dir/file, otherwise it will be
     * group root */
    if ( (group=glexec_conf_value("glexec:log_file_group")) )  {
	if ( (gr = glexec_getgrnam(group))!=NULL )	{
	    global_log_gid=gr->gr_gid;
	    glexec_log(GLEXEC_LOG_DEBUG,
		"Logfile will be created (if needed) using gid=%d(%s).\n",
		(int)global_log_gid,group);
	} else /* group does not exist */
	    glexec_log(GLEXEC_LOG_WARNING,
		"Specified glexec:log_file_group '%s' not found on system, " \
		"using GID %d instead\n", group, (int)global_log_gid);
    } else  /* Default group */
	glexec_log(GLEXEC_LOG_DEBUG,
	    "Logfile will be created (if needed) using GID %d.\n",
	    (int)global_log_gid);

    /* Temporarily set back effective root uid (if it isn't already) */
    if (oldeuid!=glexec_loguid && seteuid(glexec_loguid))   {
	glexec_log(GLEXEC_LOG_WARNING,
	    "Setting effective uid to %d failed, reverting to syslog!\n",
	    (int)glexec_loguid);
	return;
    }
    /* Set the effective gid, so that file and dir will be created with the
     * right ownership */
    if (setegid(global_log_gid))
	glexec_log(GLEXEC_LOG_WARNING,
	    "Setting effective gid to %d failed.\n", (int)global_log_gid);
    /* Now try to open the logfile with file permission file_mask and directory
     * permissions dir_mask. Force it to be trusted: last arg == 1 */
    rc=cgul_open_logfile(log_file, file_mask, dir_mask, &global_log_file, 1);
    /* reset effective gid */
    if (setegid(oldegid))
	glexec_log(GLEXEC_LOG_WARNING,
	    "Setting effective gid to %d failed.\n", (int)oldegid);
    /* reset effective uid */
    if (seteuid(oldeuid))
	glexec_log(GLEXEC_LOG_WARNING,
	    "Setting effective uid to %d failed.\n", (int)oldeuid);
    /* Parse return code of cgul_open_logfile */
    switch (rc)	{
	case 0: break;
	case -1: 
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error: error creating dir for logfile '%s'.\n", log_file);
	    glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	case -2:
	    glexec_log(GLEXEC_LOG_ERROR,
		"Error: I/O error during opening logfile: %s.\n",
		strerror(errno));
	    glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	case -3:
	    glexec_exit(GLEXEC_LOG_FILE_WRONG_PERMISSION, 
		        GLEXEC_NO_CHILD_EXIT_CODE);
	case -4:
	    glexec_exit(GLEXEC_OUT_OF_MEMORY, GLEXEC_NO_CHILD_EXIT_CODE);
	case -5:
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error: logfile '%s' is not an absolute path.\n",log_file);
	    glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	case -6:
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error: logfile is no longer the one that was opened!.\n");
	    glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	case -7:
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error: logfile '%s' is not a regular file.\n",log_file);
	    glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
	case -8:
	default:
	    glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Set filedescriptor for logfile */
    if ( (global_log_fd=fileno(global_log_file)) == -1)	{
	glexec_log(GLEXEC_LOG_ERROR,
		   "Error: Cannot get file descriptor for opened logfile.\n");
	glexec_exit(GLEXEC_LOG_FILE_OPEN_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Reset logging to file */
    glexec_set_log_to_file();
    if (global_syslog_opened) /* Close syslog if it was opened */
	closelog();
}

/*!
 * Sets up the (global) logging prefix as used in logging to file
 * \return 0 upon success, -1 on error (buffer too small)
 */
static int glexec_setup_log_prefix(void)	{
    time_t timevar;
    struct tm *lt;
    pid_t glexec_pid;
    int length;

    /* Setup prefix */
    if (time(&timevar)==-1 || (lt = gmtime(&timevar))==NULL)	{
	global_log_prefix[0]='\0';
	global_log_prefix_length=0;
	return -1;
    }

    /* Initialize glexec_pid: do this each time to make diff between child and
     * parent */
    glexec_pid=getpid();
   
    /* Fill buffer */
    length = snprintf(global_log_prefix, (size_t)GLEXEC_LOGPREFIX_LENGTH,
	    "%s[%d] %d%.2d%.2dT%.2d%.2d%.2dZ: ",
	    GLEXEC_PACKAGE_NAME, (int)glexec_pid,
	    lt->tm_year + 1900, lt->tm_mon + 1, lt->tm_mday,
	    lt->tm_hour, lt->tm_min, lt->tm_sec);

    if (length<0 || length>=GLEXEC_LOGPREFIX_LENGTH) { /* too long or error */
	global_log_prefix[0]='\0';
	global_log_prefix_length=0;
	return -1;
    }
    
    global_log_prefix_length=(size_t)length;
    return 0;

}


