/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 * \file main.c
 */

#ifdef __linux__
#ifdef DEBUG
    #include <sys/prctl.h>
#endif /* DEBUG */
#endif /* __linux__ */

#include "glexec.h"
#include "main_util.h"
#include "glexec_util.h"
#include "glexec_environ.h"
#include "glexec_log.h"
#include "glexec_exit.h"
#include "glexec_ipc.h"
#include "glexec_epilog.h"

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>

/*************************************************************************/
/* START OF main()                                                       */
/*************************************************************************/
int main(int argc, char *argv[]) {
    int glexec_print_defines_mode=0;	/* whether -V is specified */
    char *pem_string= NULL;	/* GLEXEC_CLIENT_CERT */
    char *source_proxy= NULL;	/* GLEXEC_SOURCE_PROXY */
    char *payload_proxy= NULL;	/* name of payload proxy file or NULL */
    int t_argc;			/* argc for command to be executed */
    int exec_idx;		/* index of argv pointing to payload */
    char **t_argv;		/* argv for command to be executed */
    const char *target_cwd=NULL;/* current working directory for target */
    char *fullcmd=NULL;		/* absolute command to be executed */
    char *realcmd=NULL;		/* canonical absolute command to be executed */
    pid_t pid;			/* child process pid */
    int childrc;		/* exit code for child and epilog */
    int linger_mode;		/* whether to fork or not for payload */
    int close_fds;		/* whether lingering gLExec closes fds */
    glexec_account_t caller,ecaller,target; /* real-, effect. caller, target */

#ifdef __linux__
#ifdef DEBUG
    /* echo "1" > /proc/sys/kernel/core_setuid_ok */
    /* echo "1" > /proc/sys/kernel/suid_dumpable */
    prctl(PR_SET_DUMPABLE, 1, 42, 42, 42);
    sleep(5);
#endif /* DEBUG */
#endif /* __linux__ */

    /* Check stdin/stdout/stderr are open and have right type. Do even before
     * cleaning the environment as any I/O (including logging) can fail
     * otherwise */
    glexec_check_stdinouterr();
    
#ifdef YES_I_AM_REALLY_SURE_TO_DISABLE_THIS_SECURITY_MEASURE_IN_GLEXEC
    fprintf (stderr, "[gLExec]: Warning: security measures disabled at build time! For testing purposes only, don't use this build in production.\n"); 
#endif /* YES_I_AM_REALLY_SURE_TO_DISABLE_THIS_SECURITY_MEASURE_IN_GLEXEC */

    /* Clean the internal environment, this will also chdir to / */
    glexec_clean_environment();

    /* Setup signal handlers (the extra signal handlers will only be set up once
     * the config options are available, in the get_linger_mode function) */
    glexec_install_sig_handlers();

    /* change to safe directory and setup umask: user has no more influence */
    glexec_set_runtime_cwd_umask();

    /* Check number of arguments, simple modes are handled here. Only -V mode
     * has to wait till after glexec_get_caller() is done. */
    glexec_get_runmode(argc, argv, &glexec_print_defines_mode, &exec_idx);
    
    /* Set command and argument list for target process */
    t_argc=argc-exec_idx;
    t_argv=&(argv[exec_idx]);

    /* initialize switching variable */
    glexec_set_switchingmode();

    /* Read in configuration */
    glexec_read_configuration();

    /* Setup logging */
    glexec_setup_log();

    /* Fill caller structs */
    glexec_get_caller(&caller,&ecaller);

    /* Now we are ready for the 'print_defines' / -V mode if requested, we need
     * to know if we are allowed */
    if (glexec_print_defines_mode)
	glexec_print_defines(stdout, &caller);

    /* Whether we're in linger mode, this also sets the signal options which
     * depend on the linger mode, including whether to setup handlers which are
     * used by debuggers, profilers etc. */
    linger_mode=glexec_get_linger_mode(&close_fds);

    /* Check to see if the user running this program is allowed to, based on
     * either whitelist or group membership */
    glexec_check_caller_allowed(&caller);

    /* Setup the gLExec internal environment variables */
    glexec_setup_glexec_env(caller.uname);

    /* Read GLEXEC_CLIENT_CERT and GLEXEC_SOURCE_PROXY or its default */
    glexec_get_proxies(&pem_string, &source_proxy);
    
    /* Get target account info: will exit upon error.
     * When LCMAPS should do the switch, this is where it's going to happen */
    glexec_auth((const char * const *)t_argv, pem_string, &target);

    /* Re-setup signal handlers since they might have changed in e.g. LCMAPS */
    glexec_install_sig_handlers();
    glexec_install_extra_sig_handlers();

    /* Check wether target user may run...*/
    glexec_check_target_allowed(&target);

    /* Stuff to be done when glexec is (configured to be) doing the switch */
    if (glexec_posix_enf_by_glexec()) {
	/* First create a backlog item if needed */
	glexec_create_back_log_item(&caller, &target);

	/* If we are not in logging only mode, do the switch, this sets up the
	 * groups for payload, and changes effective uid to payload. Real and
	 * saved stay original ones and will be enforced in
	 * glexec_fork_and_enforce. */
	if (glexec_is_switching())
	    glexec_userswitch(&target);
    } else
	/* Check we are no longer root! */
	glexec_check_no_longer_root(&caller, &target);

    /* Try to chdir to the proper homedir: must be (effective) target user.
     * The new cwd is returnes in target_cwd. */
    target_cwd=glexec_setup_target_cwd(&caller,&target);

    /* Check prog and directory: must be (effective) target user. */
    glexec_check_exe_and_dir(t_argv[0],
			     target_cwd, &fullcmd, &realcmd, &caller, &target);

    /* Write proxy: must be (effective) target user, name will be returned or be
     * NULL when none. */
    payload_proxy=glexec_write_source_proxy(&caller,&target,&source_proxy);

    /* Revert to effective caller when glexec is doing a userswitch (function
     * tests itself for switching mode): in
     * switching mode, we're then uid/euid caller, but groups target. */
    if (glexec_posix_enf_by_glexec())
	glexec_set_effective_uid(caller.uid);

    /* Setup remaining target environment, starting with a "clean" one. */
    glexec_setup_target_account(&caller,&target,payload_proxy);

    /* Figure out whether we have to set up an epilogue script, only do that
     * now, since we have the target proxy variables. */
    if (linger_mode)
	glexec_setup_epilog(&caller,&target,t_argc, t_argv);
    
    /* Inform the sysadmin what is being executed */
    glexec_log_argument_list(t_argc, t_argv,
			     target_cwd, fullcmd, realcmd, &caller, &target);

    /* Fork off a child when needed, otherwise, 0 will be both, also starts
     * timer and enforces identity after fork. It closes all inherited
     * filedescriptors where needed */
    pid=glexec_fork_and_enforce(linger_mode,close_fds,&caller, &target);
    if (pid==0)	{ /* I.e. child */
    	/* Execute with mapped privileges (or non-mapped privs) and from the
	 * child process the following command with specified argv[].
	 * NOTE: We always use the realcmd so that even in case of a symlink
	 * attack we still now which executable has been run. */
        execve(realcmd, t_argv, glexec_get_target_env() );
	
	/* Either NEVER come here or execve() was unsuccessful */
	glexec_execve_exit(target_cwd,fullcmd,realcmd);
    }

    /* Remainder code is the parent! */
    /* Wait for the child to exit */
    childrc=glexec_waitchild();

    /* Run epilog (when needed, will exit itself when failing */
    glexec_run_epilog();

    /* All done: exit */
    glexec_exit(GLEXEC_SUCCESS, childrc);

    /* This is to get rid of the warning: "control reaches end of non-void
     * function", which is true, but the compiler doesn't know glexec_exit()
     * never returns. */
    return 0;
}
