/**U
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 * main_util.c -- helper functions for gLExec main()
 * \file main_util.c
 */

#include "main_util.h"
#include "fileutil/fileutil.h"
#include "realpath/realpath.h"

#include "glexec_log.h"
#include "glexec_util.h"
#include "glexec_account.h"
#include "glexec_conf.h"
#include "glexec_environ.h"
#include "glexec_lcas.h"
#include "glexec_lcmaps.h"
#include "glexec_ipc.h"
#include "glexec_closefrom.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/param.h>
#include <ctype.h>

#include <pwd.h>
#include <grp.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
/* For UINT_MAX */
#include <limits.h> 
/* For strcasecmp */
#include <strings.h>

/************************************************************************/
/* DEFINES	                                                        */
/************************************************************************/

#define _JOBREPOID_LEN 50

/* Use following as a custom maximum number of groups, in case neither
 * sysconf(_SC_NGROUPS_MAX) nor NGROUPS_MAX gives something meaningful */
#ifndef NGROUPS_MAX
#define CUSTOM_NGROUPS_MAX  32
#endif

/************************************************************************/
/* INTERNAL FUNCTION PROTOTYPES                                         */
/************************************************************************/

/*!
 *  Check if the real user name is in the whitelist specified in the
 *  configuration file. A whitelist entry can be of the form:
 *   * - all users
 *   .<pool>
 *   <pool>* - where * is purely numerical
 *   <username> - no stars allowed
 *  \param user  user name.
 *  \return 1 in case the user is allowed to execute glexec, 0 otherwise.
 */
static int glexec_uname_allowed(const char *user);

/*!
 *   Check if the user executing gLExec has to present a proxy which
 *   contains a private key or that the users is allowed not to present
 *   the private key. Only when it is determined that the private key
 *   maybe omitted from the proxy, false is returned.
 *  \param user  Symbolic user name.
 *  \returns 1 in case the user needs to present a proxy *with* a private key,
 *           0 when the user is allowed to present a proxy without.
 */
static int glexec_user_needs_private_key(const char *user);

/*!
 * Determine and update values of the TERM and KILL delay settings and whether
 * to use setpgid() in child processes.
 * \param whether we run in linger_mode
 */
static void glexec_set_signaloptions(int linger_mode);

/*!
 * Determine and set linger_as_payload flag
 * \param int value to set:
 *        <0 undefined
 *        =0 different
 *        1 equal, switching mode (unsafe)
 *        2 equal, no (longer) setuid root possible
 * \return actual value found
 */
static int glexec_get_linger_as_payload(void);

/*!
 * Determines and returns specified locking mechanism for reading input proxies,
 * or its default value.
 * \return locking mechanism
 */
static GLEXEC_LOCK_MECH_T glexec_get_input_lock_mechanism(void);

/*!
 * Determines and returns specified locking mechanism for writing target proxy,
 * or its default value.
 * \return locking mechanism
 */
static GLEXEC_LOCK_MECH_T glexec_get_target_lock_mechanism(void);

/*!
 * Determines whether we want the pedantic and no-symlink exec checks and then
 * carries them out. Exits on failure.
 * \param target_cwd current working directory
 * \param fullcmd absolute path executable
 * \param realcmd canonical absolute path executable
 * \param caller
 * \param target
 * \param exe_info stat information on executable
 */
static void glexec_pedantic_checks(const char *target_cwd,
			    char *fullcmd, char *realcmd,
			    glexec_account_t *caller, glexec_account_t *target,
			    struct stat *exe_info);

/************************************************************************/
/* FUNCTIONS								*/
/************************************************************************/

/*!
 * Ensures that stderr, stdout and stdin are open filedescriptors and are resp.
 * writable and readable. If not, it (re)opens them onto /dev/null. In case of
 * error gLExec exits with GLEXEC_EXIT_INTERNAL_ERROR as we cannot yet log at
 * this stage.
 */
void glexec_check_stdinouterr(void)  {
    int flags,fd;
    struct stat buf;

    /* STDIN */
    flags=fcntl(STDIN_FILENO, F_GETFL);
    if (flags<0 && errno!=EBADF)
	/* fcntl failed in unrecoverable way */
	exit(GLEXEC_EXIT_INTERNAL_ERROR);

    /* If stdin non-existing, write-only or a dir: (re)open */
    if (flags<0 || (flags & O_ACCMODE) == O_WRONLY ||
	fstat(STDIN_FILENO, &buf)<0 || S_ISDIR(buf.st_mode) ) {
	/* Open /dev/null and use dup2 to make it stdin.
	 * close new fd only when different from the original */
	if ( (fd=open("/dev/null", O_RDONLY))==-1 ||
	     (fd!=STDIN_FILENO &&
	      (dup2(fd, STDIN_FILENO)==-1 || close(fd)==-1) ) )	{
	    /* open, dup2 or close (when fd!=0) failed */
	    exit(GLEXEC_EXIT_INTERNAL_ERROR);
	}
    }

    /* STDOUT */
    flags=fcntl(STDOUT_FILENO, F_GETFL);
    if (flags<0 && errno!=EBADF)
	/* fcntl failed in unrecoverable way */
	exit(GLEXEC_EXIT_INTERNAL_ERROR);

    /* If stdout non-existing, read-only or a dir: (re)open */
    if (flags<0 || (flags & O_ACCMODE) == O_RDONLY ||
	fstat(STDOUT_FILENO, &buf)<0 || S_ISDIR(buf.st_mode) ) {
	/* Open /dev/null and use dup2 to make it stdout.
	 * close new fd only when different from the original */
	if ( (fd=open("/dev/null", O_WRONLY | O_APPEND))==-1 ||
	     (fd!=STDOUT_FILENO &&
	      (dup2(fd, STDOUT_FILENO)==-1 || close(fd)==-1) ) )    {
	    /* open, dup2 or close (when fd!=1) failed */
	    exit(GLEXEC_EXIT_INTERNAL_ERROR);
	}
    }

    /* STDERR */
    flags=fcntl(STDERR_FILENO, F_GETFL);
    if (flags<0 && errno!=EBADF)
	/* fcntl failed in unrecoverable way */
	exit(GLEXEC_EXIT_INTERNAL_ERROR);

    /* If stderr non-existing or read-only: (re)open */
    if (flags<0 || (flags & O_ACCMODE) == O_RDONLY ||
	fstat(STDERR_FILENO, &buf)<0 || S_ISDIR(buf.st_mode) ) {
	/* Open /dev/null and use dup2 to make it stdout.
	 * close new fd only when different from the original */
	if ( (fd=open("/dev/null", O_WRONLY | O_APPEND))==-1 ||
             (fd!=STDERR_FILENO && 
	      (dup2(fd, STDERR_FILENO)==-1 || close(fd)==-1) ) ) {
	    /* open, dup2 or close (when fd!=2) failed */
	    exit(GLEXEC_EXIT_INTERNAL_ERROR);
	}
    }

    /* All done */
    return;
}

/*!
 * Parses the commandline arguments for simple running modes: i.e. no arguments
 * (=error), or a single option argument (-h, -v, -V etc.). In case of -V the
 * parameter print_defines is set. For other simple run modes, the corresponding
 * function (usage, version etc). is run and glexec will exit.
 * \param argc, argv same as for main()
 * \param print_defines set when option -V/--defines is given
 * \param exec_idx index of executable (payload) argument).
 */
void glexec_get_runmode(int argc, char *argv[],
		       int *print_defines, int *exec_idx)	{
    int i;

    /* Handle 0 or 1 arguments separately */
    if (argc<1)    {
	glexec_log(GLEXEC_LOG_ERROR,
		"gLExec is called without even a valid argv[0].\n");
        glexec_exit(GLEXEC_WRONG_SET_OF_ARGS, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Initialize print_defines */
    *print_defines=0;

    /* Loop over arguments */
    for (i=1; i<argc; i++)  {
	if (argv[i][0]!='-')	{
	    /* We reached target executable: return index of executable */
	    *exec_idx=i;
	    return;
	}

	/* Help */
	if (strcmp(argv[i], "-h")==0 || strcmp(argv[i], "--help")==0)   {
	    glexec_usage(argv[0]); 
	    glexec_exit(GLEXEC_SUCCESS, GLEXEC_NO_CHILD_EXIT_CODE);
	}

	/* Version */
	if (strcmp(argv[i], "-v")==0 || strcmp(argv[i], "--version")==0) {
	    printf("gLExec version: %s\n", GLEXEC_VERSION);
	    glexec_exit(GLEXEC_SUCCESS, GLEXEC_NO_CHILD_EXIT_CODE);
	}

	/* Print defines: set option and return. */
	if (strcmp(argv[i], "-V")==0 || strcmp(argv[i], "--defines")==0)    {
	    *print_defines=1;
	    return;
	}

	/* Payload in background? */
	if (strcmp(argv[i], "-b")==0 || strcmp(argv[i], "--background")==0) {
	    glexec_log(GLEXEC_LOG_INFO,
		    "User specified -b option to gLExec.\n");
	    glexec_set_payload_should_run_in_background(1);
	    continue;
	}
    }

    /* No executable found */
    /* Only argv[0] */
    glexec_log(GLEXEC_LOG_ERROR,
	    "No executable has been specified on the cmdline.\n");
    glexec_exit(GLEXEC_WRONG_SET_OF_ARGS, GLEXEC_NO_CHILD_EXIT_CODE);
}


/*!
 * Prints usage text on stdout. exec_name is the executable name.
 */
void glexec_usage(char *exec_name)	{
    printf("Executes a command as another user based on grid credentials.\n");
    printf("Usage: %s [opts] <cmd> [args]\n", exec_name);
    printf("  [opts]  options to gLExec itself, see below\n");
    printf("  <cmd>   command to be executed\n");
    printf("  [args]  arguments of <cmd>\n");
    printf("\nOptions:\n");
    printf("  -h, --help         print this helptext\n");
    printf("  -v, --version      print gLExec version\n");
    printf("  -V, --defines      print gLExec built-in defaults (only root)\n");
    printf("  -b, --background   run payload in background\n");
    printf("\nReturn values:\n");
    printf("  %d -   User/Client error   :\n", GLEXEC_EXIT_USER_ERROR);
    printf("          An error that the user is able to address.\n");
    printf("  %d -   gLExec/System error :\n", GLEXEC_EXIT_INTERNAL_ERROR);
    printf("          An error that the administrator has to address.\n");
    printf("  %d -   Authorization error :\n",GLEXEC_EXIT_AUTHZ_ERROR);
    printf("          The user calling gLExec or the user credentials\n");
    printf("          presented to gLExec failed authorization.\n");
    printf("  %d -   Return code overlap :\n",
	    GLEXEC_EXIT_CHILD_PROCESS_RC_OVERLAP);
    printf("          The child process (payload process) returned\n");
    printf("          a value of %d, %d, %d or %d.\n", 
	    GLEXEC_EXIT_USER_ERROR,
	    GLEXEC_EXIT_INTERNAL_ERROR,
	    GLEXEC_EXIT_AUTHZ_ERROR,
	    GLEXEC_EXIT_CHILD_PROCESS_RC_OVERLAP);
    printf("\n  %d   - execve() call failed due to permissions on the\n",
	    GLEXEC_EXECVE_FAILED_EXIT_CODE);
    printf("          executable that was tried to be executed.\n");
    printf("  %d+n - The executed process (child process) finished\n" \
	   "          due to signal number n.\n",GLEXEC_EXECVE_SIGEXIT_BASE);
    printf("\nFor more information see the manpages ");
    printf("glexec(1) and glexec.conf(5).\n");
}

/*!
 * Prints build-in defaults on filedescriptor out: -V mode.
 * Only root and special user are allowed, which are compared to the caller.
 * \return GLEXEC_SUCCESS upon success, or GLEXEC_USER_TRIED_V when user is not
 * allowed.
 */
void glexec_print_defines(FILE *out, glexec_account_t *caller){
    if ( (caller->uid==0) || strcmp(GLEXEC_USER, caller->uname)==0 ) {

#ifdef YES_I_AM_REALLY_SURE_TO_DISABLE_THIS_SECURITY_MEASURE_IN_GLEXEC
	fprintf(stderr, " Warning: this version gLExec is unsafe!\n Disable the pre-build: #define YES_I_AM_REALLY_SURE_TO_DISABLE_THIS_SECURITY_MEASURE_IN_GLEXEC\n");
#endif

#ifdef GLEXEC_PACKAGE_NAME
	fprintf(out," -D GLEXEC_PACKAGE_NAME=\"%s\"\n", GLEXEC_PACKAGE_NAME);
#endif
#ifdef GLEXEC_VERSION
	fprintf(out," -D GLEXEC_VERSION=\"%s\"\n", GLEXEC_VERSION);
#endif
#ifdef GLEXEC_CONFIG_FILE
	fprintf(out," -D GLEXEC_CONFIG_FILE=\"%s\"\n", GLEXEC_CONFIG_FILE);
#endif
#ifdef GLEXEC_CONF_ELEM_MAX
	fprintf(out," -D GLEXEC_CONF_ELEM_MAX=%lu\n", (long unsigned int)GLEXEC_CONF_ELEM_MAX);
#endif
#ifdef GLEXEC_USER
	fprintf(out," -D GLEXEC_USER=\"%s\"\n", GLEXEC_USER);
#endif
#ifdef GLEXEC_CONFIG_GROUP
	fprintf(out," -D GLEXEC_CONFIG_GROUP=\"%s\"\n", GLEXEC_CONFIG_GROUP);
#endif
#ifdef GLEXEC_WHITELIST_GROUP
	fprintf(out," -D GLEXEC_WHITELIST_GROUP=\"%s\"\n",
		GLEXEC_WHITELIST_GROUP);
#endif
#ifdef GLEXEC_UID_MIN
	fprintf(out," -D GLEXEC_UID_MIN=%u\n", GLEXEC_UID_MIN);
#endif
#ifdef GLEXEC_GID_MIN
	fprintf(out," -D GLEXEC_GID_MIN=%u\n", GLEXEC_GID_MIN);
#endif
#ifdef GLEXEC_SYSLOG_IDENT
	fprintf(out," -D GLEXEC_SYSLOG_IDENT=%s\n", GLEXEC_SYSLOG_IDENT);
#endif
#ifdef GLEXEC_SYSLOG_FAC_NAME
	fprintf(out," -D GLEXEC_SYSLOG_FACILITY=%s\n", GLEXEC_SYSLOG_FAC_NAME);
#endif
#ifdef GLEXEC_RUNTIME_UMASK
	fprintf(out," -D GLEXEC_RUNTIME_UMASK=%03o\n", GLEXEC_RUNTIME_UMASK);
#endif
#ifdef GLEXEC_LOG_FILE
	fprintf(out," -D GLEXEC_LOG_FILE=%s\n", GLEXEC_LOG_FILE);
#endif
#ifdef GLEXEC_LOG_DIRMASK
	fprintf(out," -D GLEXEC_LOG_DIRMASK=%03o\n", GLEXEC_LOG_DIRMASK);
#endif
#ifdef GLEXEC_LOG_FILEMASK
	fprintf(out," -D GLEXEC_LOG_FILEMASK=%03o\n", GLEXEC_LOG_FILEMASK);
#endif
#ifdef GLEXEC_TARGET_PATH
	fprintf(out," -D GLEXEC_TARGET_PATH=\"%s\"\n", GLEXEC_TARGET_PATH);
#endif
#ifdef GLEXEC_TARGET_UMASK
	fprintf(out," -D GLEXEC_TARGET_UMASK=%03o\n", GLEXEC_TARGET_UMASK);
#endif
#ifdef GLEXEC_DEFAULT_NAME_TARGET_PROXY
	fprintf(out," -D GLEXEC_DEFAULT_NAME_TARGET_PROXY=%s\n",
		GLEXEC_DEFAULT_NAME_TARGET_PROXY);
#endif
#ifdef GLEXEC_RUNTIME_CWD
	fprintf(out," -D GLEXEC_RUNTIME_CWD=%s\n", GLEXEC_RUNTIME_CWD);
#endif
#ifdef GLEXEC_TERM_GRACE_TIME
	fprintf(out," -D GLEXEC_TERM_GRACE_TIME=%u\n", GLEXEC_TERM_GRACE_TIME);
#endif
#ifdef GLEXEC_KILL_GRACE_TIME
	fprintf(out," -D GLEXEC_KILL_GRACE_TIME=%u\n", GLEXEC_KILL_GRACE_TIME);
#endif
#ifdef GLEXEC_EPILOG_TIMEOUT
	fprintf(out," -D GLEXEC_EPILOG_TIMEOUT=%u\n", GLEXEC_EPILOG_TIMEOUT);
#endif

	/* Defines for LCAS and LCMAPS */
	glexec_lcas_defines(out);
	glexec_lcmaps_defines(out);

	/* Successful exit */
	glexec_exit(GLEXEC_SUCCESS, GLEXEC_NO_CHILD_EXIT_CODE);
    } else
	/* User not allowed to run -V */
	glexec_exit(GLEXEC_USER_TRIED_V, GLEXEC_NO_CHILD_EXIT_CODE);
}

/*!
 * Read the configuration file of glexec.
 *
 * In setuid/switching mode, the configuration file is only read in an
 * unprivileged mode, using the GLEXEC_USER account. If this useraccount does
 * not exist or switching to that account fails an error is returned.
 * Once the configuration file has been read, glexec will resume its previous
 * identity.
 * \return Error code.
 */
GLEXEC_ERROR glexec_read_configuration(void) {
    int rc;
    struct passwd *pw;
    struct group *gr;
    uid_t read_uid;
    gid_t read_gid;
    const char *read_uname;
    const char *read_gname;

    /* Get special user and group, note that they may not belong together */
    /* First user */
    if ( (pw = glexec_getpwnam(GLEXEC_USER)) )	{
	read_uid=pw->pw_uid;
	read_uname=pw->pw_name;
    } else {
	/* The special user only MUST exist in switching mode,
	 * otherwise root is the trusted user (reading as calling user) */
	if (glexec_is_switching())  {
	    glexec_log(GLEXEC_LOG_ERROR,
			"Unprivileged user '%s' was not found on this system.\n",
			GLEXEC_USER);
	    rc=GLEXEC_CONFIG_READING_ERROR;
	    goto finalize_read_configuration;
	}
	glexec_log(GLEXEC_LOG_WARNING,
		    "Unprivileged user '%s' was not found on this system.\n",
		    GLEXEC_USER);
	read_uid=0;
	read_uname="root";
    }

    /* Then group */
    if ( (gr = glexec_getgrnam(GLEXEC_CONFIG_GROUP)) )	{
	read_gid=gr->gr_gid;
	read_gname=gr->gr_name;
    } else {
	/* The special group only MUST exist in switching mode,
	 * otherwise ignore */
	if (glexec_is_switching())  {
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Unprivileged group '%s' needed for reading config "
		    "was not found on this system, "
		    "continuing with default (built-in) values.\n",
		    GLEXEC_CONFIG_GROUP);
	    rc=GLEXEC_CONFIG_READING_ERROR;
	    goto finalize_read_configuration;
	}
	glexec_log(GLEXEC_LOG_WARNING,
		    "Unprivileged group '%s' was not found on this system.\n",
		    GLEXEC_CONFIG_GROUP);
	/* Use effective gid instead */
	read_gid=getegid();
	read_gname="";
    }

    /* Read in config file with unprivileged uid/gid */
    rc = glexec_conf_read_config(GLEXEC_CONFIG_FILE,read_uid,read_gid);

    if (rc!=GLEXEC_SUCCESS) {
	/* Reading failed, now log what we tried to do */
	if (glexec_is_switching())
	    glexec_log(GLEXEC_LOG_WARNING,
		"Reading config file as %u(%s)/%d(%s) failed, "
		"continuing with built-in defaults.\n",
		read_uid,read_uname,read_gid,read_gname);
	else
	    glexec_log(GLEXEC_LOG_WARNING,
		"Continuing with default (built-in) values.\n");
    }

finalize_read_configuration:

    return rc;
}

/*!
 * Sets up the logging based on the configuration file: destination and
 * loglevel.
 */
void glexec_setup_log(void) {
    const char *value;
    char *endptr;
    long level=GLEXEC_LOG_INFO;	/* default cutoff for log_level */
    /* Whether gLExec loglevels should translate into different syslog levels,
     * default is yes (1) */
    int diff_syslog_levels=glexec_conf_boolean("glexec:diff_syslog_levels", 1);

    /* Set correct syslog logging daemon in case one is needed */
    if ((value = glexec_conf_value("glexec:syslog_facility")))
	glexec_set_syslog_facility(value);

    /* Find destination and open as soon as possible */
    if ((value = glexec_conf_value("glexec:log_destination"))) {
        if ((strcasecmp(value, "file")==0)) {
	    glexec_set_log_to_file();
	    glexec_open_log();
	} else	{ /* In any case use syslog in this case */
	    glexec_set_log_to_syslog();
	    glexec_open_log();
	    if ((strcasecmp(value, "syslog")!=0)) {
		glexec_log(GLEXEC_LOG_ERROR,
		    "Value (%s) for key 'glexec:log_destination' "
		    "not understood. Use either syslog or file\n", value);
		glexec_exit(GLEXEC_CONF_VALUE_MALFORMED,
			    GLEXEC_NO_CHILD_EXIT_CODE);
	    }
        }
    }

    /* Set correct log level */
    if ((value = glexec_conf_value("glexec:log_level")))    {
	errno=0;
	level=strtol(value,&endptr,10);
	if (errno!=0 || endptr[0]!='\0')    {
	    glexec_log(GLEXEC_LOG_ERROR,
		"Value (%s) for key 'glexec:log_level' not understood. "\
		"Use number in the range [%d:%d].\n", value,
		GLEXEC_LOG_LEVEL_LOW,GLEXEC_LOG_LEVEL_HIGH-1);
            glexec_exit (GLEXEC_CONF_VALUE_MALFORMED,GLEXEC_NO_CHILD_EXIT_CODE);
        }
	if (level<GLEXEC_LOG_LEVEL_LOW)	{
	    /* truncate a too large loglevel */
	    glexec_log(GLEXEC_LOG_WARNING,
		"Specified loglevel %ld is too small, "
		"using min. log level %u.\n",
		level,GLEXEC_LOG_LEVEL_LOW);
	    level=GLEXEC_LOG_LEVEL_LOW;
	} else if (level>=GLEXEC_LOG_LEVEL_HIGH)	{
	    /* truncate a too large loglevel */
	    glexec_log(GLEXEC_LOG_WARNING,
		"Specified loglevel %ld is too large, "
		"using max. log level %d.\n",
		level,GLEXEC_LOG_LEVEL_HIGH-1);
	    level = GLEXEC_LOG_LEVEL_HIGH-1;
	}
    }

    /* Set the default log level and whether to differentiate */
    glexec_set_log_level((glexec_log_level_t)level,diff_syslog_levels);
}


/*!
 * Determines the real and effective caller of gLExec and fills the two account
 * structures with their information. Exits when error is encountered. The
 * fields uid, uname, dir (not for ecaller), gid, gname are set. nsgid and sgid
 * are initialized to -1 and NULL.
 * \param caller real caller
 * \param ecaller effective caller
 */
void glexec_get_caller(glexec_account_t *caller, glexec_account_t *ecaller) {
    uid_t uid=getuid(),euid=geteuid();
    gid_t gid=getgid(),egid=getegid();
    struct passwd *pw;
    struct group *gr;

    /* Real uid */
    pw=glexec_getpwuid(uid);
    if (pw==NULL)   {
        glexec_log(GLEXEC_LOG_ERROR,
	    "crit: gLExec is called by a process with an invalid uid: (%u)\n",
	    uid);
        glexec_exit(GLEXEC_UNKNOWN_USER, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    /* Real gid */
    gr=glexec_getgrgid(gid);
    if (gr==NULL)  { 
	glexec_log(GLEXEC_LOG_ERROR,
	    "crit: gLExec is called by a process with an invalid gid: (%u)\n",
	    gid);
	glexec_exit(GLEXEC_UNKNOWN_GROUP, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    /* Fill caller struct */
    caller->uid=pw->pw_uid;
    caller->uname=pw->pw_name;
    caller->dir=pw->pw_dir;
    caller->gid=gr->gr_gid;
    caller->gname=gr->gr_name;
    /* Now check secondary group membership */
    caller->nsgid=getgroups(0,NULL);
    /* Not strictly necessary, but a lot safer not to run malloc etc. with
     * numgroups == 0 */
    if (caller->nsgid > 0)	{
	/* Can safely cast nsgid to size_t iff it is positive */
	caller->sgid=gmalloc((size_t)caller->nsgid * sizeof(gid_t));
	if (getgroups(caller->nsgid, caller->sgid)<0)	{
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error: getgroups() failed: %s\n",strerror(errno));
	    free(caller->sgid);
	    caller->sgid=NULL;
	    caller->nsgid=0;
	}
    } else  {
	if (caller->nsgid<0)
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error: getgroups() failed: %s\n",strerror(errno));
	caller->sgid=NULL;
	caller->nsgid=0;
    }

    /* Effective uid */
    pw=glexec_getpwuid(euid);
    if (pw==NULL)   {
        glexec_log(GLEXEC_LOG_ERROR,
	    "crit: gLExec is called by a process with an invalid euid: (%u)\n",
	    euid);
        glexec_exit(GLEXEC_UNKNOWN_USER, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    /* Effective gid */
    gr=glexec_getgrgid(egid);
    if (gr==NULL)   { 
	glexec_log(GLEXEC_LOG_ERROR,
	    "crit: gLExec is called by a process with an invalid egid: (%u)\n",
	    egid);
	glexec_exit(GLEXEC_UNKNOWN_GROUP, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    /* Fill caller struct */
    ecaller->uid=pw->pw_uid;
    ecaller->uname=pw->pw_name;
    ecaller->gid=gr->gr_gid;
    ecaller->gname=gr->gr_name;
    ecaller->nsgid=-1;
    ecaller->sgid=NULL;

    /* Store the information also in the internal glexec_ipc variable */
    glexec_set_caller_uid(caller->uid);

    /* Log the uid/gid information */
    glexec_log(GLEXEC_LOG_INFO,
	    "Caller uid: real/eff %s(%u)/%s(%u), gid: real/eff %s(%u)/%s(%u)\n",
            caller->uname, caller->uid, ecaller->uname, ecaller->uid,
            caller->gname, caller->gid, ecaller->gname, ecaller->gid);
}

/*!
 * Determines if given caller is allowed to run gLExec based on either explicit
 * whitelisting in the config file, or membership of the special group
 * GLEXEC_WHITELIST_GROUP.
 * \param caller account structure of the gLExec caller
 * \return GLEXEC_SUCCESS when user is allowed or exit when not when not.
 */
GLEXEC_ERROR glexec_check_caller_allowed(glexec_account_t *caller) {
    int i,numgroups;
    gid_t *groups;
    struct group *this_gr=NULL;
    const char* glexec_whitelist_group;

    /*
     * If the whitelist is available in the configuration (when readable) 
     * and if the user is allowed according to it, otherwise check the 
     * secondary groups of the user to figure out if the special user is 
     * available
     */
    if ( glexec_uname_allowed(caller->uname) ) {
        glexec_log(GLEXEC_LOG_DEBUG,
	    "The user \"%s\" is in the configuration file whitelist\n",
	    caller->uname);
        return GLEXEC_SUCCESS;
    }

    /* Determine the gLExec whitelist group: default is GLEXEC_WHITELIST_GROUP,
     * which presumably is glexec */
    glexec_whitelist_group=
	glexec_conf_value_default("glexec:group_white_list",
				  GLEXEC_WHITELIST_GROUP);
    /* Check primary group */
    if (strcmp(glexec_whitelist_group, caller->gname) == 0) {
	glexec_log(GLEXEC_LOG_INFO,
	    "The user \"%s\" is whitelisted because (s)he " \
	    "is a member of the group %s\n",
	    caller->uname, glexec_whitelist_group);
	return GLEXEC_SUCCESS;
    }

    /* Now check secondary group membership */
    numgroups = getgroups (0, NULL);
    /* Not strictly necessary, but a lot safer not to run malloc etc. with
     * numgroups == 0 */
    if (numgroups > 0)	{
	/* Can safely cast numgroups to size_t iff it is positive */
	groups = gmalloc ((size_t)numgroups * sizeof(gid_t));
	if (getgroups(numgroups, groups)==-1)
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Error: getgroups() failed: %s\n",strerror(errno));
	else {
	    for (i = 0; i < numgroups; i++) 
	    {
		this_gr = glexec_getgrgid(groups[i]);

		/* 
		 * Compare Secondary GIDs of the calling user with the special
		 * account to see if (s)he is a member of the special account
		 * Being a member of the special account group means that you
		 * automagically whitelisted
		 */ 
		if (this_gr && this_gr->gr_name &&
		    (strlen(this_gr->gr_name) > 0) &&
		    strcmp(glexec_whitelist_group, this_gr->gr_name) == 0) {
		    glexec_log(GLEXEC_LOG_INFO,
			    "the user \"%s\" is whitelisted because (s)he " \
			    "is a member of the group %s\n",
			    caller->uname, glexec_whitelist_group);

		    free(groups);
		    return GLEXEC_SUCCESS;
		}
	    }
	}

	/* No match */
	free(groups);
    }

    glexec_log(GLEXEC_LOG_INFO,
	"No secondary group matched \"%s\". "
	"The Unix user \"%s\" is not whitelisted.\n",
	GLEXEC_WHITELIST_GROUP, caller->uname);

    glexec_exit(GLEXEC_USER_NOT_WHITELISTED, GLEXEC_NO_CHILD_EXIT_CODE);

    return GLEXEC_USER_NOT_WHITELISTED; /* NOTE: We never get here ... */
}

/*!
 * sets up the internal gLExec environment, based on the configuration file.
 * \param user calling username
 * \return GLEXEC_SUCCESS on success or a GLEXEC_ERROR.
 */
GLEXEC_ERROR glexec_setup_glexec_env(const char *user)    {
    struct timeval mytime;
    struct tm *l_time;
    char jobrepoid[_JOBREPOID_LEN];
    const char *value;
    int mypid=(int)getpid(),rc_sn;

    /* Setup JOB REPOSITORY ID */
    gettimeofday(&mytime, NULL);
    l_time = gmtime(&mytime.tv_sec); /* always use UTC! */
    rc_sn=snprintf(jobrepoid, (size_t)_JOBREPOID_LEN,
	"%04d-%02d-%02d.%02d:%02d:%02dZ-%05d",
	1900+l_time->tm_year, 1+l_time->tm_mon, l_time->tm_mday,
	l_time->tm_hour, l_time->tm_min, l_time->tm_sec,
	mypid&0xFFFF);
    if (rc_sn<0 || rc_sn>=_JOBREPOID_LEN || 
	    setenv("JOB_REPOSITORY_ID", jobrepoid, 1))     { 
	glexec_log(GLEXEC_LOG_ERROR,"Setting JOB_REPOSITORY_ID failed.\n");
	glexec_exit(GLEXEC_OUT_OF_MEMORY, GLEXEC_NO_CHILD_EXIT_CODE);
    } else
	glexec_log(GLEXEC_LOG_DEBUG, "JOB_REPOSITORY_ID = %s\n", jobrepoid);

    /* Check wether user may leave out a private key */
    if (!glexec_user_needs_private_key(user)) {
	glexec_log(GLEXEC_LOG_INFO,
	    "User '%s' is allowed not to present "
	    "its private key in its proxy.\n", user);
	glexec_log(GLEXEC_LOG_DEBUG,
	    "Setting environment variable: "
	    "VERIFY_PROXY_DISCARD_PRIVATE_KEY_ABSENCE\n");
	if (setenv("VERIFY_PROXY_DISCARD_PRIVATE_KEY_ABSENCE","true",1) != 0) {
	    glexec_log(GLEXEC_LOG_WARNING,
		"The environment variable "
		"'VERIFY_PROXY_DISCARD_PRIVATE_KEY_ABSENCE' has not been set. "
		"Continuing.\n");
	}
    }

    /* Check for the certdir variable */
    if ((value = glexec_conf_value("glexec:certdir"))) {
	/* Check it exists as a directory */
	if (glexec_dir_is_abs(value))	{	
	    if (setenv("X509_CERT_DIR",value,1))    {
		glexec_log(GLEXEC_LOG_DEBUG,
		    "Setting X509_CERT_DIR failed, exiting.\n");
		glexec_exit(GLEXEC_OUT_OF_MEMORY, GLEXEC_NO_CHILD_EXIT_CODE);
	    }
	} else	{
	    glexec_log(GLEXEC_LOG_WARNING,
		"Warning: ignoring certdir config option: "\
		"%s is not an absolute path to a valid directory.\n", value);
	}
    }

    /* Check for the vomsdir variable */
    if ((value = glexec_conf_value("glexec:vomsdir"))) {
	/* Check it exists as a directory */
	if (glexec_dir_is_abs(value))	{	
	    if (setenv("X509_VOMS_DIR",value,1))    {
		glexec_log(GLEXEC_LOG_DEBUG,
		    "Setting X509_VOMS_DIR failed, exiting.\n");
		glexec_exit(GLEXEC_OUT_OF_MEMORY, GLEXEC_NO_CHILD_EXIT_CODE);
	    }
	} else	{
	    glexec_log(GLEXEC_LOG_WARNING,
		"Warning: ignoring vomsdir config option: "\
		"%s is not an absolute path to a valid directory.\n", value);
	}
    }

    return GLEXEC_SUCCESS;
}

/*!
 * Reads in GLEXEC_CLIENT_CERT and when needed GLEXEC_SOURCE_PROXY. When
 * GLEXEC_SOURCE_PROXY is not specified, its default value is GLEXEC_CLIENT_CERT
 * which is then just copied in memory.
 * \param pem_string will contain the GLEXEC_CLIENT_CERT
 * \param source_proxy will contain the GLEXEC_SOURCE_PROXY
 * \param source_proxy_size will contain the size of GLEXEC_SOURCE_PROXY
 * \return GLEXEC_SUCCESS on success or exits upon error
 */
GLEXEC_ERROR glexec_get_proxies(char **pem_string, char **source_proxy) {
    int rc;
    char *client_cert_file, *client_cert_fullfile=NULL, *tmp=NULL;
    char *src_proxy_file, *src_proxy_fullfile=NULL;
    char *target_location=NULL;
    GLEXEC_LOCK_MECH_T input_lock_mech=glexec_get_input_lock_mechanism();

    /* First get GLEXEC_client_cert_file: has to exist  */
    if ( (client_cert_file = getenv("GLEXEC_CLIENT_CERT"))==NULL ||
         strlen(client_cert_file)==0 ) {
            glexec_log(GLEXEC_LOG_ERROR, "$GLEXEC_CLIENT_CERT is empty.\n");
	    glexec_exit(GLEXEC_FAILURE_READING_PROXY_ENV,
		        GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Make path absolute */
    if (glexec_make_var_absolute("GLEXEC_CLIENT_CERT", 
			        client_cert_file,&client_cert_fullfile))
	glexec_exit(GLEXEC_FAILURE_READING_PROXY_ENV,GLEXEC_NO_CHILD_EXIT_CODE);

    /* Read glexec_client_cert */
    glexec_log(GLEXEC_LOG_INFO, "Reading in GLEXEC_CLIENT_CERT='%s'.\n",
	       client_cert_fullfile);
    if ((rc=glexec_read_proxy(client_cert_fullfile,input_lock_mech,&tmp))!=
	    GLEXEC_SUCCESS) {
	glexec_log(GLEXEC_LOG_ERROR, "Reading proxy failed.\n");
	glexec_exit(rc, GLEXEC_NO_CHILD_EXIT_CODE);
    }
    *pem_string = tmp;

    /* Do we need to write a TARGET_PROXY and hence need a SOURCE_PROXY?
     * Default is yes (1). */
    if ( !glexec_conf_boolean("glexec:create_target_proxy", 1) )	{
	glexec_log(GLEXEC_LOG_DEBUG,
		"Writing of target proxy is disabled.\n");
	return GLEXEC_SUCCESS;
    }

    /* See if we used the special target /dev/null */
    target_location=getenv("GLEXEC_TARGET_PROXY");
    if (target_location && strcmp(target_location,"/dev/null")==0)  {
	glexec_log(GLEXEC_LOG_DEBUG,
		"GLEXEC_TARGET_PROXY equals \"/dev/null\", "
		"not writing a target proxy.\n");
	return GLEXEC_SUCCESS;
    }


    /*  Check if there is a GLEXEC_SOURCE_PROXY setting, if not use
     *  GLEXEC_CLIENT_CERT instead */
    if ((src_proxy_file = getenv("GLEXEC_SOURCE_PROXY"))==NULL ||
         strlen(src_proxy_file)==0 ) {
	/* Unset: use GLEXEC_CLIENT_CERT instead */
        glexec_log(GLEXEC_LOG_INFO,
	    "GLEXEC_SOURCE_PROXY not specified. " \
	    "Using GLEXEC_CLIENT_CERT instead.\n");
	/* Setup GLEXEC_SOURCE_PROXY env var */
	if ( (setenv("GLEXEC_SOURCE_PROXY",client_cert_fullfile,1))!=0 )    {
	    glexec_log(GLEXEC_LOG_ERROR,
		"Cannot set GLEXEC_SOURCE_PROXY to %s.\n",client_cert_fullfile);
	    glexec_exit( (errno==ENOMEM ? GLEXEC_OUT_OF_MEMORY : 
					  GLEXEC_SETENV_ERROR),
		        GLEXEC_NO_CHILD_EXIT_CODE);
	}
	/* Just copy the memory of pem_string into source_proxy: need
	 * copy since we free them at different times... */
	*source_proxy=gstrdup(*pem_string); /* gstrdup handles out-of-mem */

	return GLEXEC_SUCCESS;
    }

    /* Make path absolute */
    if (glexec_make_var_absolute("GLEXEC_SOURCE_PROXY", 
				src_proxy_file,&src_proxy_fullfile))
	glexec_exit(GLEXEC_FAILURE_READING_PROXY_ENV,GLEXEC_NO_CHILD_EXIT_CODE);

    /*
     *  Read the proxy now, because the process should have sufficient
     *  privileges to read it, i.e. the owner of this process should match
     *  the owner of the proxy to be read. After LCMAPS has finished this
     *  no-longer needs to be. So, do it here rather then later.
     */
    glexec_log(GLEXEC_LOG_DEBUG, "Reading in GLEXEC_SOURCE_PROXY='%s'.\n",
	    src_proxy_fullfile);
    if ( (rc=glexec_read_proxy(src_proxy_fullfile,input_lock_mech,source_proxy))
	    != GLEXEC_SUCCESS )
	glexec_exit(rc,GLEXEC_NO_CHILD_EXIT_CODE);
    
    return GLEXEC_SUCCESS;
}

/*!
 * Does the authentication and authorization for gLExec (LCAS and LCMAPS), based
 * on pem_string (GLEXEC_CLIENT_CERT) and returns a mapped user account in
 * account_info.
 * \param argv cmd+arguments for executable to be run, used to construct a RSL.
 * \param pem_string contains the proxy from GLEXEC_CLIENT_CERT which will be
 *        freed before returning.
 * \param account_info will contain the target account.
 * \return GLEXEC_SUCCESS on success or a GLEXEC_ERROR.
 */
GLEXEC_ERROR glexec_auth(const char * const *argv, char *pem_string,
			 glexec_account_t *account_info) {
    char *rsl=NULL;
    GLEXEC_ERROR rc = GLEXEC_SUCCESS;

    /* Check if LCAS needs to run: either not specified in the config file
     * or not equal to either no or false.
     * Default is yes (1) */
    if ( glexec_conf_boolean("glexec:use_lcas", 1) )  {
	/* Get RSL string */
	rc=glexec_generate_rsl_string(argv[0], argv+1, &rsl);
	if (rc == GLEXEC_SUCCESS)   {
	    /* Now really run LCAS */
	    rc=glexec_lcas(pem_string, rsl);
	    /* Free RSL string (local variable) */
	    free(rsl);
	}
    }

    /* Setup and run LCMAPS */
    if (rc==GLEXEC_SUCCESS)
	rc=glexec_lcmaps(pem_string, account_info);

    /* Free the read PEM string and finalize */
    free(pem_string);

    if (rc!=GLEXEC_SUCCESS) {
	/* Only log this on debug level, normal error output should be coming
	 * from glexec_exit() */
	glexec_log(GLEXEC_LOG_DEBUG,
		"glexec_auth() failed with rc=%d, exiting.\n",rc);
	glexec_exit(rc, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    return GLEXEC_SUCCESS;
}

/*!
 * Performs uid/gid checks on the obtained target account:
 * any uid and gid has to be non-zero and >= GLEXEC_UID_MIN/GLEXEC_GID_MIN
 * \param target
 * \return GLEXEC_SUCCESS when successful or exit gLExec.
 */
GLEXEC_ERROR glexec_check_target_allowed(glexec_account_t *target)	{
    int i;

    /*
     * Error out if attempt is made to execute as root or as
     * a UID less than GLEXEC_UID_MIN.  Tsk tsk.
     */
    if ((target->uid == 0) || (target->uid < GLEXEC_UID_MIN)) {
      glexec_log(GLEXEC_LOG_ERROR, "cannot run as forbidden uid %u\n",
	         target->uid);
      glexec_exit(GLEXEC_MAPPED_TO_FORBIDDEN_USER, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /*
     * Error out if attempt is made to execute as root group
     * or as a GID less than GLEXEC_GID_MIN.  Tsk tsk.
     */
    if ((target->gid == 0) || (target->gid < GLEXEC_GID_MIN)) {
	glexec_log(GLEXEC_LOG_ERROR,
		    "cannot run as forbidden primary gid %u\n",
		    target->gid);
	glexec_exit(GLEXEC_MAPPED_TO_FORBIDDEN_GROUP,GLEXEC_NO_CHILD_EXIT_CODE);
    }
    /*
     * Perform similar Anti-ROOT test for all secondary GIDs - includes the same
     * error code!
     */
    for (i=0; i<target->nsgid; i++) {
	if ((target->sgid[i] == 0) || (target->sgid[i] < GLEXEC_GID_MIN)) {
	    glexec_log(GLEXEC_LOG_ERROR,
			"cannot run as forbidden secondary gid %u\n",
			target->sgid[i]);
	    glexec_exit(GLEXEC_MAPPED_TO_FORBIDDEN_GROUP,
			GLEXEC_NO_CHILD_EXIT_CODE);
	}
    }

    return GLEXEC_SUCCESS;
}

/*!
 * Sets up the current working directory for target, trying a chdir to the
 * correct homedirectory (for logging-only-mode this will be that of the
 * caller), if that fails, it tries to return to the startup dir, if that also
 * fails it will remain in GLEXEC_RUNTIME_CWD.
 * \param caller account of the gLExec caller.
 * \param target account of the payload user.
 * \return current working directory for payload.
 */
const char *glexec_setup_target_cwd(glexec_account_t *caller,
			      glexec_account_t *target)	{
    const char *target_cwd=glexec_is_switching() ? target->dir : caller->dir;

    /* Change to homedir of target user, if it fails, stay in current working
     * directory at calling time (behaviour as in 0.6.8-3 versions) and warn the
     * site-admin. Don't warn on stderr, because it might be default
     * site-behaviour. */
    if (chdir(target_cwd))	{
	/* Treat non-existing homedir only as info */
	glexec_log(GLEXEC_LOG_INFO,"Cannot chdir to target dir \"%s\": %s\n",
		   target_cwd,strerror(errno));
	/* Try to go back to directory at calling time */
	target_cwd=glexec_get_caller_cwd();
	if (chdir(target_cwd))	{
	    /* Cannot return to directory at calling time: treat as notice */
	    glexec_log(GLEXEC_LOG_NOTICE,"Cannot chdir back to dir \"%s\": %s\n",
		   target_cwd,strerror(errno));
	    target_cwd=GLEXEC_RUNTIME_CWD;
	}
    }

    glexec_log(GLEXEC_LOG_DEBUG,"Current working directory is now %s.\n",
	    target_cwd);

    return target_cwd;
}

/*!
 * Sets up the target account for target. This includes a chdir to the correct
 * homedirectory (for logging-only-mode this will be that of the caller) and
 * setting the env variables PATH, HOME, USER and LOGNAME in addition to those
 * whitelisted in the gLExec configuration file.
 * The environment can be obtained using the glexec_get_target_env() function.
 * \param caller account of the gLExec caller.
 * \param target account of the payload user.
 * \param dest_proxy_name filename of proxy for payload user, or NULL on none,
 * will be freed on succesful use.
 */
void glexec_setup_target_account(glexec_account_t *caller,
				 glexec_account_t *target,
				 char *dest_proxy_name)	{
    const char *preservevars=glexec_conf_value("glexec:preserve_env_variables");

    /* Setup target variables */
    glexec_setup_target_environment(preservevars);
    
    /* Set target PATH variable */
    glexec_target_setenv("PATH", GLEXEC_TARGET_PATH);
    /* Set target HOME variable */
    glexec_target_setenv("HOME",
	    glexec_is_switching() ? target->dir : caller->dir);
    /* Set target USER variable */
    glexec_target_setenv("USER", target->uname);
    /* Set target LOGNAME variable */
    glexec_target_setenv("LOGNAME", target->uname);

    /* Setup proxy variables when needed */
    if (dest_proxy_name)    {
	/* Set target user X509_USER_PROXY variable */
	glexec_target_setenv("X509_USER_PROXY",dest_proxy_name);
	
	/* When GLEXEC_TARGET_PROXY was unset, make sure we set it up,
	 * for backwards compatibility! */
	if (getenv("GLEXEC_TARGET_PROXY")==NULL)
	    glexec_target_setenv("GLEXEC_TARGET_PROXY", dest_proxy_name);

	free(dest_proxy_name);
    } else {
	/* Make sure GLEXEC_SOURCE_PROXY and GLEXEC_TARGET_PROXY are unset! */
	glexec_target_unsetenv("GLEXEC_SOURCE_PROXY");
	glexec_target_unsetenv("GLEXEC_TARGET_PROXY");
    }
}

/*!
 * Returns whether gLExec should do the user switch or not.
 */
int glexec_posix_enf_by_glexec(void)	{
    static int posix_enf_by_glexec=-1;
    const char *value;

    if (posix_enf_by_glexec==-1)    { /* As yet undefined */
	value=glexec_conf_value("glexec:user_identity_switch_by");
	if (value)  {
	    if (strcmp(value, "glexec") == 0)
		posix_enf_by_glexec=1;
	    else if (strcmp(value, "lcmaps") == 0) 
		 posix_enf_by_glexec=0;
	    else {
		glexec_log(GLEXEC_LOG_ERROR,
		    "Value '%s' is not valid for option "
		    "'user_identity_switch_by'.\n",
		    value);
		glexec_exit(GLEXEC_CONF_FAIL_USER_ID_SWITCH_BY,
			    GLEXEC_NO_CHILD_EXIT_CODE);
	    }
	} else /* default is gLExec does the switch */
	    posix_enf_by_glexec=1;
    }
    return posix_enf_by_glexec;
}
/*!
 * Performs the userswitch, except for saved-setuid root. We are only called in
 * switching mode, so we should be (effective) root! 
 * \param target target account
 * Upon error gLExec exits.
 */
void glexec_userswitch(glexec_account_t *target)	{
    int rc_sg;
    long ngroups_max;

    /* Check we are actually effective root: if someone has already changed
     * identity, we should not be here, and not try to switch again, even though
     * it probably won't fail. */
    if (geteuid()!=0 && getuid()!=0)	{
	/* User switch has already taken place.*/
	glexec_exit(GLEXEC_DOUBLE_USERSWITCH, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Check number of sGIDs isn't too large */
    if (target->nsgid > 0)  {
	ngroups_max = sysconf(_SC_NGROUPS_MAX);
	if (ngroups_max==-1) {
#ifdef NGROUPS_MAX
	    ngroups_max=NGROUPS_MAX;
#else
	    glexec_log(GLEXEC_LOG_WARNING,
		"Maximum number of groups is unknown. Will assume '%ld'.\n",
		CUSTOM_NGROUPS_MAX);
	    ngroups_max=CUSTOM_NGROUPS_MAX;
#endif
	}
	/* Now check actual number of groups */
	if (target->nsgid > ngroups_max)   {
	    glexec_log(GLEXEC_LOG_ERROR,
		"Number of groups %d is larger than maximum %ld.\n",
		target->nsgid, ngroups_max);
	    glexec_exit(GLEXEC_TOO_MANY_GROUPS, GLEXEC_NO_CHILD_EXIT_CODE); 
	}
    }

    /* secondary group IDs:
     * NOTE 1: we need setgroups to remove the calling user's secondary gids,
     * i.e. even when nsgid==0.
     * NOTE 2: BSD uses the first element in the group list to set the effective
     * GID. When none are present MacOSX uses group 0! Hence to prevent any
     * problems, always call setgroups with at least 1 group, and use the
     * primary gid as default.
     * NOTE 3: Also protect against situation where nsgid somehow is negative */
#ifdef __linux__
    if (target->nsgid<=0)
	rc_sg=setgroups((size_t)1, &(target->gid));
    else
	/* Can safely cast nsgid to size_t iff it is positive */
	rc_sg=setgroups((size_t)target->nsgid, target->sgid);
#else
    if (target->nsgid<=0)
	rc_sg=setgroups(1, &(target->gid));
    else
	rc_sg=setgroups(target->nsgid, target->sgid);
#endif
    /* Check the result */
    if (rc_sg != 0) {
	switch (errno) {
	    case EPERM:
		glexec_log(GLEXEC_LOG_ERROR,
		    "You are not (effective) ROOT: %s\n",
		    strerror(errno));
		glexec_exit(GLEXEC_INSUFFICIENT_PRIVILEGES,
			    GLEXEC_NO_CHILD_EXIT_CODE);
	    case EFAULT:
		glexec_log(GLEXEC_LOG_WARNING,
		    "Not that fatal but serious error in setgroups: %s\n",
		    strerror(errno));
		glexec_exit(GLEXEC_UNCLASIFIED, GLEXEC_NO_CHILD_EXIT_CODE);
	    case EINVAL:
		glexec_log(GLEXEC_LOG_ERROR, "Invalid sGID list: %s\n",
		       strerror(errno));
		glexec_exit(GLEXEC_UNCLASIFIED, GLEXEC_NO_CHILD_EXIT_CODE);
	    default:
		glexec_log(GLEXEC_LOG_ERROR,
		       "Unspecified error in setgroups(): %s\n",
		       strerror(errno));
		glexec_exit(GLEXEC_UNCLASIFIED, GLEXEC_NO_CHILD_EXIT_CODE);
	}
    }

    /* primary group ID. We use setgid, we thought setgid does not work
     * correctly on Debian-5, but it does (the problem was unrelated) and setgid
     * is preferred, at least according to Darwin manpage */
    if ((setgid(target->gid)) != 0) {
        glexec_log(GLEXEC_LOG_ERROR, "failed to setgid %u\n", target->gid);
        switch (errno) {
	    case EPERM:
		/* This probably never happens here... */
		glexec_log(GLEXEC_LOG_ERROR, "You are not (effective) ROOT: %s\n",
			    strerror(errno));
		glexec_exit(GLEXEC_INSUFFICIENT_PRIVILEGES,
			    GLEXEC_NO_CHILD_EXIT_CODE);
	    case EINVAL:
		glexec_log(GLEXEC_LOG_ERROR, "Invalid pGID: %s\n",
		       strerror(errno));
		glexec_exit(GLEXEC_UNCLASIFIED, GLEXEC_NO_CHILD_EXIT_CODE);
	    default:
		glexec_log(GLEXEC_LOG_ERROR,
			    "Unspecified error in setgid(): %s\n",
			    strerror(errno));
		glexec_exit(GLEXEC_UNCLASIFIED, GLEXEC_NO_CHILD_EXIT_CODE); 
	}
    }

    /* Set effective uid to target, real will be set later on, this also
     * prevents target user (on Linux) from being able to send us any signal */
    if (seteuid(target->uid)!=0) {
	glexec_log(GLEXEC_LOG_ERROR,"Failed to seteuid %u\n",target->uid);
	glexec_exit(GLEXEC_INSUFFICIENT_PRIVILEGES, GLEXEC_NO_CHILD_EXIT_CODE);
    }

}

/*!
 *  Create a new backlog item in the backlog_path specified in the config file,
 *  or will return when none is specified. The path used for backlogging needs
 *  to be an absolute path, if not glexec will exit with an error. It will
 *  create a filename based on the absolute path and target user name and then
 *  write the caller user name to it.
 *  \param caller    uid of the source user.
 *  \param target    uid of the destination user.
 *   \return 0 on success, != 0 otherwise.
 */
int glexec_create_back_log_item(glexec_account_t *caller,
                                glexec_account_t *target) {
  const char *path;
  char filename[AP_MAXPATH];
  size_t namelen;
  int fd,rc_sn;
  int pid=(int)getpid();

  /* See if we specified a path, otherwise return immediately */
  if ((path = glexec_conf_value("glexec:backlog_path")) == NULL)
      return 0;

  /* Relative paths a disallowed. */
  if (path[0] != '/') {
    glexec_log(GLEXEC_LOG_ERROR,"Backlogging directory is relative: %s.\n",path);
    glexec_exit (GLEXEC_BACKLOG_FAILED, GLEXEC_NO_CHILD_EXIT_CODE);
  }

  /* Create directory */
  if (cgul_mkdir_with_parents(path, S_IRUSR | S_IWUSR | S_IXUSR) == -1) {
    glexec_log(GLEXEC_LOG_ERROR,
	    "Failed to create backlogging directory \"%s\".\n", path);
    glexec_exit (GLEXEC_BACKLOG_FAILED, GLEXEC_NO_CHILD_EXIT_CODE);
  }

  /* Construct the file name of the backlog item.
   * Format: <path>/<target user name>:<pid> */
  rc_sn=snprintf(filename, (size_t)AP_MAXPATH, "%s/%s:%d", path, target->uname, (int)pid);
  if (rc_sn < 0)	{
      glexec_log(GLEXEC_LOG_ERROR, "Could not create backlog filename.\n");
      glexec_exit (GLEXEC_BACKLOG_FAILED, GLEXEC_NO_CHILD_EXIT_CODE);
  }
  if (rc_sn >= AP_MAXPATH)    {
      glexec_log(GLEXEC_LOG_ERROR, "Backlog filename is too long.\n");
      glexec_exit (GLEXEC_BACKLOG_FAILED, GLEXEC_NO_CHILD_EXIT_CODE);
  }

  /* Open the file by creating it and truncating if necessary, open it write
   * only. */
  if ((fd = open(filename, O_CREAT | O_TRUNC | O_WRONLY |
			   O_SYNC, S_IRUSR | S_IWUSR)) == -1)	{
      glexec_log(GLEXEC_LOG_ERROR,"Failed to open backlog file %s.\n",filename);
      glexec_exit (GLEXEC_BACKLOG_FAILED, GLEXEC_NO_CHILD_EXIT_CODE);
  }

  glexec_log(GLEXEC_LOG_DEBUG,
	  "Writing back-log entry '%s' to '%s'\n", caller->uname, filename);
  /* Write the calling user name to the open file. */
  namelen=strlen(caller->uname);
  /* Should compare with ssize_t, as it might be <0 */
  if (write(fd, caller->uname, namelen) != (ssize_t)namelen) {
    glexec_log(GLEXEC_LOG_ERROR, "Failed to write to backlog: %s.\n",filename);
    glexec_exit (GLEXEC_BACKLOG_FAILED, GLEXEC_NO_CHILD_EXIT_CODE);
  }

  if (close(fd))   {
      glexec_log(GLEXEC_LOG_ERROR,
	      "Error while closing %s: %s\n", filename, strerror(errno));
      glexec_exit (GLEXEC_LOG_FILE_CLOSE_FAIL, GLEXEC_NO_CHILD_EXIT_CODE);
  }

  return 0;
}

/*!
 * Runs a simple test that gLExec is no longer root.
 * \return effective uid when non-zero or exits when it is.
 */
uid_t glexec_check_no_longer_root(glexec_account_t *caller,
				  glexec_account_t *target)    {
    uid_t euid,uid;

    /* Only in switching mode we need to check something */
    if (glexec_is_switching())	{
	/* Check we are target... */
	if ( (uid=getuid()) != target->uid)	{
	    /* Something is wrong, check what! */
	    /* Check if we are still caller (most likely) */
	    if (uid==caller->uid)
		glexec_exit(GLEXEC_DID_NOT_USERSWITCH,
			    GLEXEC_NO_CHILD_EXIT_CODE);
	    /* Even more wrong: we switched to another account?! */
	    glexec_log(GLEXEC_LOG_ERROR,
		"UID should have been %u but is %u.\n", target->uid, uid);
	    glexec_exit(GLEXEC_WRONG_USERSWITCH,GLEXEC_NO_CHILD_EXIT_CODE);
	}
	/* Now check effective root is gone: otherwise exit */
	if ( (euid=geteuid()) == 0)
	    glexec_exit(GLEXEC_STILL_EFFECTIVE_ROOT,GLEXEC_NO_CHILD_EXIT_CODE);
    } else
	euid=caller->uid;

    return euid;
}

/*!
 * This function checks the cmd and target cwd against some security issues: Not
 * set[ug]id, regular file, not changing and optionally the pedantic and
 * no-symlink checks. It obtains the absolute (fullcmd) and canonical absolute
 * (realcmd) path for cmd. In case of problems glexec exists.
 * \param cmd command to be executed
 * \param target_cwd current working directory at payload time
 * \param fullcmd will contain absolute path executable
 * \param realcmd will contain canonical absolute path executable
 * \param caller
 * \param target
 */
void glexec_check_exe_and_dir(char *cmd, const char *target_cwd,
			      char **fullcmd, char **realcmd,
			      glexec_account_t *caller,
			      glexec_account_t *target)  {
    struct stat fullcmd_stat, realcmd_stat;    /* program info holder       */
    static char fullcmd_buf[AP_MAXPATH]="";
    int rc_sn;

    *fullcmd=fullcmd_buf;

    /* Create non-relative path for command. */
    if (cmd[0]=='/')	{
	rc_sn=snprintf(*fullcmd, (size_t)AP_MAXPATH,"%s", cmd);
	if (rc_sn>=AP_MAXPATH)   {
	    glexec_log(GLEXEC_LOG_ERROR, "Exec name '%s' is too long.\n",cmd);
	    glexec_exit(GLEXEC_STAT_ERROR_EXECUTABLE_CMD,
			GLEXEC_NO_CHILD_EXIT_CODE);
	}
    } else {
	rc_sn=snprintf(*fullcmd, (size_t)AP_MAXPATH,"%s/%s",target_cwd,cmd);
	if (rc_sn>=AP_MAXPATH) {
	    glexec_log(GLEXEC_LOG_ERROR, "Exec name '%s/%s' is too long.\n",
		    target_cwd,cmd);
	    glexec_exit(GLEXEC_STAT_ERROR_EXECUTABLE_CMD,
			GLEXEC_NO_CHILD_EXIT_CODE);
	}
    }
    if (rc_sn < 0)	{
	glexec_log(GLEXEC_LOG_ERROR, "Error creating fullcmd\n");
	glexec_exit(GLEXEC_STAT_ERROR_EXECUTABLE_CMD,
		    GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Now stat the non-relative path (can still contain symlinks, /../ etc. */
    if (stat(*fullcmd, &fullcmd_stat)!=0)	{
	glexec_log(GLEXEC_LOG_ERROR,"Cannot stat program: (%s)\n", *fullcmd);
	glexec_log(GLEXEC_LOG_ERROR,"  errno: %d (%s)\n",errno,strerror(errno));
	glexec_exit(GLEXEC_STAT_ERROR_EXECUTABLE_CMD,GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Get realpath: absolute without symlinks */
    if ((*realcmd=cgul_realpath(*fullcmd, &fullcmd_stat))==NULL)  {
	glexec_log(GLEXEC_LOG_ERROR,
	    "Cannot get canonical absolute name for program '%s'\n", *fullcmd);
	glexec_log(GLEXEC_LOG_ERROR,"  errno: %d (%s)\n",errno,strerror(errno));
	glexec_exit(GLEXEC_STAT_ERROR_EXECUTABLE_CMD,GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Now stat the absolute, canonical path */
    if (stat(*realcmd, &realcmd_stat)!=0)	{
	glexec_log(GLEXEC_LOG_ERROR,"Cannot stat program: (%s)\n", *realcmd);
	glexec_log(GLEXEC_LOG_ERROR,"  errno: %d (%s)\n",errno,strerror(errno));
	glexec_exit(GLEXEC_STAT_ERROR_EXECUTABLE_CMD,GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* First check it's a real file */
    if (!S_ISREG(realcmd_stat.st_mode)) {
	glexec_log(GLEXEC_LOG_ERROR,"Program %s (=%s) is not a regular file.\n",
		   *fullcmd, *realcmd);
	glexec_exit(GLEXEC_STAT_ERROR_EXECUTABLE_CMD,GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Check they match: warn if not */
    if (fullcmd_stat.st_dev   != realcmd_stat.st_dev ||	    /* device */
	fullcmd_stat.st_ino   != realcmd_stat.st_ino ||	    /* inode */
	fullcmd_stat.st_mtime != realcmd_stat.st_mtime ||   /* mod time */
	fullcmd_stat.st_size  != realcmd_stat.st_size)  {   /* size */
	glexec_log(GLEXEC_LOG_WARNING,
	    "Program %s (=%s) has changed during examination. "
	    "Will execute '%s'.\n", *fullcmd, *realcmd, *realcmd);
    }

    /* Check it's neither setuid nor setgid */
    if ((realcmd_stat.st_mode & S_ISUID) || (realcmd_stat.st_mode & S_ISGID)) {
	glexec_log(GLEXEC_LOG_ERROR,
		"Program %s (=%s) is either setuid or setgid.\n",
		*fullcmd, *realcmd);
	glexec_exit(GLEXEC_WRONG_PERMISSION_ON_CMD, GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Do the pedantic checks */
    glexec_pedantic_checks(target_cwd,*fullcmd,*realcmd,
			    caller,target,&realcmd_stat);
}


/*!
 * Prints out a one-line log entry with all relevant information:
 * real calling username/groupname, target username/groupname, command to be
 * executed and its arguments.
 * \param argc for target process
 * \param argv for target process
 * \param cwd current working directory for target
 * \param fullcmd absolute path for target (can contain symlinks)
 * \param realcmd canonical absolute path for target
 * \param caller real calling user
 * \param target target user
 */
void glexec_log_argument_list(int argc, char *argv[],
			      const char *target_cwd,
			      char *fullcmd, char *realcmd,
			      glexec_account_t *caller,
			      glexec_account_t *target) {
    char print_arglist[GLEXEC_MAX_PRINT_ARGLIST]=""; /* will contain list of args */
    char *pos=print_arglist;
    int i,sn_rc;
    size_t print_arglen=0;

    /* Setup argument list array, print_arglen==0 now, pos==print_arglist */
    print_arglen=0;
    pos=print_arglist;
    for (i=1; i<argc && print_arglen<GLEXEC_MAX_PRINT_ARGLIST ; i++)	{
	/* make sure not to add whitespace at the start...*/
	sn_rc=snprintf(pos,GLEXEC_MAX_PRINT_ARGLIST-print_arglen,
				(print_arglen==0 ? "%s" : " %s"),argv[i]);
	/* when sn_rc is positive, it's safe to convert to a size_t */
	if (sn_rc>=0) print_arglen+=(size_t)sn_rc;
	pos=print_arglist+print_arglen*sizeof(char);
    }
    /* Truncate long argument list. Make sure to leave also space for \0 */
    if (print_arglen>=GLEXEC_MAX_PRINT_ARGLIST)
	sprintf(print_arglist+(GLEXEC_MAX_PRINT_ARGLIST-4)*sizeof(char),"...");

    /* Are there arguments to command? */
    if (print_arglen)    {
	glexec_log (GLEXEC_LOG_NOTICE,
	    "real_user(%s),real_group(%s)->" \
	    "mapped_user(%s),mapped_group(%s) " \
	    "cwd: \"%s\" cmd: \"%s\" (=\"%s\") args: \"%s\"\n",
	    caller->uname, caller->gname,  target->uname, target->gname,
	    target_cwd, fullcmd, realcmd, print_arglist);
    } else {
	glexec_log (GLEXEC_LOG_NOTICE,
	    "real_user(%s),real_group(%s)->" \
	    "mapped_user(%s),mapped_group(%s) " \
	    "cwd: \"%s\" cmd: \"%s\" (=\"%s\")\n",
	    caller->uname, caller->gname,  target->uname, target->gname,
	    target_cwd, fullcmd, realcmd);
    }
}

/*!
 *  This function writes the proxy for the target user/process, using
 *  target_lock_mechanism.
 *  When GLEXEC_TARGET_PROXY is set, this will be the output filename. When it
 *  is unset, the behaviour depends on the mode: in switching mode, a default
 *  proxy name will be constructed from the uid of the target user and a random
 *  string of characters. In logging only mode, no file will be written.
 *  In all cases X509_USER_PROXY and GLEXEC_TARGET_PROXY are set in the target
 *  environment to the correct proxyfile. In logging only mode without a set
 *  value for GLEXEC_TARGET_PROXY (when no file is written, see above) the value
 *  will be made equal to that of GLEXEC_SOURCE_PROXY or its default
 *  GLEXEC_CLIENT_CERT.
 *  \param caller for caller user
 *  \param target for target user
 *  \param source_proxy contents of GLEXEC_SOURCE_PROXY (or its default), will
 *	   be freed.
 *  \return filename of target proxy or NULL when none
 */
char *glexec_write_source_proxy(glexec_account_t *caller,
				       glexec_account_t *target,
				       char **source_proxy)
{
    const char *dest_proxy_name;
    char *sourceproxyvar, *targetproxyvar, buffer[AP_MAXPATH];
    int rc=0,rc_sn;
    int lock_type=0;
    mode_t old_umask;
    uid_t uid;
    gid_t gid;
    
    if (*source_proxy==NULL)	{
	glexec_log(GLEXEC_LOG_DEBUG,"NOT writing a target proxy.\n");
	return NULL;
    }

    /* Now get the proxy variables */
    sourceproxyvar=getenv("GLEXEC_SOURCE_PROXY");
    targetproxyvar=getenv("GLEXEC_TARGET_PROXY");

    if (glexec_is_switching())	{
	uid=target->uid; gid=target->gid;
    } else {
	uid=caller->uid; gid=caller->gid;
    }

    /* Set correct locking mechanism */
    switch(glexec_get_target_lock_mechanism()) {
	case GLEXEC_LOCK_FLOCK:	    lock_type=LCK_FLOCK; break;
	case GLEXEC_LOCK_FCNTL:	    lock_type=LCK_FCNTL; break;
	case GLEXEC_LOCK_DISABLED:  lock_type=LCK_NOLOCK; break;
	case GLEXEC_LOCK_UNKNOWN:
	default:
	    glexec_exit(GLEXEC_CONF_FAIL_LOCK_MECH,GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* Set the umask */
    old_umask = umask(GLEXEC_TARGET_UMASK);

    /* Set the correct dest_proxy_name and optionally write it:
     * If GLEXEC_TARGET_PROXY is set: write it, if it is not set, use default in
     * setuid case and use GLEXEC_SOURCE_PROXY in the logging only case */
    if (targetproxyvar==NULL)	{
	if (glexec_is_switching())  {
	    rc_sn=snprintf(buffer, (size_t)AP_MAXPATH,
			 GLEXEC_DEFAULT_NAME_TARGET_PROXY, uid);
	    if (rc_sn>=0 && rc_sn<AP_MAXPATH)	{
		rc=cgul_write_uniq_proxy(buffer,*source_proxy,
				         uid,gid);
		dest_proxy_name=buffer;
		/* Not so nice, but only now do we have the actual name. */
		if (rc==0)
		    glexec_log(GLEXEC_LOG_DEBUG,
			"Writing target user proxy: '%s'\n", dest_proxy_name);
	    } else  {
		/* Error or filename is too long: set rc different */
		dest_proxy_name=""; /* Set dest_proxy_name just in case */
		rc=-42;
	    }
	} else	{ /* Logging only mode: sharing homedir, don't write anything */
	    dest_proxy_name=sourceproxyvar;
	    glexec_log(GLEXEC_LOG_DEBUG,
		"Determined destination of target user proxy: '%s'\n",
		dest_proxy_name);
	    rc=0;
	}
    } else  {	/* GLEXEC_TARGET_PROXY was set */
	dest_proxy_name=targetproxyvar;
	glexec_log(GLEXEC_LOG_DEBUG,
	    "Writing target user proxy: '%s'\n", dest_proxy_name);
	/* Make sure to set explicitly the uid/gid, as we are still real uid
	 * caller at this point */
	rc=cgul_write_proxy(dest_proxy_name,lock_type,*source_proxy,
			    uid,gid);
    }
    /* Restore the umask to its original value. */
    umask(old_umask);

    /* we're done with source_proxy */
    free(*source_proxy); *source_proxy=NULL;

    /* Parse return code */
    switch(rc)	{
	case 0:
	    break;
	case -1: /* I/O error */
	    glexec_log(GLEXEC_LOG_ERROR,"Failed to write proxyfile %s.\n",
		    dest_proxy_name);
	    glexec_log(GLEXEC_LOG_ERROR,"  errno (%d): %s\n",
		    errno, strerror(errno));
	    glexec_exit(GLEXEC_FAILURE_WRITING_PROXY,
			GLEXEC_NO_CHILD_EXIT_CODE);
	case -2: /* privilege drop error */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Failed to set identity while writing proxy %s.\n",
		    dest_proxy_name);
	    glexec_exit(GLEXEC_COULD_NOT_DROP_PRIVILEGES,
			GLEXEC_NO_CHILD_EXIT_CODE);
	case -3: /* path is wrong */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Path or its permissions for writing proxy is wrong: %s.\n",
		    dest_proxy_name);
	    glexec_exit(GLEXEC_FAILURE_WRITING_PROXY,
			GLEXEC_NO_CHILD_EXIT_CODE);
	case -4: /* out of mem */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Out of memory while writing proxy %s.\n",
		    dest_proxy_name);
	    glexec_exit(GLEXEC_OUT_OF_MEMORY,
			GLEXEC_NO_CHILD_EXIT_CODE);
	case -5: /* template is incorrect (should end in 6 X)*/
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Template for writing proxy is wrong: %s.\n",
		    dest_proxy_name);
	    glexec_exit(GLEXEC_FAILURE_WRITING_PROXY,
			GLEXEC_NO_CHILD_EXIT_CODE);
	case -6: /* locking failed */
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Could not lock file during writing of proxy %s.\n",
		    dest_proxy_name);
	    glexec_exit(GLEXEC_LOCK_FAILURE, GLEXEC_NO_CHILD_EXIT_CODE);
	case -42: /* Filename in template became too long: NOT a
		     cgul_write_proxy() error code */
	    glexec_log(GLEXEC_LOG_ERROR,"Path is too long for target proxy.\n");
	    glexec_exit(GLEXEC_FAILURE_WRITING_PROXY,
			GLEXEC_NO_CHILD_EXIT_CODE);
	default:
	    glexec_log(GLEXEC_LOG_ERROR,
		    "Unknown error while writing proxy: %s.\n",
		    dest_proxy_name);
	    glexec_exit(GLEXEC_FAILURE_WRITING_PROXY,
			GLEXEC_NO_CHILD_EXIT_CODE);
    }

    return gstrdup(dest_proxy_name);
}

/*!
 * Determines if we should linger or not, if so also sets correct signalling
 * delays and determines linger account (default caller).
 * \param close_fds will be set to 1 if the lingering gLExec should close all
 * its filedescriptors (except stdin/out/err).
 * \return 1 when lingering (default), 0 otherwise
 */
int glexec_get_linger_mode(int *close_fds)	{
    int linger_mode;

    /* Should in linger? Default yes (1) */
    if ( (linger_mode=glexec_conf_boolean("glexec:linger", 1))==0 )
	glexec_log(GLEXEC_LOG_INFO,
	    "Forking disabled for executing the real user job.\n");
    else
	glexec_log(GLEXEC_LOG_DEBUG,
	    "Using fork for executing the real user job.\n");
   
    /* Now that we know we're in linger mode, we need to be able to use the
     * signal delays and know whether to use setpgid() in child processes. */
    glexec_set_signaloptions(linger_mode);

    /* Setup extra signal handlers (when the config file doesn't prevent us) */
    glexec_install_extra_sig_handlers();

    /* Following only when in linger_mode */
    if (linger_mode)	{
	/* Make sure we know who to linger as... */
	glexec_get_linger_as_payload();

	/* Set value for closing of file descriptors. */
	if ( (*close_fds=glexec_conf_boolean("glexec:close_fds", 1))==0 )
	    glexec_log(GLEXEC_LOG_INFO,
		"Not closing open file descriptors in lingering process.\n");
#ifdef _GLEXEC_CLOSEFROM_NEEDS_INIT
	else	{
	    /* If LCMAPS is doing the userswitch, we might need to initialize
	     * close_from on older RedHat Linux systems */
	    glexec_init_closefrom();
	}
#endif
    

    }

    return linger_mode;
}

/*!
 * Depending on the linger mode will optionally do a fork() and return the
 * corresponding pid_t. The start time will be set, in linger mode only for the
 * parent. It also closes all needed file descriptors: in parent all inherited
 * (when close_fds==1) and in payload for non-posix the directory fd. The
 * process for which glexec_fork_and_enforce() returns 0 is supposed to run the
 * payload.
 * \param linger_mode 1 for linger, 0 for non-linger
 * \param close_fds 1 for closing them, 0 for not.
 * \param caller account structure
 * \param target account structure
 * \return non-linger: 0, linger/parent: pid of child, linger/child: 0.
 * Upon error (failing fork) gLExec will exit.
 */
pid_t glexec_fork_and_enforce(int linger_mode, int close_fds,
			      glexec_account_t *caller,
			      glexec_account_t *target)	{
    int switching=glexec_posix_enf_by_glexec() && glexec_is_switching() ? 1 : 0;
    pid_t pid=0;
    
    /* cleanup target sgids */
    if (target->nsgid>0)    {
        free(target->sgid);
        target->sgid=NULL;
    }
	
    /* If we're in switching mode and done by gLExec: Switch to effective root,
     * such that we can easily switch to the correct uid after the fork: either
     * caller or target. */
    if (switching && seteuid(0)) {
	glexec_log(GLEXEC_LOG_ERROR, "failed to seteuid(0)\n");
	glexec_exit(GLEXEC_INSUFFICIENT_PRIVILEGES,
		    GLEXEC_NO_CHILD_EXIT_CODE);
    }

    /* In linger mode we will do the fork here */
    if (linger_mode)
	/* Actual fork: this also sets I/O for payload/parent */
	pid=glexec_forkchild();

    /* payload has pid==0, also in non-linger mode */
    if (pid==0)	{   /* Child or non-linger */
	/* Revert umask to initial one: Note that this cannot fail */
	umask(glexec_get_caller_umask());

	/* If we're in switching mode: need to enforce target uid */
	if (switching)  {
	    if ( setuid(target->uid) != 0 ) {
		glexec_log(GLEXEC_LOG_ERROR,
		    "This can not happen: failed to setuid %u\n",
		    target->uid);
		if (linger_mode)
		    /* In linger mode: need a proper child exit value: use
		     * GLEXEC_EXIT_INTERNAL_ERROR, which will trigger a
		     * GLEXEC_EXIT_CHILD_PROCESS_RC_OVERLAP, to show it's not a
		     * simple payload error */
		    exit(GLEXEC_EXIT_INTERNAL_ERROR);
		else
		    glexec_exit(GLEXEC_INSUFFICIENT_PRIVILEGES,
				GLEXEC_NO_CHILD_EXIT_CODE);
	    }
	    glexec_log(GLEXEC_LOG_DEBUG,"Payload uid=%u\n",target->uid);

#if ( (!defined(_POSIX_C_SOURCE) || _POSIX_C_SOURCE < 200112L) && \
      defined(_GLEXEC_CLOSEFROM_NEEDS_INIT) )
	    /* Make sure to close the dirp filepointer, only needed on pre-POSIX
	     * 2001 */
	    if (close_fds==1)	{
		glexec_log(GLEXEC_LOG_DEBUG,
		    "Closing dir handle used to close file descriptors.\n");
		glexec_close_dirp();
	    }
#endif
	}
    } else  {
	/* We need to upgrade back to either original or 
	 * to set real uid to payload */
	glexec_setup_linger_account(caller, target);

	/* Close all filedescriptors except STD{IN,OUT,ERR}. Close and open log
	 * */
	if (close_fds==1)   {
	    glexec_log(GLEXEC_LOG_DEBUG,"Closing all open file descriptors.\n");
	    glexec_close_fds();
	}
    }

    return pid;
}

/*!
 * Reverts the groups and (e)uid to that of the original calling account, only
 * having a saved-effective uid on root in order to have the ability to go back
 * up.
 * \param caller, target caller and target accounts
 */
void glexec_setup_linger_account(glexec_account_t *caller,
				 glexec_account_t *target)  {
    int rc_sg;

    switch(glexec_get_linger_as_payload())	{
	case -1:/* This cannot happen: linger_as_payload has to be defined */
	    glexec_log(GLEXEC_LOG_ERROR,
			"Coding error: linger status cannot be undefined "
			"at this point.\n");
	    glexec_exit(GLEXEC_UNCLASIFIED,
			GLEXEC_NO_CHILD_EXIT_CODE);
	    break;
	case 0:
	    /* We are real/eff caller/root, need to set groups and eff caller */
	    /* Protect against negative nsgid. Current code does not allow,
	     * but enforce. Also, BSD does not like setgroups(0,NULL) */
#ifdef __linux__
	    if (caller->nsgid<=0)
		rc_sg=setgroups((size_t)1,&(caller->gid));
	    else
		/* Can safely cast nsgid to size_t iff it is positive */
		rc_sg=setgroups((size_t)caller->nsgid,caller->sgid);
#else
	    if (caller->nsgid<=0)
		rc_sg=setgroups(1,&(caller->gid));
	    else
		rc_sg=setgroups(caller->nsgid,caller->sgid);
#endif
	    /* Check the setgroups() value and set the gid and euid */
	    if (rc_sg==-1 ||
		setgid(caller->gid)==-1 ||
		seteuid(caller->uid)==-1)   {
		    glexec_log(GLEXEC_LOG_ERROR,
				"This can not happen: "
				"parent failed to identity of caller: %s\n",
				strerror(errno));
		    glexec_exit(GLEXEC_INSUFFICIENT_PRIVILEGES,
				GLEXEC_NO_CHILD_EXIT_CODE);
	    }
	    glexec_log(GLEXEC_LOG_DEBUG,
			"gLExec is lingering as uid/euid=%u/%u\n",
			caller->uid,caller->uid);
	    break;
	case 1:
	    /* We are real/eff caller/root, need to go set real/eff target/root
	     * then eff target */
	    if (setreuid(target->uid,0)==-1 ||
		seteuid(target->uid)==-1)   {
		glexec_log(GLEXEC_LOG_ERROR,
			    "This can not happen: "
			    "parent failed to set(re/e)uid %u: %s\n",
			    target->uid,strerror(errno));
		glexec_exit(GLEXEC_INSUFFICIENT_PRIVILEGES,
			    GLEXEC_NO_CHILD_EXIT_CODE);
	    }
	    glexec_log(GLEXEC_LOG_DEBUG,
			"gLExec is lingering as uid/euid=%u/%u\n",
			target->uid,target->uid);
	    break;
	case 2: /* Nothing to change: we have no superpowers */
	    break;
	default: /* We handled them all, can not happen */
	    glexec_log(GLEXEC_LOG_ERROR,
			"Coding error: linger status has unexpected value.\n");
	    glexec_exit(GLEXEC_UNCLASIFIED,
			GLEXEC_NO_CHILD_EXIT_CODE);
	    break;
    }
}

/*!
 * Closes all open file descriptors except STDIN, STDOUT and STDERR. Close the
 * log during the closing, since we don't know it's fd.
 */
void glexec_close_fds(void) {
    int rc;
    sigset_t old_sigset;

    /* temporarily block signals */
    glexec_block_signals(&old_sigset);
    /* Close log to prevent if from being closed by glexec_closefrom() */
    glexec_close_log();
    /* Close everything >stderr */
    rc=glexec_closefrom(3);
    /* Reopen log */
    glexec_open_log();
    /* restore signals */
    glexec_sigprocmask(&old_sigset);

    if (rc)
	glexec_log(GLEXEC_LOG_WARNING,
		   "Closing of all filedescriptors >=3 failed: %s\n",
		   strerror(errno));
}

/************************************************************************/
/* INTERNAL FUNCTIONS                                                   */
/************************************************************************/

/*!
 *  Check if the real user name is in the whitelist specified in the
 *  configuration file. A whitelist entry can be of the form:
 *   * - all users
 *   .<pool>
 *   <pool>* - where * is purely numerical
 *   <username> - no stars allowed
 *  \param user  user name.
 *  \return 1 in case the user is allowed to execute glexec, 0 otherwise.
 */
static int glexec_uname_allowed(const char *user)
{
  char *wl_entry, **wl_list;
  int entry=0;
  size_t n;


  if ( !(wl_list = glexec_convert_csv_to_list(
		  glexec_conf_value("glexec:user_white_list")))) 
  {
    glexec_log(GLEXEC_LOG_WARNING, "no user_white_list is configured.\n");
    return 0;
  }

  wl_entry=wl_list[entry];

  while (wl_entry) {
    n = strlen(wl_entry);

    if (n > 0 ) {
      glexec_log(GLEXEC_LOG_DEBUG,
	      "checking user name: %s = %s\n", user, wl_entry);

      if (wl_entry[0] == '.') { /* Poolaccount */
	/* We have a pool account formatted like the pool accounts in gridmaps
	 * files. */
        glexec_log(GLEXEC_LOG_DEBUG,
	    "  poolaccount found, because '%s' starts with '.'.\n", wl_entry);

        if (wl_entry[n-1] == '*') { /* Poolaccount ending with * */
	    glexec_log(GLEXEC_LOG_WARNING,
		"Warning: found an ambiguous entry (%s) for 'user_white_list'"\
		" in the config file. Entry is ignored.\n", wl_entry);
          continue;
        }

        if (glexec_match_pool_account_name(wl_entry+1, user, n-1)) {
          glexec_log(GLEXEC_LOG_DEBUG, "  match found.\n");

          glexec_free_list(wl_list);
          return 1;
        }
      } else if ((n==1) && (wl_entry[0] == '*')) { /* ANY account */
        glexec_log(GLEXEC_LOG_WARNING,
	    "All users are allowed to use glexec, "
	    "this might not be what you want.\n");
        glexec_free_list(wl_list);
        return 1;
      } else if (n>1 && (wl_entry[n-1]=='*')) { /* account* MAY ONLY BE POOLACCOUNT */
	/* We have a pool account formatted like <pool>* and we need to enforce
	 * that * matches numbers only.*/
        glexec_log(GLEXEC_LOG_DEBUG,
	    "  poolaccount found, because '%s' ends with '*'.\n", wl_entry);
        if (glexec_match_pool_account_name(wl_entry, user, n-1)) {
          glexec_log(GLEXEC_LOG_DEBUG, "  match found.\n");
          glexec_free_list(wl_list);
          return 1;
        }
      } else {
        /* Just a regular account and do a straightforward match on it. */
        glexec_log(GLEXEC_LOG_DEBUG, "  regular account name found.\n");
        if (strcmp(wl_entry, user)==0) {
          glexec_log(GLEXEC_LOG_DEBUG, "  match found.\n");
          glexec_free_list(wl_list);
          return 1;
        }
      }

    }
    wl_entry=wl_list[++entry];
  }
  glexec_free_list(wl_list);
  return 0;
}

/*!
 *   Check if the user executing gLExec has to present a proxy which
 *   contains a private key or that the users is allowed not to present
 *   the private key. Only when it is determined that the private key
 *   maybe omitted from the proxy, false is returned.
 *
 *  \param user  Symbolic user name.
 *
 *  \returns 1 in case the user needs to present a proxy *with* a private key,
 *           0 when the user is allowed to present a proxy without private key.
 */
static int glexec_user_needs_private_key(const char *user)
{
    char **user_list;

    if ( !(user_list = glexec_convert_csv_to_list(
	    glexec_conf_value("glexec:omission_private_key_white_list"))))
	return 1; /* NO LIST: NO MATCH */

    while (*user_list) { /* LIST: look for username */
	glexec_log(GLEXEC_LOG_DEBUG,"checking user name: %s = %s\n",
		   user, *user_list);

	if (strcmp(user, *user_list) == 0) {
	    glexec_log(GLEXEC_LOG_DEBUG, "  Match found.\n");
	    return 0; /* MATCH */
	}
	user_list++;
    }

    return 1; /* NO MATCH */
}

/*!
 * Determine and update values of the TERM and KILL delay settings and whether
 * to use setpgid() in child processes.
 * \param whether we run in linger_mode
 */
static void glexec_set_signaloptions(int linger_mode)   {
    const char *term_val_str;
    const char *kill_val_str;
    long term_val, kill_val;
    unsigned int term_delay,kill_delay;

    /* Should we set handlers commonly used in debuggers/profilers?
     * Default yes (1) */
    if ( glexec_conf_boolean("glexec:extra_sighandlers", 1) )
	glexec_set_use_extra_sighandlers(1);
    else {
	glexec_set_use_extra_sighandlers(0);
	glexec_log(GLEXEC_LOG_NOTICE ,
	    "NOT installing extra signal handlers such as SIGTRAP etc.\n");
    }

    /* If not in linger_mode we're done */
    if (!linger_mode) return;

    /* Grace time for children after SIGTERM */
    term_val_str=glexec_conf_value("glexec:term_delay");
    if ( term_val_str != NULL ) {
	errno=0;
	term_val=strtol(term_val_str,NULL,10);
	if (errno!=0)	{
	    term_delay=GLEXEC_TERM_GRACE_TIME;
	    glexec_log(GLEXEC_LOG_WARNING,
		"Cannot convert term_delay \"%s\" to an unsigned int, "
		"using default %u.\n",
		term_val_str,term_delay);
	} else if (term_val<0 || (long unsigned)term_val>UINT_MAX)  {
	    term_delay=GLEXEC_TERM_GRACE_TIME;
	    glexec_log(GLEXEC_LOG_WARNING,
		"term_delay should be [0,%u], leaving at default %u.\n",
		UINT_MAX,term_delay);
	} else	{
	    /* We have guaranteed above 0 <= term_val <=UINT_MAX */
	    term_delay=(unsigned int)term_val;
	    glexec_log(GLEXEC_LOG_DEBUG,"Grace time after SIGTERM set to %u\n",
		    term_delay);
	}
	glexec_set_term_delay(term_delay);
    }

    /* Grace time for children after SIGKILL */
    kill_val_str=glexec_conf_value("glexec:kill_delay");
    if ( kill_val_str != NULL ) {
	errno=0;
	kill_val=strtol(kill_val_str,NULL,10);
	if (errno!=0)	{
	    kill_delay=GLEXEC_KILL_GRACE_TIME;
	    glexec_log(GLEXEC_LOG_WARNING,
		"Cannot convert kill_delay \"%s\" to an unsigned int, "
		"leaving at default %u.\n",
		kill_val_str,kill_delay);
	} else if (kill_val<=0 || (long unsigned)kill_val>UINT_MAX)  {
	    kill_delay=GLEXEC_KILL_GRACE_TIME;
	    glexec_log(GLEXEC_LOG_WARNING,
		"kill_delay should be (0,%u], leaving at default %u.\n",
		UINT_MAX,kill_delay);
	} else {
	    /* We have guaranteed above 0 <(=) kill_val <=UINT_MAX */
	    kill_delay=(unsigned int)kill_val;
	    glexec_log(GLEXEC_LOG_DEBUG,"Grace time after SIGKILL set to %u\n",
		    kill_delay);
	}
	glexec_set_kill_delay(kill_delay);
    }

    /* Whether to use setpgid(0,0) to start a new process group in the payload
     * and epilogue child processes, and then send TERM and KILL signals to the
     * child process group instead of the child process.
     * Default yes (1) */
    if ( glexec_conf_boolean("glexec:use_setpgid", 1) )	{
	glexec_set_use_setpgid(1);
	glexec_log(GLEXEC_LOG_DEBUG,
		    "Will use separate process group for payload.\n");
    } else {
	glexec_set_use_setpgid(0);
	glexec_log(GLEXEC_LOG_INFO,
		"Disabling use of setpgid() for starting new process group "
		"in child processes\n");
    }

    /* Should we close stdin for payload in case stdin==tty. If specified yes,
     * we can safely set it, if specified no, leave it untouched because user
     * might have specified -b (yes) ?
     * Default no (0) */
    if ( glexec_conf_boolean("glexec:force_payload_background", 0) )  {
	glexec_set_payload_should_run_in_background(1);
	glexec_log(GLEXEC_LOG_INFO,
	    "Payload will be forced into the background: "
	    "when stdin is a tty payload will get /dev/null.\n");
    }

}

/*!
 * Determine and set linger_as_payload flag
 * \return actual value found
 */
static int glexec_get_linger_as_payload(void)  {
    static int linger_as_payload=-1;

    /* If initialized, return it now */
    if (linger_as_payload>-1)   
        return linger_as_payload;

    if (!glexec_posix_enf_by_glexec() ||
        !glexec_is_switching()) /* Non-switching mode */
        linger_as_payload=2;
    else    { /* saved setuid root mode */
	/* Whether lingering gLExec is target (unsafe)
	 * Default no (0) */
	linger_as_payload=glexec_conf_boolean("glexec:linger_as_payload", 0);
	if (linger_as_payload)
            glexec_log(GLEXEC_LOG_WARNING,
		       "Lingering as payload user. NOTE: this is insecure!\n");
    }
    return linger_as_payload;
}

/*!
 * Determines and returns specified locking mechanism for reading input proxies,
 * or its default value.
 * \return locking mechanism
 */
static GLEXEC_LOCK_MECH_T glexec_get_input_lock_mechanism(void)   {
    const char *value=glexec_conf_value("glexec:input_lock_mechanism");

    /* Change the mechanism used to lock the input proxy files */
    if (!value || strcasecmp(value, "flock") == 0)  {
        glexec_log(GLEXEC_LOG_DEBUG,
	    "Using \"flock()\" file locking mechanism to read the "
	    "proxy files at the (default) $GLEXEC_SOURCE_PROXY and "
	    "$GLEXEC_CLIENT_CERT locations.\n");
	return GLEXEC_LOCK_FLOCK;
    }

    /* fcntl */
    if (strcasecmp(value, "fcntl") == 0)    {
        glexec_log(GLEXEC_LOG_DEBUG,
	    "Using \"fcntl()\" file locking mechanism to read the proxy "
	    "files at the (default) $GLEXEC_SOURCE_PROXY and "
	    "$GLEXEC_CLIENT_CERT locations.\n");
	return GLEXEC_LOCK_FCNTL;
    }

    /* disable */
    if (strcasecmp(value, "disabled") == 0) {
        glexec_log(GLEXEC_LOG_INFO,
	    "disabled file locking mechanism "
	    " to read the proxy files at the (default) "
	    "$GLEXEC_SOURCE_PROXY and $GLEXEC_CLIENT_CERT locations.\n");
	return GLEXEC_LOCK_DISABLED;
    }
   
    /* error */
    glexec_log(GLEXEC_LOG_ERROR,
	"Unknown file locking mechanism specified for input proxy files.\n");
    glexec_exit(GLEXEC_CONF_FAIL_LOCK_MECH, GLEXEC_NO_CHILD_EXIT_CODE);

    return GLEXEC_LOCK_UNKNOWN;
}

/*!
 * Determines and returns specified locking mechanism for writing target proxy,
 * or its default value.
 * \return locking mechanism
 */
static GLEXEC_LOCK_MECH_T glexec_get_target_lock_mechanism(void)  {
    /* Locking mechanism destination proxy */
    /* Possibilities: old only, new only, old&new, neither */
   
    /* check new syntax */
    const char *value=glexec_conf_value("glexec:target_lock_mechanism");
    /* check old syntax */
    const char *valueold=glexec_conf_value("glexec:lock_mechanism");

    /* check for old syntax: if also new, use that, otherwise, use old */
    if (valueold)   { /* if at least old: warning */
	glexec_log(GLEXEC_LOG_WARNING,
		"Warning: Using deprecated syntax \'lock_mechanism\', " \
		"please use \'target_lock_mechanism\' for target proxy " \
		"and \'input_lock_mechanism\' for input proxies.\n");
	if (value) /* if both are set: warning */
	    glexec_log(GLEXEC_LOG_WARNING,
		"Warning: Found both \"target_lock_mechanism\" and " \
		"\"lock_mechanism\", ignoring \"lock_mechanism\".\n");
	else /* if only old -> use old */
	    value=valueold;
    }

    /* Now parse the value if it is set */
    /* Flock */
    if (!value || strcasecmp(value, "flock") == 0)  {
        glexec_log(GLEXEC_LOG_DEBUG,
		"Using \"flock()\" file locking mechanism to write the proxy "\
		"file at the (default) $GLEXEC_TARGET_PROXY location.\n");
	return GLEXEC_LOCK_FLOCK;
    }
    /* fcntl */
    if (strcasecmp(value, "fcntl") == 0)    {
        glexec_log(GLEXEC_LOG_DEBUG,
		"Using \"fcntl()\" file locking mechanism to write the proxy "\
		"file at the (default) $GLEXEC_TARGET_PROXY location.\n");
	return GLEXEC_LOCK_FCNTL;
    }
    /* disabled */
    if (strcasecmp(value, "disabled") == 0) {
        glexec_log(GLEXEC_LOG_NOTICE,
		"disabled file locking mechanism to write the proxy file at "\
		"the (default) $GLEXEC_TARGET_PROXY location.\n");
	return GLEXEC_LOCK_DISABLED;
    }

    /* Unknown: error */
    glexec_log(GLEXEC_LOG_ERROR,
	"Unknown file locking mechanism specified for destination proxy file.\n");
    glexec_exit(GLEXEC_CONF_FAIL_LOCK_MECH, GLEXEC_NO_CHILD_EXIT_CODE);

    return GLEXEC_LOCK_UNKNOWN; /* this is never reached */
}

/*!
 * Determines whether we want the pedantic and no-symlink exec checks and then
 * carries them out. Exits on failure.
 * \param caller
 * \param target
 * \param fullcmd absolute path executable
 * \param realcmd canonical absolute path executable
 * \param target_cwd current working directory
 * \param exe_info stat information on executable
 */
static void glexec_pedantic_checks(const char *target_cwd,
			    char *fullcmd, char *realcmd,
			    glexec_account_t *caller, glexec_account_t *target,
			    struct stat *exe_info)	{
    struct stat lnk,dir_info;

    /* Prohibit symlink checks
     * Default no (0) */
    if ( glexec_conf_boolean("glexec:prohibit_exec_via_symlink", 0) )  {
	/* Get stat for program, first with lstat */
	if ((lstat(fullcmd, &lnk)) != 0) {
	    glexec_log(GLEXEC_LOG_ERROR,
		"cannot lstat program %s: %s\n", fullcmd, strerror(errno));
	    glexec_exit(GLEXEC_STAT_ERROR_EXECUTABLE_CMD,
			GLEXEC_NO_CHILD_EXIT_CODE);
	}
	/* Error out if the program is symlinked - and... explicitly set to
	 * check the cmd at symlinked or not. */
	if (S_ISLNK(lnk.st_mode)) {
	    glexec_log(GLEXEC_LOG_ERROR,
		"stat indicates that program is symlinked: (%s)\n", fullcmd);
	    glexec_exit(GLEXEC_STAT_ERROR_CMD_SYMLINKED,
			GLEXEC_NO_CHILD_EXIT_CODE);
	}
    }

    /* Pedantic checks
     * Default no (0) */
    if ( glexec_conf_boolean("glexec:pedantic_security_checks", 0) )   {
	glexec_log(GLEXEC_LOG_DEBUG, "Using \"pedantic_security_checks\".\n");
	/* Error out if the program is writable by others. */
	if (exe_info->st_mode & S_IWOTH) {
	    glexec_log(GLEXEC_LOG_ERROR,
		"file %s (=%s) is writable by others.\n", fullcmd,realcmd);
	    glexec_exit(GLEXEC_WRONG_PERMISSION_ON_CMD,
			GLEXEC_NO_CHILD_EXIT_CODE);
	}
	/* Get stat for directory */
	if (stat(target_cwd, &dir_info) != 0) {
	    glexec_log(GLEXEC_LOG_ERROR, "cannot stat current dir: %s.\n",
			strerror(errno));
	    glexec_exit(GLEXEC_STAT_ERROR_CURRENT_WORK_DIR,
			GLEXEC_NO_CHILD_EXIT_CODE);
	}
	/* Check ownership of exe/dir based on the mode: in switching mode it
	 * should match the target, otherwise it should match the caller */
	if (glexec_is_switching())   {
	    /* Error out if the uid of the executable is different from the
	     * caller, target and root uid. */
	    if ( caller->uid != exe_info->st_uid &&
		 target->uid != exe_info->st_uid &&
			   0 != exe_info->st_uid)  {
		glexec_log(GLEXEC_LOG_ERROR,
		    "uid of program (%u) does not match with that of "\
		    "either caller user (%u), target user (%u) or root.\n",
		    exe_info->st_uid, caller->uid, target->uid);
		glexec_exit(GLEXEC_PROG_WRONG_OWNERSHIP,
			    GLEXEC_NO_CHILD_EXIT_CODE);
	    }
	    /* Error out if the uid of the target cwd is different from the
	     * caller, target and root uid. */
	    if ( caller->uid != dir_info.st_uid && 
		 target->uid != dir_info.st_uid &&
			   0 != dir_info.st_uid)	{
		glexec_log(GLEXEC_LOG_ERROR,
		    "uid of directory (%u) does not match with that of "\
		    "either caller user (%u), target user (%u) or root.\n",
		    dir_info.st_uid, caller->uid, target->uid);
		glexec_exit(GLEXEC_DIR_WRONG_OWNERSHIP,
			    GLEXEC_NO_CHILD_EXIT_CODE);
	    }
	} else {
	    /* Error out if the uid of the executable is different from the
	     * caller/target and root uid. */
	    if ( caller->uid != exe_info->st_uid && 0 != exe_info->st_uid)  {
		glexec_log(GLEXEC_LOG_ERROR,
		    "uid of program (%u) does not match with that of " \
		    "either target=caller user (%u) or root.\n",
		    exe_info->st_uid, caller->uid);
		glexec_exit(GLEXEC_PROG_WRONG_OWNERSHIP,
			    GLEXEC_NO_CHILD_EXIT_CODE);
	    }
	    /* Error out if the uid of the target cwd is different from the
	     * caller/target and root uid. */
	    if ( caller->uid != dir_info.st_uid && 0 != dir_info.st_uid)    {
		glexec_log(GLEXEC_LOG_ERROR,
		    "uid of directory (%u) does not match with that of "\
		    "either target=caller user (%u) or root.\n",
		    dir_info.st_uid, caller->uid);
		glexec_exit(GLEXEC_DIR_WRONG_OWNERSHIP,
			    GLEXEC_NO_CHILD_EXIT_CODE);
	    }
	}
    }
}
